// Keybindings.js -- code related to customizable keybindings

// Globals
var gKeybindingMgr= null;

/* The keybindingManager controls all things keybinding-related.

    It manages:
        - configurations (sets of keybindings)
        - Which keys map to what commands
        - What commands are parametrized
        - What menuitems have what what keyboard shortcuts (acceltexts)
        
    It doesn't yet manage:
        - Tooltips
    
    A few definitions:
        - Command ID: a string, the id field of the <command> element -- e.g., 'cmd_undo'
        - key: a given key id (see keys.xul), e.g. "Ctrl-A"
        - key sequence: an array of keys e.g. ['Ctrl-A', 'Ctrl-K']
        - key label: the textual representation of a key -- it's in fact the same as the key ID
        - key sequence label: the textual representation of a key sequence -- keys are separated by ', ', e.g. "Ctrl-A, Ctrl-K"
        - configuration: A set of bindings for the factory-shipped commands (those defined in XUL at this point).
          These come in two varieties -- factory defined and custom.  The former are not modifiable by the user, the latter
          are by definition written by the user.
          
      
    API:
        Apart from the initialization, and the configuration management (done by pref-keys.js), most of the calls are to:
        
            AssignKey
            makeKeyActive
            unlabelMenuItem
            clearBinding
            
        The keybindingManager is also Observing Notify's on kb-unload, to known when to clear out 'layered' bindings.
            The "subject" of the Notify has to be the part (XXX should move to a part-independent data structure)
        

 
    Internal datastructures for those reading the code:
  
        Configuration management:
            - currentConfiguration: the currently active configuration
            - _knownconfigs: a list of know configurations (those which have a pref, show up in the menulist, etc.)
            - _factoryConfigurations: a subset of _knownconfigs which corresponds to the factory defaults (these days just one)
            
        Keys:
            - The command2key mapping maps command IDs (as in "cmd_undo") to key sequence labels (as in "Ctrl-Z")
            - The keyTree is a tree of all of the current key/command assignments, so that prefix keys map to dictionaries of
              key->command or key->subtree relationships.
            - _commandParams: a mapping of command IDs to a mapping of key sequence labels to parameters.
                e.g. if Ctrl-F -> runMacro, macro1 and
                       Ctrl-G -> runMacro, macro2 then
                    _commandParams['runMacro']['Ctrl-F'] = 'macro1'
                and _commandParams['runMacro']['Ctrl-G'] = 'macro2'
                ** A command is a parametrized command if it is a key of _commandParams -- or at least there's no other way to 
                   guess from inside this code.
                   
            - commanditems: used only for the display of known keybindings -- should move away from this class in a utility
              function
    
            - command2key: maps the commandId to a list of key sequence labels (the last assignment goes in front)          
        
    
*/

function keybindingManager() {
    this.log = getLoggingMgr().getLogger('keybindingManager');
    this.commanditems = null; // XXX
    this.command2key = {}; // what key for a command?
    this.key2command= {}; // what command assigned to a key?
    this.keyTree = {};
    this.activeCommands = {};
    var line, i;
    this.prefset = gPrefSvc.prefs;
    this._knownconfigs = [];
    this.keybindingSchemeService = Components.classes['@activestate.com/koKeybindingSchemeService;1'].getService();
    var schemes = new Array();
    this.keybindingSchemeService.getSchemeNames(schemes, new Object());
    this._knownconfigs = schemes.value;
    this.currentConfiguration = gPrefSvc.prefs.getStringPref('keybinding-scheme');
    this.currentScheme = this.keybindingSchemeService.getScheme(this.currentConfiguration);
    //dump('got currentConfiguration = ' + this.currentConfiguration + '\n');
    
    this._commandParams = {}
    this.currentPrefixString = '';
    this.commanditems = null;
    this.document = document;
    var observerSvc = Components.classes["@mozilla.org/observer-service;1"].
        getService(Components.interfaces.nsIObserverService);
    // This observer is never removed, but it doesn't matter - very light-weight and no cycles.
    observerSvc.addObserver(this, "kb-unload",false);    
    observerSvc.addObserver(this, "kb-load",false);    
}

keybindingManager.prototype.constructor = keybindingManager;

keybindingManager.prototype.observe = function(part, topic, data) {
    // Two notifications are observed by the keybinding manager -- kb-load and kb-unload.
    // see the notification HOWTO for details
    // XXX This is relatively slow, due to a large number of XPCOM calls back to the part
    // Some ideas:
    //   - pass more data in the notification, e.g. using tab-delimited
    //   - move to a uniform 'callNamedPart' model where only the kb needs to be sent
    
    if (topic == 'kb-unload' || topic == 'kb-load') {
        gTimeline.startTimer('kb-load');
        //dump("data = " + data  +'\n');
        //dump("got unload information about part: " + part.type + '\n');
        //if (part.name) { dump("part.name = " + part.name + '\n') };
        if (gKeybindingMgr && part.hasAttribute('keyboard_shortcut')) {
            var keylabels = part.getStringAttribute('keyboard_shortcut').split('###');
            var keylabel;
            for (i in keylabels) {
                keylabel = keylabels[i];
                if (keylabel != '') {
                    //dump('keylabel = ' + keylabel + '\n');
                    keysequence = keylabel.split(', ')
                    //dump('keysequence[0] = ' + keysequence[0] + '\n');
                    if (typeof(part.shortcut_command_attribute) != 'undefined') {
                        var parameter = part.getStringAttribute(part.shortcut_command_attribute);
                        if (parameter) {
                            if (topic=='kb-unload') {
                                gKeybindingMgr.clearBinding(part.shortcut_command, parameter, true);
                            } else {
                                gKeybindingMgr.assignKey(part.shortcut_command, keysequence, parameter);
                                gKeybindingMgr.makeKeyActive(part.shortcut_command, keysequence);
                            }
                        }
                    } else {
                        this.log.warn("Error removing keyboard shortcut for " + part);
                    }
                }
            }
        }
        gTimeline.stopTimer('kb-load');
        gTimeline.markTimer('kb-load');
    }
}

keybindingManager.prototype.getConfigurations = function () {
    return this._knownconfigs;
}

keybindingManager.prototype.loadConfiguration = function (configName, forceReload /* = false */) {
    try {
        gTimeline.enter('keybindingManager.loadConfiguration');
        if (typeof(forceReload) == 'undefined') forceReload = false;
        //dump("loading from " + configName + '\n');
        var data = this.keybindingSchemeService.getScheme(configName).data;
        this._clearActiveBindings();
        this.clearBindings();
        this.parseConfiguration(data, forceReload);
        this.activeCommands = cloneObject(this.command2key);
        this.currentPrefixMap = this.keyTree;
        this._configDirty = false;
        this._configUnsaved = false;
        this.currentConfiguration = configName;
        this.currentScheme = this.keybindingSchemeService.getScheme(this.currentConfiguration);
        this._configKeyTree = cloneObject(this.keyTree);
        gTimeline.leave('keybindingManager.loadConfiguration');
    } catch (e) {
        log.exception(e);
    }
}

function cloneObject(what) {
    var f = new Object();
    for (var i in what) {
        if (typeof what[i] == 'object') {
            f[i] = new cloneObject(what[i]);
        }
        else
            f[i] = what[i];
    }
    return f;
}

keybindingManager.prototype.parseConfiguration = function (data, forceReload /* =false */) {
    gTimeline.enter('keybindingManager.parseConfiguration');
    var i, line;
    var lines = data.split('\n')
    var onlywhitespace = /^\s*$/;
    if (typeof(forceReload) == 'undefined') forceReload = true; // false; XXX
    
    // Find out if we are binding keys, or if we've gotten our keybindings using
    // the Mozilla persist mechanism
    var keyset = document.getElementById('widekeyset');
    var bindingKeys = true;
    //if (!forceReload && keyset.hasAttribute('persisted_kb') && keyset.getAttribute('persisted_kb') == 'true') {
    //    this.log.info("Binding Keys is false - we got them from persisted XUL");
    //    bindingKeys = false;
    //} else {
    //    this.log.info("Binding Keys is true - we're doing it ourselves");
    //}
    onlywhitespace = onlywhitespace.compile(/^\s*$/); // this seems clumsy, but appears to be the syntax required??
    for (i in lines) {
        line = lines[i];
        //dump("looking at: " + line +'\n');
        if (onlywhitespace.test(line)) continue;
        if (line[0] == '#') continue;
        var re = /(\w+)\s+(.*)\s*$/;  // Two groups - 'kind' and 'rest', separated by whitespace
        var groups = re.exec(line);
        var desc;
        var commandId;
        var keysequence;
        switch (groups[1]) {
            case 'version':
                if (groups[2] != '1') this.log.error("unknown version number -- continuing anyway\n");
                break;
            case 'binding':
                //dump("found a binding"+ line +'\n');
                desc = groups[2].replace(/\s*$/, ''); // rstrip
                var parts = desc.split(/\s+/);
                commandId = parts[0];
                keysequence = parts.splice(1, parts.length);
                //dump("command = " + commandId + " bound to key sequence " + keysequence + '\n');
                // Note that keysequences can have spaces in the case of multi-key sequences -- e.g. "Ctrl-X Ctrl-C"
                this.assignKey(commandId, keysequence);
                if (bindingKeys) {
                    this.makeKeyActive(commandId, keysequence);
                }
                break;
            case 'parambinding':
                // Now we're dealing with a "parametrized" keybinding
                desc = groups[2].replace(/\s*$/, ''); // rstrip
                var firstQuote = desc.indexOf('"');
                var lastQuote = desc.lastIndexOf('"');
                commandId = desc.slice(0, firstQuote).replace(/\s*$/, '');
                var parameter = desc.slice(firstQuote+1, lastQuote);
                keysequence = desc.slice(lastQuote+1, desc.length).replace(/^\s*/, '').split(/\s+/);
                this.assignKey(commandId, keysequence, parameter);
                if (bindingKeys) {
                    this.makeKeyActive(commandId, keysequence);
                }
                break;
            default:
                this.log.warn("Unknown command in keybinding configuration:" + groups[1]);
                
        }
    }
    if (bindingKeys) {
        this.log.info("setting persisted_kb to true!");
        keyset.setAttribute('persisted_kb', 'true');
    }
    gTimeline.leave('keybindingManager.parseConfiguration');
}

keybindingManager.prototype.learnParameter = function (commandId, parameter, keylabel) {
    // Parameters are the parameters to "parametrized commands" -- i.e. those that require an argument
    // used e.g. in "Run_NamedCommand" and "Execute_NamedMacro"
    if (!(commandId in this._commandParams)) {
        this._commandParams[commandId] = {};
    }
    //dump("Associating parameter " + parameter + " for " + commandId + " with " + keylabel +'\n');
    this._commandParams[commandId][keylabel] = parameter;
}

keybindingManager.prototype._configKnown= function (newconfigname) {
    for (var i in this._knownconfigs) {
        if (this._knownconfigs[i] == newconfigname) return true;
    }
    return false;
}

keybindingManager.prototype.makeNewConfiguration = function (newconfigname, prefset) {
    try {
        this.prefset = prefset;
        var alreadyExists = this._configKnown(newconfigname);
        if (alreadyExists) {
            var question = "The configuration \""+newconfigname+"\" is "+
                           "already defined.  Do you wish to overwrite the "+
                           "existing configuration?";
            if (dialog_yesNo(question) == "No") { // we bail
                return '';
            }
        } else {
            this._knownconfigs.push(newconfigname);
        }
        this.currentConfiguration = newconfigname;
        this.currentScheme = this.currentScheme.clone(newconfigname);
        this._configDirty = true;
        this._configUnsaved = true;
    } catch (e) {
        log.exception(e);
    }
    return newconfigname;
}

keybindingManager.prototype.deleteConfiguration = function (configname, prefset) {
    try {
        if (! this._configKnown(configname)) {
            alert("Error deleting configuration: " + configname);
            this.log.error('KeybindingManager: Attempting to delete unknown configuration: ' + configname);
            return;
        }
        // Remove the configuration from the list of known configurations
        for (var i = this._knownconfigs.length-1; i > 0; i--) {
            if (this._knownconfigs[i] == configname) {
                this._knownconfigs.splice(i,1);
                break;
            }
        }
        this.prefset = prefset
        // Remove the preference
        this.prefset.deletePref(configname);
        var scheme = this.keybindingSchemeService.getScheme(configname);
        try {
            scheme.remove();
        } catch (e) {
            alert("Error removing scheme: " + e);
        }
        var schemes = new Array();
        this.keybindingSchemeService.getSchemeNames(schemes, new Object());
        this._knownconfigs = schemes.value;
        
        this._configDirty = false; // it's gone, but clean!
        this.switchConfiguration(this._knownconfigs[0], prefset);
        // Reset the pref corresponding to the set of known configurations
        this._saveKnownConfigs();
    } catch (e) {
        log.exception(e);
    }
}

keybindingManager.prototype.makeDirty= function() {
    this._configDirty = true;
}

keybindingManager.prototype.saveCurrentConfiguration = function() {
    if (!this.configurationWriteable(this.currentConfiguration)) {
        return;  // We don't bother saving factory default configs.
    }
    var text = '';
    var label, param, keydata;
    for (var commandname in this.command2key) {
        if (commandname in this._commandParams) {
            for (label in this._commandParams[commandname]) {
                param = '"' + this._commandParams[commandname][label].replace('"', '\\"', 'g') +'"';
                text = text.concat('parambinding ' + commandname + ' ' + param + ' ' + label.replace(', ', ' ', 'g') + '\n');
            }
        } else {
            keydata = this.command2key[commandname];
            for (i = 0; i < keydata.length; i++) {
                label = keydata[i];
                //dump("saving: commandname = " + commandname + " label = " + label + '\n');
                text = text.concat('binding ' + commandname + ' ' + label.replace(', ', ' ') +'\n');
            }
        }
    }
    this.currentScheme.data = text;
    this.currentScheme.save();
    this._configDirty = false;
}

keybindingManager.prototype.configurationWriteable = function(configname) {
    var scheme = this.keybindingSchemeService.getScheme(configname);
    return scheme.writeable;
}

function isAssociativeArray(obj) {
    return ((typeof(obj) == 'object') && (typeof(obj.length) == 'undefined'));
}

function isArray(obj) {
    return ((typeof(obj) == 'object') && (typeof(obj.length) == 'number'));
}

keybindingManager.prototype.applyCurrentConfiguration = function() {
    this.walk_and_apply(this.keyTree, []);
}

keybindingManager.prototype.walk_and_apply = function(root, keysequence) {
    // walk the prefixTree, calling makeKeyActive each time
    for (var leaf in root) {
        //dump("leaf = " + leaf +'\n');
        switch (typeof(root[leaf])) {
            case 'string':
                // XXX To review when layered commands are in
                keysequence.push(leaf);
                //if (keysequence) {
                //    dump("making active " + root[leaf] + ': ' + keysequence +'\n');
                //}
                this.makeKeyActive(root[leaf], keysequence);
                keysequence.pop();
                break;
            case 'object':
                if (isAssociativeArray(root[leaf])) {
                    //dump("found tree under " + leaf + '\n');
                    keysequence.push(leaf);
                    this.walk_and_apply(root[leaf], keysequence);
                    keysequence.pop();
                    break;
                }
                // fall through
            default:
                //dump("found a " + typeof(root[leaf]) + ' for ' + leaf +'\n');
                break;
        }
    }
}

keybindingManager.prototype._clearActiveBindings= function() {
    gTimeline.enter('keybindingManager._clearActiveBindings');
    if (! this.activeCommands) {
        //dump("null activeCommands\n");
        return;
    }
    //dump("********* clearActiveBindings called \n");
    //dump("activeCommands = " + this.activeCommands + '\n');
    for (var command in this.activeCommands){
        //dump("CLEARING " + command +'\n');
        this.clearBinding(command, false);
    }
    this.activeCommands = null;
    gTimeline.leave('keybindingManager._clearActiveBindings');
}

keybindingManager.prototype.revertToPref = function(configname) {
    if (this._configUnsaved) {
        var configs = this._knownconfigs;
        this._knownconfigs = []
        for (var i = 0; i < configs.length; i++) {
            if (configs[i] != this.currentConfiguration) {
                this._knownconfigs.push(configs[i]);
            }
        }
    }
    gPrefSvc.prefs.setStringPref('keybinding-scheme', configname);
    this.currentConfiguration = configname;
    this.currentScheme = this.keybindingSchemeService.getScheme(this.currentConfiguration);
    this.loadConfiguration(this.currentConfiguration);
}

keybindingManager.prototype.saveAndApply = function(prefset) {
    try {
        var keybindings = '';
        var i, keybinding;
        var dirty = false;
        this.prefset = prefset;
        if (prefset.getStringPref('keybinding-scheme') != this.currentConfiguration) {
            prefset.setStringPref('keybinding-scheme', this.currentConfiguration);
            dirty = true;
        }
        if (dirty || this._configDirty) {
            alert("Changes to the keybindings won't fully take effect until Komodo is restarted.");
        }
        this.saveCurrentConfiguration();
        this._saveKnownConfigs();
        this.loadConfiguration(this.currentConfiguration, true);
        var observerSvc = Components.classes["@mozilla.org/observer-service;1"].
              getService(Components.interfaces.nsIObserverService);
        observerSvc.notifyObservers(this, 'parts_reload', '');
    } catch (e) {
        log.exception(e);
    }
}

keybindingManager.prototype._saveKnownConfigs = function() {
    var keybinding;
    var keybindings = '';
    for (i in this._knownconfigs) {
        keybinding = this._knownconfigs[i];
        keybindings += keybinding +'\n';
    }
    if (keybindings) {
        this.prefset.setStringPref('keybindings', keybindings);
        //dump('saving : keybindings -> ' + keybindings +'\n');
    } else {
        this.log.error("Tried to save empty keybindings pref!");
    }
}

// On switching of configuration names, check to see if current configuration needs to be saved

keybindingManager.prototype.offerToSave= function () {
    if (this._configDirty
        && dialog_yesNo("Changes to the current keyboard configuration have "
                        +"not been saved yet.  Do you wish to save them?")
           == "Yes"
       )
    {
        this.saveCurrentConfiguration();
    }
}

keybindingManager.prototype.switchConfiguration= function (newconfigname) {
    if (newconfigname == this.currentConfiguration) return true; // Nothing to do
    if (!this._configKnown(newconfigname)) {
        alert("Unknown Keyboard configuration: " + newconfigname);
        this.log.error("Unknown Keyboard configuration: " + newconfigname);
        return false;
    }
    this.offerToSave()
    try {
        this.loadConfiguration(newconfigname, true);
    } catch (e) {
        this.log.error(e);
    }
    return true;
    //dump("switched to " + this.currentConfiguration+'\n');
}

keybindingManager.prototype.clearBindings = function () {
    this.keyTree = {};  // a sequence of key sequences
    this.command2key = {};  // a hash between commands and keys
    this.key2command = {};  // a hash from key to command
    this.activeCommands = {};
}

function keysequence2keylabel(keysequence) {
    if (!isArray(keysequence)) {
        this.log.error("keysequence2keylabel has to be called with a sequence of key labels\n");
    }
    var keyname = '';
    var part;
    for (i in keysequence) {
        part = keysequence[i];
        if (part[part.length-1] == '') part += '+';  // e.g. Ctrl--
        if (i == 0) keyname = part;
        else keyname += ', ' + part;
        //dump("part = " + part + '\n');
    }
    return keyname
}

function keylabel2keysequence(keylabel) {
    var keysequence = [];
    if (typeof(keylabel) != 'string') {
        this.log.error("keylabel2keysequence should be caled with a string, not with a" + typeof(keylabel) + " such as " + keylabel) 
    }
    var parts = keylabel.split(', ');
    return parts;
}

keybindingManager.prototype.getKey = function(keyname) {
    this.log.info("doing getKey for " + keyname);
    // XXX consider using this.key etc. to avoid creation of 4 variables on each key assignment.
    // This function is called once for each key binding at startup -- it should be as fast as possible.
    var oldkey = this.document.getElementById(keyname);
    if (oldkey) {
        this.log.debug("found old key with right id!");
        return oldkey;
    }
    if (keyname[keyname.length-1] == ',') {
	keyname = keyname.slice(0,keyname.length-1);
    }
    oldkey = this.document.getElementById(keyname);
    if (oldkey) {
        this.log.debug("found old key with right id!");
        return oldkey;
    }
    this.log.error("unknown key for keyname:" + keyname)
    return null;
}

function dumpKey(key) {
    var i;
    dump('<key ');
    for (i = 0; i < key.attributes.length; i++) {
        dump(key.attributes[i].nodeName+'="'+key.attributes[i].nodeValue+'" ');
    }
    dump('/>');    
}

//utility function that I don't need at this time.
//function DOMdiffer(a, b) {
//    if (a.nodeName != b.nodeName) return true;
//    if (a.attributes.length != b.attributes.length) return true;
//    var i;
//    for (i = 0; i < a.attributes.length; i++) {
//        if (a.attributes[i].nodeName != b.attributes[i].nodeName) return true;
//    }
//    return false;
//}

keybindingManager.prototype.clearBinding = function(commandId, commandParam, restore) {
    var root;
    var i, j, k;
    var keylabel;
    var param = '';
    if (typeof(commandParam) != 'undefined') {
        param = commandParam;
    }
    var parametrized = commandId in this._commandParams;
    var text;
    if (!(commandId in this.command2key)) {
        //dump("-- clearing binding for " + commandId + " but no binding exists \n");
        //text = commandId;
        //if (commandParam) {
        //    text += ':' + commandParam;
        //}
        //dump("trying to clear the binding for a command w/ no bindings: " + text)
        return;
    }
    //dump("Clearing " + commandId + " , " + commandParam  +'\n');

    var kbs = this.command2key[commandId];
    for (k = this.command2key[commandId].length-1 ; k >= 0; k--) {
        keylabel = this.command2key[commandId][k];
        if (typeof(keylabel) == 'undefined') {
            continue;
        }
        if (parametrized && this._commandParams[commandId][keylabel] != param) {
            // skip parametrized commands not bound to this parameter
            continue;
        }
        if (keylabel) {
            this.unAssignKey(keylabel);
        }
        this.clearSequence(commandId, keylabel, restore);
    }
    if (!parametrized && typeof(keylabel) != 'undefined') {
        // if this is not a parametrized command, then there should be no more key bindings
        // for this command.  If it is, then there may still be some.
        //if (this.command2key[commandId].length != 0) {
        //    this.log.error("After clearing: this.command2key["+commandId+"] = " + this.command2key[commandId] +'\n');
        //}
    }
    this.unlabelMenuItem(commandId);  // XXX restrict to unparametrized commands?
};

keybindingManager.prototype.unAssignKey= function(keylabel) {
    var keysequence = keylabel.split(', ')
    if (keysequence.length == 1) { // We need to do this, but only if this is the _last_ command with that prefix. ugh.
        var keyname = keysequence[0]; // for purpose of assignment, we bind the command to the first key
        var oldkey = document.getElementById(keyname);
        var key = this.getKey(keyname);
        if (key.hasAttribute('command')) {
            key.removeAttribute('command');
        }
        if (key.hasAttribute('oncommand')) {
            key.removeAttribute('oncommand');
        }
    }
}
keybindingManager.prototype.clearSequence = function(commandId, keylabel, restore) {
    if (typeof(keylabel) != 'string') {
        this.log.error("clearSequence should be caled with a string, not with a" + typeof(keylabel) + " such as " + keylabel) 
    }
    var root;
    var i, j;
    var keysequence;
    root = this.keyTree;
    keysequence = keylabel2keysequence(keylabel);
    for (i=0; i < keysequence.length - 1; i++) {
        
        if (!isAssociativeArray(root)) {
            return;
        }
        root = root[keysequence[i]];
    }
    if (root[keysequence[keysequence.length-1]] != commandId) {
        return;
    } else {
        delete root[keysequence[keysequence.length-1]];
    }
    var newkeylabels = []
    for (j = this.command2key[commandId].length-1; j >= 0; j--) {
        if (this.command2key[commandId][j] == keylabel) {
            this.command2key[commandId].slice(j,1)
            // we need to delete the binding in the keyTree as well.
        } else {
            newkeylabels.push(this.command2key[commandId][j]);
        }
    }
    //if (this.command2key[commandId].length != newkeylabels.length +1) {
    //    dump("DIDN't remove a binding!" + keylabel);
    //}
    this.command2key[commandId] = newkeylabels;
    this.unlabelMenuItem(commandId); // XXX Actually should update with the last keybinding if available
    if (commandId in this._commandParams) {
        delete this._commandParams[commandId][keylabel];
    }

    // XXX should walk _up_ the tree to see if any of the parents are empty, and remove them

    var command = this.document.getElementById(commandId);
    if (command) {
        if (command.hasAttribute('key') && command.getAttribute('key') == keylabel) {
            command.removeAttribute('key');
            //dump('removing acceltext of '+ commandId + ' in clearSequence\n'); 
            command.removeAttribute('acceltext');
        }
        // XXX need to see if there's another keybinding to assign to this
        // menuitem based on its command...
    }
    // need to restore "configuration" keys if they exist.
    if (restore) {
        root = this._configKeyTree;
        for (i=0; i < keysequence.length - 1; i++) {
            if (!isAssociativeArray(root)) {
                return;
            }
            root = root[keysequence[i]];
        }
        var cmd = root[keysequence[keysequence.length-1]];
        if (typeof(cmd) != 'undefined'){
            this.assignKey(cmd, keysequence);
            this.makeKeyActive(cmd, keysequence);
        }
    }
};

keybindingManager.prototype.makeKeyActive = function(commandId, keysequence) {
  try {
    var keyname = keysequence[0]; // for purpose of assignment, we bind the command to the first key
    var oldkey = document.getElementById('keyname');
    var hadKey = 0;
    var key = this.getKey(keyname);
    if (!key) {
        this.log.error("Could not find key with name: " + keyname);
        return;
    }
    this.log.info("found key: " + keyname);
    if (keysequence.length > 1) {
        this.key2command[keysequence[0]] = 'cmd_prefix';
        key.setAttribute('oncommand', "gKeybindingMgr.prefixHandler(event)");
        key.removeAttribute('command');
    } else {
        this.keyTree[keysequence[0]] = commandId;
        this.key2command[keysequence[0]] = commandId;
        if (commandId in this._commandParams) {
            this.key2command[keysequence[0]] = commandId;
            if (key.hasAttribute('command')) {
                key.removeAttribute('command');
            }
            var keylabel = keysequence2keylabel(keysequence);
            try {
                var c = document.getElementById(commandId).getAttribute('oncommand');
            } catch (e) {
                this.log.error("Looked for commandId = " + commandId + " and found none!");
                return;
            }
            // Because of the JS eval that happens, we need to escape quotes and backslashes
            var s = this._commandParams[commandId][keylabel].replace('"', '\\"', 'g').replace("\\", "\\\\", "g");
            var onc = 'var _param = "' + s + '";' + c
            key.setAttribute('oncommand', onc);
        } else {
            key.setAttribute('command', commandId);
            if (key.hasAttribute('oncommand')) {
                key.removeAttribute('oncommand');
            }
        }
    }
    //document.persist(keyname, 'command');
    //document.persist(keyname, 'oncommand');
    var command = document.getElementById(commandId);
    if (command) {
        //dump('setting key of ' + commandId + ' to ' + keyname + '\n');
        command.setAttribute('key', keyname);
        // Build acceltext by hand, since otherwise
        // only the first key is shown, and sometimes the key doesn't
        // know how to display some characters
        var acceltext;
        acceltext = keysequence2keylabel(keysequence);
        if (acceltext[acceltext.length-1] == '+') {
            acceltext+= '+';
        }
        //dump('setting acceltext of ' + commandId + ' to ' + acceltext + '\n');
        command.setAttribute('acceltext', acceltext);
        //document.persist(commandId, 'key');
        //document.persist(commandId, 'acceltext');
    } else {
        this.log.error("couldn't find command: '" + commandId + "'");
    }
    if (!hadKey) {
        var parent = document.getElementById('widekeyset')
        parent.appendChild(key);
    }
  } catch (e) {
    this.log.error(e);
  }
}

keybindingManager.prototype.clearUsedBys= function(commandId, keysequence) {
    var usedbys = this.usedBy(keysequence)
    for (var i in usedbys) {
        this.clearSequence(usedbys[i].command, usedbys[i].label);
    }
}

keybindingManager.prototype.assignKey = function(commandId, keysequence, parameter) {
    var sequencelabel = keysequence2keylabel(keysequence)
    if (typeof(parameter) == 'string' && parameter != '') {
        this.learnParameter(commandId, parameter, sequencelabel);
    }
    if (!(commandId in this.command2key)) {
        this.command2key[commandId] = [sequencelabel];
    } else {
        this.command2key[commandId] = [sequencelabel].concat(this.command2key[commandId]);
    }
    var i;
    var root = this.keyTree;
    var key;
    for (i in keysequence) {
        key = keysequence[i];
        if (i != keysequence.length-1) {
            if (!(key in root)) {
                root[key] = {}
            } else {
                if (typeof(root[key]) != 'object') {
                    root[key] = {}
                }
            }
            root = root[key];
        } else {
            root[key] = commandId;
            return;
        }
    }
}

keybindingManager.prototype.unlabelMenuItem= function(commandId) {
    var command = document.getElementById(commandId);
    if (command) {
        command.removeAttribute('key');
        //dump("unlabeling command " + commandId + '\n')
        command.removeAttribute('acceltext');
    }
}

keybindingManager.prototype.parseGlobalData= function() {
    var i;
    var keyname, key, commandname, commanddesc, command;
    var keys = document.getElementsByTagName('key');
    var commands = document.getElementsByTagName('command');
    var broadcasters = document.getElementsByTagName('broadcaster');
    this.keynames = new Array();
    for (i=0; i< keys.length; i++) {
        this.keynames[keys[i].getAttribute('id')] = 1;
    }
    this.commandnames = new Array();
    this.commanditems = new Array();
    var already_got = new Array();
    for (i=0; i< commands.length; i++) {
        command = commands[i];
        commandname = command.getAttribute('id');
        if (already_got[commandname]) continue;
        already_got[commandname] = 1;
        this.commandnames.push(commandname);
        commanddesc = ''
        if (command.hasAttribute('desc')) {
            commanddesc = command.getAttribute('desc');
        }
        if (commanddesc == '') {
            commanddesc = commandname;
        }
        var commanditem = {};
        commanditem.desc = commanddesc;
        commanditem.name = commandname
        this.commanditems.push(commanditem);
        //dump("adding " + commandname + ' --> ' + commanddesc + '\n');
    }
    for (i=0; i< broadcasters.length; i++) {
        command = broadcasters[i];
        commandname = command.getAttribute('id');
        if (already_got[commandname]) continue;
        already_got[commandname] = 1;
        this.commandnames.push(commandname);
        commanddesc = ''
        if (command.hasAttribute('desc')) {
            commanddesc = command.getAttribute('desc');
        }
        if (commanddesc == '') {
            commanddesc = commandname;
        }
        var commanditem = {};
        commanditem.desc = commanddesc;
        commanditem.name = commandname
        this.commanditems.push(commanditem);
        //dump("adding " + commandname + ' --> ' + commanddesc + '\n');
    }
    this.commanditems.sort(sort_commanditems);
}

function sort_commanditems(a, b) {
    if (a.desc < b.desc) return -1;
    if (a.desc == b.desc) return 0;
    return 1;
}

var cssTable = "/* Copyright (c) 2000-2001 ActiveState SRL.\
   See the file LICENSE.txt for licensing information. */\
\
UL {\
   font : small verdana, arial, helvetica, sans-serif; \
   color : black; \
}\
\
TH {\
   font : small verdana, arial, helvetica, sans-serif; \
   color : black; \
}\
\
TEXTAREA {\
   font : bold small verdana, arial, helvetica, sans-serif; \
   color : black; \
   background-color : white; \
}\
\
TD.foot {\
   font : medium sans-serif; \
   color : #eeeeee; \
}\
\
TD {\
   font : small verdana, arial, helvetica, sans-serif; \
   color : black; \
   padding: 0; \
}\
\
PRE.programlisting {\
   font-size : 9pt; \
   margin-top : -14pt; \
   margin-left : 16pt; \
   list-style-type : disc; \
}\
\
P {\
   font : small verdana, arial, helvetica, sans-serif; \
   font-size: 9pt;\
   list-style-type : circle; \
   list-style-position : inside; \
}\
\
OL {\
   font : small verdana, arial, helvetica, sans-serif; \
   color : black; \
}\
\
LI {\
   font : small verdana, arial, helvetica, sans-serif; \
   color : black; \
}\
\
INPUT {\
   font : bold small verdana, arial, helvetica, sans-serif; \
   color : black; \
   background-color : white; \
}\
\
H6 {\
   font : bold small verdana, arial, helvetica, sans-serif; \
   color : black; \
}\
\
H5 {\
   font : bold small verdana, arial, helvetica, sans-serif; \
   color : blue; \
}\
\
H4 {\
   font : bold small verdana, arial, helvetica, sans-serif; \
   color : maroon; \
}\
\
H3 {\
   font : bold medium verdana, arial, helvetica, sans-serif; \
   color : blue; \
}\
\
H2 {\
   font : bold large verdana, arial, helvetica, sans-serif; \
   color : maroon; \
}\
\
H1 {\
   font : bold large verdana, arial, helvetica, sans-serif; \
   color : black; \
}\
\
DT {\
   font : small verdana, arial, helvetica, sans-serif; \
   color : black; \
}\
\
DL {\
   font : small verdana, arial, helvetica, sans-serif; \
   color : black; \
}\
\
DD {\
   font : small verdana, arial, helvetica, sans-serif; \
   color : black; \
}\
\
BODY {\
   font : x-small verdana, arial, helvetica, sans-serif; \
   color : black; \
   background-color : white; \
}\
\
A:visited {\
   color : #666666; \
}\
\
A:link {\
   color : #0000ff; \
}\
\
A:active {\
   color : #ff0000; \
}";

keybindingManager.prototype.makeCurrentKeyBindingTable = function() {
    //if (this.commanditems == null) {
        this.parseGlobalData();
    //}
    var infoSvc = Components.classes["@activestate.com/koInfoService;1"].
                  getService(Components.interfaces.koIInfoService);
    var title = 'Komodo Keybindings [' + this.currentConfiguration + ']';
    var text = "<html><HEAD><title>" + title + "</title></HEAD><STYLE TYPE=\"text/css\" MEDIA=screen><!-- " + cssTable + " --></STYLE><body size='-2'><table width=\"100%%\">";
    text = text.concat('<h2>' + title + '</h2>');
    var desc;
    var category;
    var key;
    this.lastcategory = '';
    var commandname;
    this.rownum = 0;
    var i, j;
    
    // First non-parametrized commands
    for (i = 0; i < this.commanditems.length; i++) {
        commandname = this.commanditems[i].name;
        if (commandname in this.command2key) {
            desc = this.commandId2tabledesc(commandname, '')
            category = desc.slice(0, desc.indexOf(': '));
            if (! this._commandParams[commandname]) {
                desc = desc.slice(desc.indexOf(': ')+2);
                text = text.concat(this._addRow(category, desc, this.command2key[commandname]));
            }
        }
    }
    // Then parametrized commands (these tend to be the custom ones)
    for (i = 0; i < this.commanditems.length; i++) {
        commandname = this.commanditems[i].name;
        if (commandname in this.command2key) {
            desc = this.commandId2tabledesc(commandname, '')
            category = desc.slice(0, desc.indexOf(': '));
            if (this._commandParams[commandname]) {
                for (key in this._commandParams[commandname]) {
                    desc = this.commandId2tabledesc(commandname, this._commandParams[commandname][key])
                    category = desc.slice(0, desc.indexOf(': '));
                    desc = desc.slice(desc.indexOf(': ')+2);
                    text = text.concat(this._addRow(category, desc, [key]));
                }
            }
        }
    }
    text = text.concat('</table></body></html>');
    return text;
}

keybindingManager.prototype._addRow = function(category, desc, keys) {
    var text = '';
    // We may be missing some characters.
    desc = desc.replace('&', '&amp;', 'g');
    desc = desc.replace('<', '&lt;', 'g');
    desc = desc.replace('>', '&gt;', 'g');
    if (category != this.lastcategory) {
        if (this.lastcategory != '') {
            text = text.concat('</table>');
        }
        this.lastcategory = category;
        text = text.concat('<h4>' + category + '</h4><table width=\"100%%\">')
        this.rownum = 0;
    }
    if (this.rownum % 2) {
        text = text.concat('<tr>')
    } else {
        text = text.concat('<tr bgcolor="#eeeeee">')
    }
    this.rownum += 1;
    text = text.concat('<td width="50%%"><p>' + desc + '</p></td><td width="50%%">');
    text = text.concat('<table>');
    for (var j in keys) {
        text = text.concat('<tr><td><p>'+ keys[j] +'</p></td></tr>');
    }
    text = text.concat('</table>');
    text = text.concat('</td></tr>\n');
    return text;
}

keybindingManager.prototype.commandId2desc= function(commandname, label) {
    var command = document.getElementById(commandname);
    if (! command) {
        return '';
    }
    var commanddesc = ''
    if (command.hasAttribute('desc')) {
        commanddesc = command.getAttribute('desc');
    }
    if (commanddesc == '') {
        commanddesc = commandname;
    }
    if (commandname in this._commandParams) {
        var param = this._commandParams[commandname][label];
        commanddesc = commanddesc + " (called with \'" + param + "') [" + label + ']';
    } else {
        commanddesc = commanddesc + ' [' + label + ']';        
    }
    return commanddesc;
}

keybindingManager.prototype.commandId2shortdesc= function(commandname, param) {
    var command = document.getElementById(commandname);
    if (! command) {
        return '';
    }
    var commanddesc = ''
    if (command.hasAttribute('desc')) {
        commanddesc = command.getAttribute('desc');
    }
    if (commanddesc == '') {
        commanddesc = commandname;
    }
    if (typeof(param) != 'undefined' && param) {
        if (commandname == 'cmd_executeNamedMacro') {
            return 'the "'+param+'" Macro';
        } else if (commandname == 'cmd_runNamedCommand') {
            return 'the "'+param+'" Command';
        } else if (commandname == 'cmd_insertNamedSnippet') {
            return 'the "'+param+'" Snippet';
        } else if (commandname == 'cmd_callNamedPart') {
            return 'the "'+param+'" Item';
        }
        commanddesc = commanddesc + " (called with \'" + param + "')";
    } else {
        if (commanddesc.indexOf(': ') != -1) {
            return commanddesc.slice(commanddesc.indexOf(': ')+2, commanddesc.length)
        }
    }
    return commanddesc;
}

keybindingManager.prototype.commandId2tabledesc= function(commandname, param) {
    var command = document.getElementById(commandname);
    if (! command) {
        return '';
    }
    var commanddesc = ''
    if (command.hasAttribute('desc')) {
        commanddesc = command.getAttribute('desc');
    }
    if (commanddesc == '') {
        commanddesc = commandname;
    }
    if (typeof(param) != 'undefined' && param) {
        if (commandname == 'cmd_executeNamedMacro') {
            return 'Custom Macros: ' + param;
        } else if (commandname == 'cmd_runNamedCommand') {
            return 'Custom Commands: ' + param;
        } else if (commandname == 'cmd_insertNamedSnippet') {
            return 'Snippets: '+param;
        }
    }
    return commanddesc;
}

keybindingManager.prototype.commandId2parameter= function(commandname, label) {
    if (commandname in this._commandParams) {
        return this._commandParams[commandname][label];
    } else {
        return '';
    }
}

keybindingManager.prototype.stashIn= function(vessel, keylabel) {
    vessel.setStringAttribute('keyboard_shortcut', keylabel);
    if (toolboxMgr.findItemByAttributeValue('name', vessel.getStringAttribute('name'))) {
        try {
            toolboxMgr.toolbox.save();
        } catch(ex) {
            var lastErrorSvc = Components.classes["@activestate.com/koLastErrorService;1"].
                getService(Components.interfaces.koILastErrorService);
            alert("There was an error saving the toolbox: " +
                  lastErrorSvc.getLastErrorMessage());
        }
    }
}

keybindingManager.prototype.command2keysequences= function(commandId, commandParam) {
    if (!(commandId in this.command2key)) return [];
    if (!(commandId in this._commandParams)) return this.command2key[commandId]
    var seqs = [];
    var seq;
    for (seq in this._commandParams[commandId]) {
        if (this._commandParams[commandId][seq] == commandParam) {
            seqs.push(seq);
        }
    }
    return seqs;
    
}

keybindingManager.prototype._usedbys = function(root, remainingsequence, sofarsequence, usedbys) {
    // This is a fairly delicate piece of code -- the logic is tricky to follow
    // so don't touch it if you don't really know what you're doing.
    //dump("-- remainingsequence = " + remainingsequence +'\n');
    //dump("-- sofarsequence = " + sofarsequence +'\n');
    //dump("-- usedbys = ");
    //for (i in usedbys) { dump('\t'+usedbys.desc +'\n');}
    //if (root != this.keyTree) {
        //dump("root has:\n");
        //for (var j in root) {
            //dump('\t'+j+'\t'+root[j]+'\n');
        //}
    //}
    // We need to walk the tree and collect _all_ the children of all the branch points
    // being visited.
    // Are we being specific still -- do we have more parts of the trees to prune?
    var usedby;
    var key;
    var param;
    if (remainingsequence.length > 0) {
        //dump("remaining sequence = " + remainingsequence + " and has length " + remainingsequence.length +'\n');
        for (var i in remainingsequence) {
            // for each key in the sequence yet to walk
            key = remainingsequence[i];
            //dump("phase 1, looking at " + key +'\n');
            switch (typeof(root[key])) {
                case 'string':
                    usedby = new Object();
                    usedby.command = root[key];
                    sofarsequence.push(key);
                    usedby.label = keysequence2keylabel(sofarsequence);
                    usedby.desc = this.commandId2desc(usedby.command, usedby.label);
                    if (usedby.command in this._commandParams) {
                        param = this._commandParams[usedby.command][usedby.label];
                    } else {
                        param = null;
                    }
                    usedby.shortdesc = this.commandId2shortdesc(usedby.command, param);
                    usedby.parameter = this.commandId2parameter(usedby.command, usedby.label);
                    usedbys.push(usedby);
                    break;
                case 'object':
                    sofarsequence.push(key);
                    // recurse
                    //dump("recursing into " + key +'\n');
                    this._usedbys(root[key],
                                  remainingsequence.slice(1, remainingsequence.length),
                                  sofarsequence, usedbys);
                    sofarsequence.pop();
                    root = root[key];
                    break;
                default:
                    //dump("root["+key+"] is undefined!\n");
                    // type is undefined, presumably
            }
        }
    } else {
        // We've reached the end of the specification of the sequence -- the tree
        // from here on out has to be gathered.
        for (key in root) {
            //dump("phase 2, looking at "+key+'\n');
            if (typeof(root[key]) == 'string') {
                usedby = new Object();
                usedby.command = root[key];
                //dump('YYYY - pushing ' + usedby.command +'\n');
                sofarsequence.push(key);
                usedby.label = keysequence2keylabel(sofarsequence);
                usedby.desc = this.commandId2desc(usedby.command, usedby.label);
                sofarsequence.pop();
                usedbys.push(usedby);
            } else if (typeof(root[key]) == 'object') {
                //dump("recursing in phase 2!\n");
                sofarsequence.push(key);
                // recurse
                this._usedbys(root[key], [], sofarsequence, usedbys);
                sofarsequence.pop();
            }
        }
    }
}

keybindingManager.prototype.usedBy = function(sequence) {
    // This function needs to walk the keyTree starting at root and return a list
    // of objects describing the commands which would be "intercepted" by the sequence
    // being sent in.
    if (!isArray(sequence)) {
        this.log.error("usedBy has to be called with a sequence of key labels\n");
    }
    var usedbys = new Array();
    // we start at the root of the tree, given the entire sequence, a working array,
    // and the return array
    this._usedbys(this.keyTree, sequence, [], usedbys);
    return usedbys;
}

function dumpUsedBys(usedbys) {
    var i;
    dump("USED BY RESULTS\n");
    for (i in usedbys) {
        dump('\t'+usedbys[i].label + ': ' + usedbys[i].desc + ': ' + usedbys[i].command +'\n');
    }
}

keybindingManager.prototype.event2keylabel = function (event) {
  var label = '';
  try {
    // From an event, builds the string which corresponds to the key
    // conveniently also the menu label, pretty description, etc.

    // This is a little tricky because while most keystrokes that we care about spawn keypress
    // events, some only show up in keywowns.

    //gTimeline.dumpEvent(event);

    var isCtrl = event.ctrlKey;
    var isAlt = event.altKey;
    var isShift = event.shiftKey;
    var charCode = event.charCode;
    var chr = String.fromCharCode(charCode);
    if (typeof(chr) == 'undefined') chr = '';
    var keyCode = event.keyCode;

    switch (event.type) {
        case 'keydown':
            // Here we handle "special cases" -- keys that never generate keypressed events
            // These only seem to be keys like the function keys and the keypad keys _IF_ they are modified
            if (!isCtrl && !isAlt && !isShift) return '';
            // characters shouldn't be handled by keydowns
            if (event.charCode) return '';
            // We're not interested in "just" shift, ctrl, alt, capslock
            if (event.keyCode == 0x10 || // shift
                event.keyCode == 0x11 || // ctrl
                event.keyCode == 0x12 || // alt
                event.keyCode == 0x14) { // caps lock 
                return '';
            }
            if (keyCode == 109 && !isCtrl) return ''; // This is when we do Ctrl+Shift+-, and I have no idea what it's doing here.
            if (!(keyCode in gKeydownVKs)) return '';
            label = gVKCodes[keyCode];
            break;
        case 'keypress':
            if (charCode == 32) return '';  // we handle this with VK_SPACE
            if (charCode == 45) return '';  // This is when we do Ctrl+Shift+-, and it's handled by the keydown
            if (keyCode == 31) return '';  // This is when we do Ctrl+Shift+-, and I have no idea what it's doing here.
            if (keyCode in gVKCodes) {
                label = gVKCodes[keyCode];
            } else {
                label = chr.toUpperCase();
            }
            //gTimeline.dumpEvent(event);
            break;
        default:
            this.log.warn("Got unknown event type: " + event.type + '\n');
            return '';
    }
    var shiftedDigitsMap = {'~': '`',
                      '!': '1',
                      '@': '2',
                      '#': '3', 
                      '$': '4',
                      '%': '5',
                      '^': '6',
                      '&': '7',
                      '*': '8',
                      '(': '9',
                      ')': '0',
                      '_': '-'};
    if (isShift && !(chr in shiftedDigitsMap)) {
        label = "Shift+" + label;
    }
    if (isAlt) {
        label = "Alt+" + label;
    }
    if (isCtrl) {
        label = "Ctrl+" + label;
    }
    if (label == 'Ctrl+Escape') label = 'Ctrl-[';
    //dump('type = ' + event.type+'\n');
    //dump('label = ' + label+'\n');
    //dump('keycode = ' + event.keyCode+'\n');
    //dump('charcode = ' + charCode+'\n');
    // Check that it's a key we know about in keys.xul
    if (document.getElementById(label) == null) return '';
  } catch (e) {
    log.exception(e);
  }
  return label;
}

keybindingManager.prototype.cancelPrefix = function (why) {
    // This function is called whenever the prefix handler business needs to "stop"
    // this happens internally when a sequence of keystrokes doesn't correspond
    // to a predefined binding.
    
    // If the first argument is null, then the status bar message is canceled.
    // If the first argument is _NOT_ null, then it is used as the
    // message to display on the status bar message for 3 seconds.
    
    this.currentPrefixString = '';
    this.currentPrefixMap = this.keyTree;
    if (this._originalFocus) {
        try {
            this._originalFocus.removeEventListener('keypress', gKeyHandler, true);
            this._originalFocus.removeEventListener('keydown', gKeyHandler, true);
            this._originalFocus.removeEventListener('blur', gCancelKeyHandler, false);
        } catch (e) {
            this.log.warn(e)
        }
    }
    if (why == null) {
        StatusBar_AddMessage(null, "prefix", 0, false)
    } else {
        StatusBar_AddMessage(why, "prefix", 3000, false)
    }
}

var gKeyHandler = function (event) {
    gKeybindingMgr.prefixHandler(event);
}

var gCancelKeyHandler = function (event) {
    // If called from <scintilla>, 'this' will be the XBL widget.  Weird JS.
    if (this && this.key_handler) {
        this.key_handler = null;
    }
    if (this && this.mouse_handler) {
        this.mouse_handler = null;
    }
    gKeybindingMgr.cancelPrefix('Clicked away')
}

keybindingManager.prototype.prefixHandler = function (event) {
    event.cancelBubble = true;
    event.preventDefault();

    var key = this.event2keylabel(event);
    if (key == '') return;  // but _stay_ in prefixHandler mode
    if (event.keyCode == event.DOM_VK_ESCAPE) {
        this.cancelPrefix(null);
        return;
    }

    if (this.currentPrefixString == '') {
        this.currentPrefixString = key;
    } else {
        this.currentPrefixString += ', ' + key;
    }

    var scintilla = gViewMgr.currentView.scintilla;
    if (!(key in this.currentPrefixMap)) {
        this.cancelPrefix('('+this.currentPrefixString + ") is not a valid key sequence.");
        this.currentPrefixMap = this.keyTree;
        if (scintilla) {
            if (scintilla.key_handler) {
                scintilla.key_handler = null;
            }
            if (scintilla.mouse_handler) {
                scintilla.mouse_handler = null;
            }
        }
        return;
    }

    if (typeof(this.currentPrefixMap[key]) == 'string') {
        // it's a command
        var keylabel = this.currentPrefixString;
        var commandname = this.currentPrefixMap[key];
        var command;
        this.cancelPrefix(null);
        if (scintilla) {
            if (scintilla.key_handler) {
                scintilla.key_handler = null;
            }
            if (scintilla && scintilla.mouse_handler) {
                scintilla.mouse_handler = null;
            }
        }
        command = document.getElementById(commandname);
        if (!command) {
            this.log.error("FAILURE to find command " + commandname);
        }
        var expr;
        try {
            if (commandname in this._commandParams) {
                // It's a "parametrized command
                if (command.hasAttribute('oncommand')) {
                    expr = "var _param = '"+this._commandParams[commandname][keylabel].replace('"', '\\"', 'g') +"'; " + command.getAttribute('oncommand');
                    //dump("would eval" + expr +'\n');
                    eval(expr);
                }
            } else {
                if (!command_doCommand(commandname)) {
                    // if it's not a "real" command but just has an oncommand handler, then use that
                    if (command.hasAttribute('oncommand')) {
                        expr = command.getAttribute('oncommand');
                        //dump("would eval: " + expr+'\n');
                        eval(expr);
                    }
                }
            }
        } catch (e) {
            // This is important We really don't want a command evaluation to screw up the prefix map.
            this.log.error(e);
        }
        //dump("resetting prefix map\n");
        this.currentPrefixMap = this.keyTree;
    } else {
        // The way we handle prefix commands has to differ depending on whether we're in a scintilla
        // window or not.  This code will have to be generalized if we add keybindings to other
        // scintilla widgets
        if (view_editorHasFocus()) {
            // <scintilla> is pretty smart in that if a key_handler is set, then keystroke events
            // make their way to that handler, not to the real scintilla.
            this._originalFocus = gViewMgr.currentView.scintilla;
            this._originalFocus.key_handler = gKeyHandler;
        } else {
            // unfortunately, I don't know how to _hijack_ the key events on regular XUL
            // events (such as the trees).  So Ctrl-K, Down arrow will change the focus
            // in the tree. =( XXX
            this._originalFocus = document.commandDispatcher.focusedElement;
            this._originalFocus.addEventListener('keypress',gKeyHandler, true);
            this._originalFocus.addEventListener('keydown', gKeyHandler, true);
        }
        this._originalFocus.addEventListener('blur', gCancelKeyHandler, false);
        this.currentPrefixMap = this.currentPrefixMap[key]
        StatusBar_AddMessage(this.currentPrefixString + " was pressed.  Awaiting further keys.", "prefix", 0, false)
    }
}

function keybindingManagerOnLoad()
{
    try {
        gTimeline.enter('keybindingManagerOnLoad');
        gKeybindingMgr = new keybindingManager();
        gKeybindingMgr.loadConfiguration(gKeybindingMgr.currentConfiguration);
    } catch (e) {
        log.error(e);
    }
    gTimeline.leave('keybindingManagerOnLoad');
}

var gKeydownVKs = {
    0x1b:1, // ESC
    0x1d:1, // ]
    0x6a:1, //*
    0x6b:1, //+
    0x6d:1, //-
    0x6f:1, //divide
    0x09:1, // "Tab",
    0x20:1  // "Space",
}

var gVKCodes = {0x03: "cancel",
0x08: "Backspace",
0x09: "Tab",
0x0c: "Clear",
0x0d: "Return",
0x0e: "Enter",
0x10: "Shift",
0x11: "Control",
0x12: "Alt",
0x13: "Pause",
0x14: "Caps_Lock",
0x1b: "Escape",
0x1d: "]",
0x20: "Space",
0x21: "Page_Up",
0x22: "Page_Down",
0x23: "End",
0x24: "Home",
0x25: "Left",
0x26: "Up",
0x27: "Right",
0x28: "Down",
0x2c: "PrintScreen",
0x2d: "Insert",
0x2e: "Delete",
0x3b: ";",
0x3d: "=",
0x60: "NumPad-0",
0x61: "NumPad-1",
0x62: "NumPad-2",
0x63: "NumPad-3",
0x64: "NumPad-4",
0x65: "NumPad-5",
0x66: "NumPad-6",
0x67: "NumPad-7",
0x68: "NumPad-8",
0x69: "NumPad-9",
0x6a: "*",
0x6b: "+",
0x6c: "Separator",
0x6d: "-",
0x6e: "Decimal",
0x6f: "/",
0x70: "F1",
0x71: "F2",
0x72: "F3",
0x73: "F4",
0x74: "F5",
0x75: "F6",
0x76: "F7",
0x77: "F8",
0x78: "F9",
0x79: "F10",
0x7A: "F11",
0x7B: "F12",
0x7C: "F13",
0x7D: "F14",
0x7E: "F15",
0x7F: "F16",
0x80: "F17",
0x81: "F18",
0x82: "F19",
0x83: "F20",
0x84: "F21",
0x85: "F22",
0x86: "F23",
0x87: "F24",
0x90: "Num_Lock",
0x91: "Scroll_Lock",
0xbc: ",",
0xbe: ".",
0xbf: "/",
0xc0: "`",
0xdb: "[",
0xdc: "\\",
0xdd: "]",
0xde: "\"",
0xe0: "Meta"
}

