/* Copyright (c) 2000-2001 ActiveState Corporation.
   See the file LICENSE.txt for licensing information. */

// A project extension to manage source code control operations in Komodo.

// start the extension
var gpeSCC = new peSCC();
registeredExtensions[registeredExtensions.length] = gpeSCC;

//---- Registry of active SCC services
// 'sccModules' is a dictionary of active SCC services conforming to the
// virtual (so virtual we have defined a common base class) koISCC interface.
// Generally there is one such service for each SCC system (e.g. CVS) that
// Komodo supports.
// 
// "Active" means that the SCC system is (1) functional (cannot support CVS
// if CVS is not installed on the user's machine) and (2) enabled (user
// preference). Monitoring (2) can be done via Komodo pref observer system.
// Monitoring (1) is done by observing the 'subsystem_changed'
// notification.
//
var sccModules = {};
var _sccObserver = new _SCCObserver();
var _sccPrefObserver = new _SCCPrefObserver();

function _activateCVSService() {
    var _cvsSvc = Components.classes["@activestate.com/koCVS;1"].
                  getService(Components.interfaces.koICVS);
    // Activate of deactivate the CVS service depending on current state.
    if (_cvsSvc.isFunctional && gPrefs.getBooleanPref("cvsEnabled")) {
        //dump("peSCC: CVS service enabled\n");
        sccModules['cvs'] = _cvsSvc;
    } else {
        //dump("peSCC: CVS service disabled\n");
        sccModules['cvs'] = null;
    }
}

function _activateP4Service() {
    // Activate of deactivate the P4 service depending on current state.
    var _p4Svc = Components.classes["@activestate.com/koP4;1"].
                 getService(Components.interfaces.koIP4);
    if (_p4Svc.isFunctional && gPrefs.getBooleanPref("p4Enabled")) {
        //dump("peSCC: P4 service enabled\n");
        sccModules['p4'] = _p4Svc;
    } else {
        //dump("peSCC: P4 service disabled\n");
        sccModules['p4'] = null;
    }
}

function _SCCObserver() {}
_SCCObserver.prototype.observe = function(subject, topic, data)
{
    // The "subsystem_changed" notification indicates that a Komodo
    // subsystem has changed. "data" indentifies the subsystem. If data is
    // 'p4' or 'cvs' then we might need to update the list of active SCC
    // services.
    switch (topic) {
    case 'scc_status_message':
        SCCScinAppendText(data);
        break;
    case 'subsystem_changed':
        switch (data) {
        case 'cvs':
            _activateCVSService();
            break;
        case 'p4':
            _activateP4Service();
            break;
        }
        break;
    }    
}
_SCCObserver.prototype.init = function() {
    var _obSvc = Components.classes["@mozilla.org/observer-service;1"].
                 getService(Components.interfaces.nsIObserverService);
    _obSvc.addObserver(this, 'subsystem_changed',false);
    _obSvc.addObserver(this, 'scc_status_message',false);
}
_SCCObserver.prototype.destructor = function() {
    var _obSvc = Components.classes["@mozilla.org/observer-service;1"].
                 getService(Components.interfaces.nsIObserverService);
    _obSvc.removeObserver(this, 'subsystem_changed');
    _obSvc.removeObserver(this, 'scc_status_message');
}

function _SCCPrefObserver() {};
_SCCPrefObserver.prototype.observe = function(prefSet, prefSetID, prefName)
{
    //dump("peSCC: observed pref notification: prefName='"+prefName+"'\n");
    switch (prefName) {
    case "cvsEnabled":
        _activateCVSService();
        break;
    case "p4Enabled":
        _activateP4Service();
        break;
    }
};
_SCCPrefObserver.prototype.init = function() {
    gPrefs.addObserver(this);
}
_SCCPrefObserver.prototype.destructor = function() {
    if (gPrefSvc && gPrefs) {
        gPrefs.removeObserver(this);
    }
}


//---- SCC project extension implementation.

function peSCC() {
    this.name = 'SCC';
    this.log = getLoggingMgr().getLogger('peSCC');
    this.sysutils = Components.classes["@activestate.com/koSysUtils;1"].
              getService(Components.interfaces.koISysUtils);
    this.observerSvc = Components.classes["@mozilla.org/observer-service;1"].
                       getService(Components.interfaces.nsIObserverService);
    this._lastErrorSvc = Components.classes["@activestate.com/koLastErrorService;1"]
               .getService(Components.interfaces.koILastErrorService);
}

// The following two lines ensure proper inheritance (see Flanagan, p. 144).
peSCC.prototype.constructor = peSCC;

peSCC.prototype.init = function() {
    // register our command handlers
    this.registerCommands();
    
    // Setup dynamic registration of active SCC services.
    _activateCVSService();
    _activateP4Service();
    _sccPrefObserver.init();
    _sccObserver.init();
}

peSCC.prototype.registerCommands = function() {
    extensionManager.registerCommand('cmd_SCCedit', this);
    extensionManager.registerCommand('cmd_SCCadd', this);
    extensionManager.registerCommand('cmd_SCCremove', this);
    extensionManager.registerCommand('cmd_SCCupdate', this);
    extensionManager.registerCommand('cmd_SCCcommit', this);
    extensionManager.registerCommand('cmd_SCCdiff', this);
    extensionManager.registerCommand('cmd_SCCrevert', this);

    extensionManager.registerCommand('cmd_SCCadd_folder', this);
    extensionManager.registerCommand('cmd_SCCupdate_folder', this);
    extensionManager.registerCommand('cmd_SCCcommit_folder', this);
    extensionManager.registerCommand('cmd_SCCdiff_folder', this);
    extensionManager.registerCommand('cmd_SCCrevert_folder', this);
}

peSCC.prototype.registerEventHandlers = function() {
}

peSCC.prototype.registerMenus = function() {
    var menupopup = document.getElementById('menu_SCCmenu_context');
    if (!menupopup) {
        log.error("Couldn't get 'menu_SCCmenu'");
        return;
    }
    var folderpopup = document.getElementById('menu_SCCmenu_folder_context');
    if (!folderpopup) {
        log.error("Couldn't get 'menu_SCCmenu_folder'");
        return;
    }
    extensionManager.addMenuItem(Components.interfaces.koIPart_SCC_file, document.createElement('menuseparator'));
    extensionManager.addMenuItem(Components.interfaces.koIPart_SCC_file, menupopup);
    extensionManager.addMenuItem(Components.interfaces.koIPart_SCC_collection, document.createElement('menuseparator'));
    extensionManager.addMenuItem(Components.interfaces.koIPart_SCC_collection, folderpopup);
    extensionManager.addMenuItem(Components.interfaces.koIPart_SCC_directory, document.createElement('menuseparator'));
    extensionManager.addMenuItem(Components.interfaces.koIPart_SCC_directory, folderpopup);
}

/* This takes an item, and determines what type of scc
   flags we are dealing with. returns true if we can do anything, or
   false if we cannot do anything */
peSCC.prototype.determineItemSCCSupport = function(item, sccObj) {
    // prevent non-cvs directories from having to go through this torture
    if (item.properties['sccExclude']) {
        return false;
    }
    // assume the worst =)
    sccObj.isscc_file = false;
    sccObj.isscc_directory = false;
    sccObj.isscc_container = false;
    sccObj.isscc_collection = false;
    
    if (!item.part) {
        sccObj.isscc_file = true;
    } else {
        try {
            ok = item.part.QueryInterface(Components.interfaces.koIPart_SCC_file) != null;
            if (ok) {
                sccObj.isscc_file = true;
            }
        } catch(e) {}
        try {
            ok = item.part.QueryInterface(Components.interfaces.koIPart_SCC_collection) != null;
            if (ok) {
                sccObj.isscc_collection = true;
            }
        } catch(e) {}
        try {
            ok = item.part.QueryInterface(Components.interfaces.koIPart_SCC_directory) != null;
            if (ok) {
                sccObj.isscc_directory = true;
            }
        } catch(e) {}
        sccObj.isscc_container = sccObj.isscc_directory || sccObj.isscc_collection;
    }
    sccObj.has_scc = typeof(item.properties['sccType']) != 'undefined' && item.properties['sccType'] != '';
    sccObj.is_scc = ('isscc_file' in sccObj && sccObj.isscc_file) || sccObj.isscc_container;
    if (!sccObj.is_scc) {
        return false;
    }

    sccObj.can_scc = sccObj.has_scc || item.properties['sccDirType'];
    sccObj.missing = item.properties['missing'];
    
    var in_repository = null;
    if (item.url && !sccObj.can_scc && !sccObj.missing) {
        item.properties['sccDirType'] = '';
        in_repository = gFileStatusSvc.isRepository(item.url);
        if (in_repository) {
            sccObj.can_scc = item.properties['sccDirType'] = in_repository;
        }
        item.properties['sccExclude'] = !sccObj.has_scc && !sccObj.can_scc;
    }

    if (!sccObj.has_scc && !sccObj.can_scc) return false;
    return true;
}

peSCC.prototype.supportsCommand = function(command, item) {
    // do we have either cvs or scc?
    if (!gPrefs.getBooleanPref('cvsEnabled') && !gPrefs.getBooleanPref('p4Enabled')) {
        return false;
    }
    // If scintilla has focus, should be talking about the file in scintilla
    if (!item) {
        if (view_editorHasFocus()) {
            var view = gViewMgr.currentView;
            if (!view || !view.document || view.document.isUntitled) {
                // This is an unsaved file -- no SCC commands can apply.
                return 0;
            }
            item = view.item;
            item.url = view.document.file.URI;
        } else {
            if (!view_projectHasFocus() && !view_toolboxHasFocus()) {
                return 0;
            } else {
                // otherwise, look at the focused project view's selected parts.
                item = gFocusedProjectView.getSelectedItem();
                // item must be a file
                if (!item) {
                    return false;
                }
            }
        }
    }
    // XXX look into cacheing sccObj into the item
    var sccObj = new Object();
    if (!this.determineItemSCCSupport(item,sccObj)) {
        return false;
    }
    var isfoldercmd = command.indexOf('_folder')>1;
    if (isfoldercmd && !sccObj.isscc_container) {
        return false;
    }
    /* folder handling is different between p4 and cvs */
    var folders_in_repository = item.properties['sccType'] == 'cvs' || item.properties['sccDirType'] == 'cvs';
    var folders_revertable = item.properties['sccType'] == 'p4' || item.properties['sccDirType'] == 'p4';
    
    /* if cvs, and this is a directory, then the directory MUST be in cvs */
    var foldercmd_ok = (sccObj.isscc_container && isfoldercmd) && 
            ((folders_in_repository && sccObj.isscc_directory && sccObj.has_scc) || 
            (!folders_in_repository && sccObj.can_scc));
    var retval;
    switch (command) {
    case 'cmd_SCCedit':
        retval = sccObj.has_scc && item.properties['sccType'] != 'cvs' && item.properties['sccAction'] == '';
        //dump('asked about '+ command + ', returning' + retval +'\n');
        return retval;
        break;
    case 'cmd_SCCadd_folder':
        retval = sccObj.can_scc && !sccObj.has_scc && isfoldercmd 
                && sccObj.isscc_directory && folders_in_repository;
        //dump('asked about '+ command + ', returning' + retval +'\n');
        return retval;
        break;
    case 'cmd_SCCadd':
        retval = sccObj.can_scc && !sccObj.has_scc;
        //dump('asked about '+ command + ', returning' + retval +'\n');
        return retval;
        break;
    case 'cmd_SCCupdate_folder':
        return foldercmd_ok;
        break;
    case 'cmd_SCCupdate':
        retval = sccObj.has_scc && item.properties['sccNeedSync'];
        //dump('asked about '+ command + ', returning' + retval +'\n');
        return retval;
        break;
    case 'cmd_SCCremove':
        return sccObj.has_scc && !item.properties['sccNeedSync'] &&
            item.properties['sccAction'] == '';
        break;
    case 'cmd_SCCdiff_folder':
        return foldercmd_ok;
        break;
    case 'cmd_SCCdiff':
        retval = sccObj.has_scc &&(item.properties['sccNeedSync'] ||
                item.properties['sccAction'] == 'edit');
        //dump('asked about '+ command + ', returning' + retval +'\n');
        return retval;
        break;
    case 'cmd_SCCrevert_folder':
        return foldercmd_ok && folders_revertable;
        break;
    case 'cmd_SCCrevert':
        retval = sccObj.has_scc && (item.properties['sccNeedSync'] ||
                item.properties['sccAction'] != '');
        //dump('asked about '+ command + ', returning' + retval +'\n');
        return retval;
        break;
    case 'cmd_SCCcommit_folder':
        return foldercmd_ok;
        break;
    case 'cmd_SCCcommit':
        return sccObj.has_scc && item.properties['sccAction'] != ''
            && (!item.properties['sccNeedSync'] || 
                item.properties['sccAction'] == 'delete');
        break;
    default:
        break;
    }
    return false;
}

peSCC.prototype.isCommandEnabled = peSCC.prototype.supportsCommand;

/* recursivly build a list of items that are being used in
   this scc operation */
peSCC.prototype._addVirtualFolderChildren = function(item)
{
    var items = [];
    if (!item.hasWrappedChildren() ||
        !item.children) return items;
        
    for (var i = 0; i < item.children.length; i++) {
        if (item.children[i].isContainer &&
            item.children[i].isContainerOpen && 
            !item.children[i].url && 
            item.children[i].hasWrappedChildren()) {
            
            var c = this._addVirtualFolderChildren(item.children[i]);
            if (c.length > 0) {
                items = items.concat(c);
            }
        } else if (item.children[i].url) {
            items.push(item.children[i]);
        }
    }
    return items;
}

peSCC.prototype.doCommand = function(command) {
    window.setTimeout('peSCCdoCommand("'+command+'");',1);
}

function peSCCdoCommand(command) {
    gpeSCC._doCommand(command);
}

peSCC.prototype._doCommand = function(command) {
    var cmd = null;
    var error = null;
    var urls=[];
    var _sccUrls = [];
    _sccUrls['p4'] = []
    _sccUrls['cvs'] = []
    var item = null;
    var items = [];
    var i;
    if (view_editorHasFocus()) {
        //dump("Scintilla has focus -- actions should be based on the current file\n");
        var view = gViewMgr.currentView;
        if (view && !view.document.isUntitled) {
            item = view.item;
            item.url = view.document.file.URI;
            items.push(item);
        }
    } else if (view_projectHasFocus() || view_toolboxHasFocus()) {
        items = gFocusedProjectView.getSelectedItems();
    }
    
    if (items.length < 1) return;

    var isfoldercmd = command.indexOf('_folder')>1;
    for (i=0; i<items.length; i++) {
        // XXX look into cacheing sccObj into the item
        var sccObj = new Object();
        if (!this.determineItemSCCSupport(items[i],sccObj)) return;

        var type = items[i].properties['sccType'];
        if (!type) type = items[i].properties['sccDirType'];
        if (!type) continue;
        
        if (isfoldercmd && sccObj.isscc_directory && items[i].type == 'project') {
            _sccUrls[type].push(uriparse_dirName(items[i].url));
        } else if (isfoldercmd && sccObj.isscc_collection) {
            if (typeof(items[i].url) != 'undefined' && items[i].url)
                _sccUrls[type].push(items[i].url);
            // add the dialogs children
            items = items.concat(this._addVirtualFolderChildren(items[i]));
        } else {
            _sccUrls[type].push(items[i].url);
        }
        
        // XXX implement virtual folder support
        // check for virtual folders, add the children of the virtual
        // folders to our items array.
        //if (sccObj.isscc_collection && items.children.length > 0) {
        //    var c = this._addVirtualFolderChildren(items[i]);
        //    if (c.length > 0) {
        //        items = items.concat(c);
        //    }
        //}
    }
    var output = '';
    var title;
    //dump('selected for scc operation '+urls +'\n');
    for (var scchandler in sccModules) {
        if (!sccModules[scchandler]) continue; // skip unregistered SCC services
        if (_sccUrls[scchandler].length < 1) continue;
        urls = _sccUrls[scchandler];
        //try {
            switch (command) {
            case 'cmd_SCCedit':
                window.setCursor("wait");
                output = sccModules[scchandler].edit(urls.length, urls);
                output = this.generateCommandInfo(scchandler+' edit',urls,output);
                title = this.summarizeSCCAction(scchandler, 'edit', urls);
                SCCScinReportResults(title, output);
                break;
            case 'cmd_SCCadd_folder':
                if (!isfoldercmd) return;
            case 'cmd_SCCadd':
                this.doAdd(urls, sccModules, scchandler);
                break;
            case 'cmd_SCCremove':
                this.doRemove(urls, sccModules, scchandler);
                break;
            case 'cmd_SCCrevert_folder':
                if (!isfoldercmd) return;
            case 'cmd_SCCrevert':
                this.doRevert(urls, sccModules, scchandler);
                break;
            case 'cmd_SCCupdate_folder':
                if (!isfoldercmd) return;
            case 'cmd_SCCupdate':
                this.doUpdate(urls, sccModules, scchandler);
                break;
            case 'cmd_SCCdiff_folder':
                if (!isfoldercmd) return;
            case 'cmd_SCCdiff':
                window.setCursor("wait");
                var options = '';
                var external = '';
                var diff;
                if (scchandler == 'p4') {
                    if (gPrefs.getStringPref('p4_diff_type') == 'komododiff') {
                        options = gPrefs.getStringPref('p4DiffOptions');
                    } else {
                        options = '';
                        external = gPrefs.getStringPref('externaldiff');
                    }
                    diff = sccModules[scchandler].diff(urls.length, urls, options, external);
                } else if (scchandler == 'cvs') {
                    options = gPrefs.getStringPref('cvsDiffOptions');
                    diff = sccModules[scchandler].diff(urls.length, urls, options, '');
                }
                error = this._lastErrorSvc.getLastErrorMessage();
                //output = this.generateCommandInfo(scchandler+' diff',urls,diff);
                //SCCScinAppendText(output);
                if (!error && external == '') {
                    if (diff) {
                        var diffDisplayStyle = gPrefs.getStringPref('diffDisplayStyle');
                        if (diffDisplayStyle == 'tab') {
                            try {
                                diffView = gViewMgr.doNewView('Diff');
                                diffView.document.buffer = diff;
                            } catch (e) {
                                log.exception(e);
                            }
                        } else if (diffDisplayStyle == 'diffwin') {
                            // Put together a title for the diff output win.
                            title = "Diff:";
                            for (i=0; i < urls.length; i++) {
                                if (i > 0) title += ",";
                                title += " "+uriparse_baseName(urls[i]);
                            }
                            launch_DiffWindow(diff, title);
                        }
                    } else {
                        alert("Files are in sync with repository.");
                    }
                }
                break;
            case 'cmd_SCCcommit_folder':
                if (!isfoldercmd) return;
            case 'cmd_SCCcommit':
                this.doCommit(urls, sccModules, scchandler);
                break;
            default:
                break;
            }
        //} catch (e) {
        //    alert("An exception occured during SCC operation. (" + e + ')');
        //    this.log.error(e);
        //}
    }
    window.updateCommands('SCC'); // We've changed things hopefully, and no one else knows.
    window.setCursor("auto");
    error = this._lastErrorSvc.getLastErrorMessage();
    if (error)
        alert(error);
}

peSCC.prototype.generateCommandInfo = function(command,urls,text) {
    var out = command+'\n';
    for (var url in urls) {
        out += '    '+urls[url]+'\n'
    }
    out += text;
    return out;
}

peSCC.prototype.doAdd = function(urls, sccModules, scchandler) {
    var selectedURLs = dialog_selectFromList(
            "Add Files to Source Control", // title
            "Select the files you would like to add to the Source Code Control repository.",
            urls,
            null, // selectionCondition (null means default)
            uriparse_displayPath); // stringifier
    if (selectedURLs == null) { // user cancelled dialog
        return false;
    }
    window.setCursor("wait");
    output = sccModules[scchandler].add(selectedURLs.length, selectedURLs,
                                        '', '');
    window.setCursor("auto");
    output = this.generateCommandInfo(scchandler+' add', selectedURLs, output);
    var title = this.summarizeSCCAction(scchandler, 'add', selectedURLs);
    SCCScinReportResults(title, output);
    return true;
}

peSCC.prototype.doRemove = function(urls, sccModules, scchandler) {
    var prompt = null;
    if (scchandler == 'p4') {
        prompt = "Select the file you would like to remove from the " +
                 "repository. (The selected files will be removed from " +
                 "your disk.)";
    } else { 
        prompt = "Select the file you would like to remove from the " +
                 "repository.";
    }
    var selectedURLs = dialog_selectFromList(
            "Remove Files from Source Control", // title
            prompt,
            urls,
            null, // selectionCondition (null means default)
            uriparse_displayPath); // stringifier
    if (selectedURLs == null) { // user cancelled dialog
        return false;
    }

    window.setCursor("wait");
    var force = true;
    var recursive = false;
    var output = sccModules[scchandler].remove(selectedURLs.length,
                                               selectedURLs, force,
                                               recursive);
    window.setCursor("auto");
    output = this.generateCommandInfo(scchandler+' remove', selectedURLs, output);
    var title = this.summarizeSCCAction(scchandler, 'remove', selectedURLs);
    SCCScinReportResults(title, output);
    return true;
}

peSCC.prototype.doRevert = function(urls, sccModules, scchandler) {
  try {
    var selectedURLs = dialog_selectFromList(
            "Revert Files to Repository", // title
            "Select the files to revert to the repository version. "+
                "(Changes in opened documents will be lost).",
            urls,
            null, // selectionCondition (null means default)
            uriparse_displayPath); // stringifier
    if (selectedURLs == null) { // user cancelled dialog
        return false;
    }

    window.setCursor("wait");
    try {
        var output = sccModules[scchandler].revert(selectedURLs.length, selectedURLs,'');
        output = this.generateCommandInfo(scchandler+' revert',selectedURLs,output);
        var title = this.summarizeSCCAction(scchandler, 'revert', selectedURLs);
        SCCScinReportResults(title, output);
    } catch (e) {
        this.log.error(e);
    }
    gViewMgr.revertViewsByURL(selectedURLs);
    // handle reverting projects properly
    var isProjectFile;
    for (var i in selectedURLs) {
        isProjectFile = selectedURLs[i].slice(-4).toLowerCase() == '.kpf';
        if (isProjectFile) {
            gProjectManager.revertProjectByURL(selectedURLs[i]);
        }
    }
    window.setCursor("auto");
  } catch (e) {
    this.log.error(e);
  }
    return true;
}

peSCC.prototype.summarizeSCCAction = function(scchandler, action, urls)
{
    var retval = "Last command: " + action + ' ('+scchandler + ') on ';
    var display = uriparse_displayPath(urls[0]);
    if (urls.length > 1) {
        retval = retval + ' ' + display + ', ... (' + urls.length + ' files/directories).';
    } else {
        retval = retval + ' ' + display + '.';
    }
    return retval;
}

peSCC.prototype.doCommit= function(urls, sccModules, scchandler) {
    gViewMgr.offerToSave(urls,
                         "Save Modified Files",
                         "Save selected files before performing SCC operation?");

    var answer = sccdialog_commit(urls);
    if (answer == null) { // user cancelled
        return false;
    }

    window.setCursor("wait");
    var output = sccModules[scchandler].commit(answer.selected.length,
                                               answer.selected,
                                               answer.message, '');
    gViewMgr.revertViewsByURL(answer.selected);
    window.setCursor("auto");
    output = this.generateCommandInfo(scchandler+' commit', answer.selected,
                                      output);
    var title = this.summarizeSCCAction(scchandler, 'commit', answer.selected);
    SCCScinReportResults(title, output);
    return true;
}

peSCC.prototype.doUpdate = function(urls, sccModules, scchandler) {
    OfferToSaveFilesProjectsDialogs(urls, "Save Modified Files",
        "Save selected files before performing SCC operation?");

    var selectedURLs = dialog_selectFromList(
            "Update Files from Source Control", // title
            "Select the files to update from the repository.",
            urls,
            null, // selectionCondition (null means default)
            uriparse_displayPath); // stringifier
    if (selectedURLs == null) { // user cancelled dialog
        return false;
    }

    window.setCursor("wait");
    var changes = sccModules[scchandler].update(selectedURLs.length, selectedURLs,'');
    gViewMgr.revertViewsByURL(selectedURLs);
    window.setCursor("auto");
    if (changes) {
        this.handleChanges(changes);
    }
    output = this.generateCommandInfo(scchandler+' update',selectedURLs,output);
    var title = this.summarizeSCCAction(scchandler, 'update', selectedURLs);
    SCCScinReportResults(title, changes);
    return true;
}

peSCC.prototype.handleChanges = function(changes) {
    //XXX This pattern is specific to CVS update output. What about Perforce
    //    sync output? Is this a bug or is .handleChanges only called for CVS
    //    updates? --TrentM

    // re ^(?P<status>[UPARMC\?]) (?P<file>.*)$
    RegExp.multiline = true;
    var pattern = new RegExp("^([UCDP\?]) (.*?) (.*?)$","g");
    var result;
    var conflicts = [];
    var haveConflict = false;
    var removed = [];
    var haveRemoved = false;
    var added = [];
    var haveNew = false;
    var file;
    var i,j;
    var alreadyHaveIt;
    while ((result = pattern.exec(changes)) != null) {
        //dump('found: '+result[1]+' '+result[2]+' '+result[3]+'\n');
        switch (result[1]) {
        case 'C':
            // conflict
            haveConflict = true;
            conflicts.push(result[3]);
            break;
        case 'D':
            // removed
            haveRemoved = true;
            removed.push(result[3]);
            break;
        case 'P':
            break; // Patched file can't be a new file, even though it may not be in our project.
        case 'U':
            // either updated or new file -- 
            // XXX need to check to make sure that it really is new -- we add all files that aren't in the project
            alreadyHaveIt = false;
            for (j=0; j < gPartManagers.length; j++) {
                list = gPartManagers[j].getItemsByURL(result[3]);
                if (list.length > 0) {
                    alreadyHaveIt = true;
                    break;
                }
            }
            if (! alreadyHaveIt) {
                haveNew = true;
                added.push(result[3]);
            }
            break;
        }
    }
    // look to see if files were added, ask if they should be
    // added to the current project.
    var filepart, selectedURLs, obj;
    if (haveNew && gProjectManager.getCurrentProject()) {
        //XXX I think either (1) we should 'allowNone' here so the user could
        //    chose to not add new files, yet continue on to remove files, or
        //    better yet (2) we should group the added and removed files and
        //    ask the user to select which updates to carry to their project in
        //    one dialog.
        selectedURLs = dialog_selectFromList(
                "Add New Files", // title
                "Some new files from the repository are not in your "+
                    "project.  Select the files to add to your project.",
                added,
                null, // selectionCondition (null means default)
                uriparse_displayPath); // stringifier
        if (selectedURLs == null) { // user cancelled dialog
            return false;
        }

        for (i = 0; i < selectedURLs.length; i++) {
            filepart = Components.classes["@activestate.com/koPart?type=file;1"]
                                    .createInstance(Components.interfaces.koIPart_file);
            filepart.setStringAttribute('url', selectedURLs[i]);
            gProjectManager.addItem(filepart);
        }
    }
    // look to see if files were removed, notify of removal,
    // ask to remove from projects.
    if (haveRemoved) {
        //XXX This code is broken: what guarantee is there that selectedURLs
        //    has been set? !!!  [TrentM]
        var aroundToRemove = [];
        for (i = 0; i < selectedURLs.length; i++) {
            for (j=0; j < gPartManagers.length; i++) {
                list = gPartManagers[j].getItemsByURL(selectedURLs[i]);
                if (list.length > 0) {
                    aroundToRemove.push(selectedURLs[i]);
                }
            }
        }

        selectedURLs = dialog_selectFromList(
                "Remove Old Files", // title
                "Some files have been removed from the repository. "+
                    "Select the files to remove from your projects and "+
                    "toolbox.",
                aroundToRemove,
                null, // selectionCondition (null means default)
                uriparse_displayPath); // stringifier
        if (selectedURLs == null) { // user cancelled dialog
            return false;
        }

        var items, list;
        for (i = 0; i < selectedURLs.length; i++) {
            for (j=0; j < gPartManagers.length; j++) {
                list = gPartManagers[j].getItemsByURL(selectedURLs[i]);
                for (j = 0; j < list.length; j++) {
                    gPartManagers[i].removeItem(list[j]);
                }
            }
        }
    }
    
    // look to see if there are conflicts, notify of conflicts,
    // and ask to open the files in komodo.  also mark any
    // items with conflict status
    if (haveConflict) {
        selectedURLs = dialog_selectFromList(
                "Merge Conflicts", // title
                "There are conflicts in some updated file(s). "+
                    "Would you like to open those file(s)?",
                conflicts,
                null, // selectionCondition (null means default)
                uriparse_displayPath); // stringifier
        if (selectedURLs == null) { // user cancelled dialog
            return false;
        }
        open_openMultipleURIs(selectedURLs);
    }
    
    // XXX reset the static var, is this necessary?
    RegExp.multiline = false;
    return true;
}

function editSCCFile(item)
{
    var urls = [];
    urls.push(item.url);
    window.setCursor("wait");
    try {
        sccModules[item.properties['sccType']].edit(urls.length, urls);
    } catch (e) {
        item.dump()
        alert("An exception occured during an SCC operation:" + e);
    }
    window.setCursor("auto");
    var lastErrorSvc = Components.classes["@activestate.com/koLastErrorService;1"]
                        .getService(Components.interfaces.koILastErrorService);
    error = lastErrorSvc.getLastErrorMessage();
    if (error) {
        alert(error);
        return false;
    }
    return true;
}


/* SCC Scintilla control */

function SCCScinInit() {
    var scin = document.getElementById('scin-sccoutput');
    scin.initWithBuffer('', 'Text')
}

function SCCScinReportResults(actiondescription, actiontext)
{
    document.getElementById('sccoutput-desc').setAttribute('value', actiondescription);
    SCCTab_Show();
    SCCScinAppendText(actiontext);
}

function SCCTab_Show()
{
    var tabWidget = document.getElementById("sccoutput_tab");
    var tabsWidget = document.getElementById("output_tabs");
    if (! uilayout_outputPaneShown() || tabsWidget.selectedItem != tabWidget) {
        if (dialog_yesNo("Show SCC Output Tab?", // prompt
                         "Yes", // response
                         null, // text
                         null, // title
                         'show_scc_tab') == "Yes") {
            uilayout_ensureOutputPaneShown(window);
            tabsWidget.selectedItem = tabWidget; 
        }
    }
}

function SCCScinAppendText(text) {
    if (typeof(text)=='undefined' || !text) return;
    var scin = document.getElementById('scin-sccoutput').scimoz;
    scin.readOnly = false;
    scin.currentPos = scin.textLength-1;
    // move cursor to end of current text
    text = "-------------------------------------------------------------------\n" +
        text+'\n';
    scin.addText(text.length, text);
    scin.gotoLine(scin.lineFromPosition(scin.textLength));
    scin.readOnly = true;
}

function SCCScinClearText() {
    var scin = document.getElementById('scin-sccoutput').scimoz;
    scin.readOnly = false;
    scin.text = '';
    scin.readOnly = true;
}


function SCCWarnAboutSSHIfNecessary() {
    var osSvc = Components.classes["@activestate.com/koOs;1"].getService();
    var CVS_RSH = osSvc.getenv('CVS_RSH');

    // Only bother warning if CVS_RSH is set and CVS integration and CVS
    // background checking are enabled.
    if (CVS_RSH
        && gPrefs.getBooleanPref("cvsEnabled")
        && gPrefs.getBooleanPref("cvsBackgroundCheck")
        // Checking this pref (seemingly redundantly) here is necessary
        // because the pref is being used to mean:
        //  "don't ask or do anything if true"
        // rather than the usual meaning:
        //  "don't ask, just presume user wants same action as last time"
        && !gPrefs.getBooleanPref("donotask_cvs_ssh_setup_warning")
       )
    {
        var msg = "The CVS_RSH environment variable is set to '"+CVS_RSH+"' and CVS Background Status Checking is currently turned on.  Failure to properly configure CVS authentication may result in process lockups within Komodo.  Please read the Komodo documentation regarding the use of external protocols, such as SSH, with Komodo.  \nIf you are sure you have properly configured CVS/SSH authentication on your machine, click 'Keep Enabled'.  Otherwise click 'Disable' to disable Komodo's CVS integration for now. When you are sure you have setup CVS properly you can re-enable CVS integration in Edit|Preferences|Source Control|CVS.";
        var answer = dialog_customButtons(msg,
            ["Disable", "Disable and Tell Me More", "Keep Enabled"],
            "Disable", null, "Komodo CVS Integration",
            "cvs_ssh_setup_warning");
        switch (answer) {
            case "Keep Enabled":
                break;
            case "Disable and Tell Me More":
                launch_MainHelp("chrome://jaguar/content/doc/komodo/komodo-doc-scc.html#config_cvs");
                // fall through
            case "Disable":
            case "Cancel":
                gPrefs.setBooleanPref("cvsEnabled", false);
        }
    }
}
