/* Copyright (c) 2000-2001 ActiveState Corporation.
   See the file LICENSE.txt for licensing information. */

/* -*- Mode: Java; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

// The status bar API for Komodo.
//
// Here is an excerpt from the status bar requirements document:
//  --------------------------------------------------------------------------
//  |  <message>                  | <ck> | <enc.> | <pos.> | <lang.> | <ro> /,
//  --------------------------------------------------------------------------
//     ^--(a) Message                ^       ^        ^         ^       ^    ^
//        (b) Linter/Checker status -'       |        |         |       |    |
//        (c) Encoding ----------------------'        |         |       |    |
//        (d) Position (line and column) -------------'         |       |    |
//        (e) Language -----------------------------------------'       |    |
//        (f) Read-only ------------------------------------------------'    |
//        (z) Grippy --------------------------------------------------------'
//
// Implementation:
//
// The Status Bar observes various status notification messages (e.g.
// 'current_view_encoding_changed') via Mozilla's nsIObserver mechanism.
// Because some of the
// status bar segments are specific to the currently focused view ((b)-(f)),
// the preference is that notifications only be sent for changes to the current
// view. This is not always easy, so in some cases the logic in this module
// must ensure that notifications are for the current view.
//
// How to make a status message:
// Method 1.
//      If you are in JavaScript code you can just issue a
//      StatusBar_AddMessage(...) call. See that function for a description
//      of the arguments.
// Method 2.
//      You can send a 'status_message' notification via the nsIObserver
//      mechanism where the 'subject' argument of the Notify() is a
//      koIStatusMessage object. Here is an example in Python code:
//
//        observerSvc = components.classes["@mozilla.org/observer-service;1"]\
//                      .getService(components.interfaces.nsIObserverService)
//        sm = components.classes["@activestate.com/koStatusMessage;1"]\
//             .createInstance(components.interfaces.koIStatusMessage)
//        sm.category = "<some string>"
//        sm.msg = "<some string>"
//        sm.timeout = 0    # 0 for no timeout, else a number of milliseconds
//        sm.highlight = 0  # boolean, whether or not to highlight
//        try:
//            observerSvc.notifyObservers(sm, 'status_message', None)
//        except COMException, e:
//            # do nothing: Notify sometimes raises an exception if (???)
//            # receivers are not registered?
//            pass
//
// TODO:
//  - double-click and right-click functionality as per the requirements doc
//  - read-only status icon
//  - better icons for showing check status (more obvious, the inprogress one
//    must at least be replaced)                                             


//---- globals

var gStatusLogger = getLoggingMgr().getLogger('statusbar');
//gStatusLogger.setLevel(LOG_DEBUG);
var gStatusBarObserver;
var gStatusMessageStack = null;
var gStatusBarInitialized = false;


//---- helper functions

function _StatusBar_UpdateEncoding(view) {
    if (typeof(view)=='undefined' || !view || !view.document)
        return;
    try {
        //XXX It would probably be cleaner to add an "encoding_name"
        //    attribute to koIView and then let each view type override that.
        //    Then the view-startpage could return null to indicate N/A.
        if (view.getAttribute("type") == "startpage") {
            _StatusBar_ClearEncoding();
        } else {
            var encoding = view.document.encoding.short_encoding_name;
            var encodingWidget = document.getElementById('statusbar-encoding');
            encodingWidget.setAttribute("label", encoding);
        }
    } catch(e) {
        _StatusBar_ClearEncoding();
    }
}


function _StatusBar_ClearEncoding() {
    var encodingWidget = document.getElementById('statusbar-encoding');
    encodingWidget.removeAttribute("label");
}


function _StatusBar_UpdateLanguage(view) {
    if (typeof(view)=='undefined' || !view || !view.document)
        return;
    try {
        //XXX It would probably be cleaner to handle the "startpage language
        //    is N/A" logic in the view system, but I don't know how to
        //    easily do that right now.
        if (view.getAttribute("type") == "startpage") {
            _StatusBar_ClearLanguage();
        } else {
            language = view.document.language;
            var languageWidget = document.getElementById('statusbar-language');
            languageWidget.setAttribute("label", language);
        }
    } catch(e) {
        _StatusBar_ClearLanguage();
    }
}


function _StatusBar_ClearLanguage() {
    var languageWidget = document.getElementById('statusbar-language');
    languageWidget.removeAttribute("label");
}

var _updateLineColTimer = null;
function _StatusBar_doUpdateLineCol(lineColText) {
    _updateLineColTimer = null;
    var lineColWidget = document.getElementById('statusbar-line-col');
    lineColWidget.setAttribute('label', lineColText);
}

function _StatusBar_UpdateLineCol(view) {
    if (typeof(view)=='undefined' || !view)
        return;
    // Update line and column text.
    try {
        var lineColText = "Ln: " + view.currentLine + " Col: " + view.currentColumn;
        if (_updateLineColTimer) {
            clearTimeout(_updateLineColTimer);
        }
        setTimeout("_StatusBar_doUpdateLineCol('"+lineColText+"')",200);
    } catch(e) {
        // not a view that supports these
        _StatusBar_ClearLineCol();
    }

    // If there are lint result messages for the current line then display
    // them.
    if (typeof(view.lintBuffer)=='undefined' || !view.lintBuffer)
        return;
        
    var lintResults = view.lintBuffer.lintResults;
    var lintError = view.lintBuffer.errorString;
    var lintMessage = null;
    if (!lintError && lintResults) {
        var resultsObj = new Object();
        var numResultsObj = new Object();
        lintResults.getResultsAtPosition(view.currentLine,
                                         view.currentColumn,
                                         resultsObj,
                                         numResultsObj);
        var results = resultsObj.value;
        var numResults = numResultsObj.value;

        for (var i = 0; i < numResults; i++) {
            if (lintMessage) {
                lintMessage += "\n" + results[i].description;
            } else {
                lintMessage = results[i].description;
            }
        }
    }
    
    if (lintMessage) {
        StatusBar_AddMessage(lintMessage, "check", 0, false);
    } else {
        StatusBar_AddMessage(null, "check", 0, false);
    }
}


function _StatusBar_ClearLineCol() {
    var lineColWidget = document.getElementById('statusbar-line-col');
    lineColWidget.removeAttribute("label");
    StatusBar_AddMessage(null, "check", 0, false);
}


function _StatusBar_ClearCheck() {
    var checkWidget = document.getElementById('statusbar-check');
    checkWidget.removeAttribute("src");
}


function _StatusBar_UpdateCheck(view) {
try {
    if (typeof(view)=='undefined' || !view || !view.prefs)
        return;

    // Update the status bar for the current check status.
    var checkWidget = document.getElementById('statusbar-check');

    // Only have linting for some view types (currently only 'editor').
    if (typeof(view.lintBuffer) == "undefined" || !view.lintBuffer)
    {
        checkWidget.removeAttribute("src");
        checkWidget.setAttribute("tooltiptext", "Syntax Checking Status");
        return;
    }

    // Is linting enabled?
    var checkingEnabled = view.prefs.getBooleanPref("editUseLinting");
    if (!checkingEnabled) {
        checkWidget.removeAttribute("src");
        checkWidget.setAttribute("tooltiptext", "Background Syntax Checking disabled");
        return;
    }

    if (typeof(view.lintBuffer)=='undefined' || !view.lintBuffer)
        return;

    // Is there an error in the linter?
    var lintError = view.lintBuffer.errorString;
    if (lintError) {
        checkWidget.setAttribute('src',
            "chrome://jaguar/skin/images/icon_check_error.png");
        checkWidget.setAttribute("tooltiptext",lintError);
        return;
    }

    // If so, get the lintResults.
    //   lintResults == null        -> inprogress
    //   lintResults.length == 0    -> ok
    //   lintResults.length > 0     -> errors/warnings
    var lintResults = view.lintBuffer.lintResults;
    if (!lintResults) {
        checkWidget.setAttribute('src',
            "chrome://jaguar/skin/images/icon_check_inprogress.png");
        checkWidget.setAttribute("tooltiptext",
                                 "Syntax Checking Status: in progress");
    } else {
        var resultsObj = new Object();
        var numResultsObj = new Object();
        lintResults.getResults(resultsObj, numResultsObj);
        if (numResultsObj.value == 0) {
            checkWidget.setAttribute('src',
                "chrome://jaguar/skin/images/icon_check_ok.png");
            checkWidget.setAttribute("tooltiptext",
                                     "Syntax Checking Status: ok");
        } else {
            checkWidget.setAttribute('src',
                "chrome://jaguar/skin/images/icon_check_error.png");
            checkWidget.setAttribute("tooltiptext",
                "Syntax Checking Status: " +
                lintResults.getNumErrors() + " error(s), "+
                lintResults.getNumWarnings() + " warning(s)");
        }
    }
} catch(ex) {
    gStatusLogger.exception(ex);
}
}

var _addMessageTimer = null;
function _StatusBar_AddMessageObject(sm)
{
    gStatusMessageStack.Push(sm);
    //dump("StatusBar: add message: current stack:\n");
    //gStatusMessageStack.Dump();
    
    // Get the latest message and show it in the UI.
    if (_addMessageTimer) {
        clearTimeout(_addMessageTimer);
    }
    _addMessageTimer = setTimeout("_StatusBar_UpdateMessage();", 200);
    // If this new message timesout then schedule an update for then as well.
    if (sm.timeout > 0) {
        // Allow for some inaccuracy in timeout scheduling to ensure that
        // the message has actually timed-out before updating.
        var epsilon = 300;
        setTimeout("_StatusBar_UpdateMessage();", sm.timeout+epsilon);
    }
}

function _StatusBar_UpdateMessage()
{
    _addMessageTimer = null;
    //dump("StatusBar: update message: current stack:\n");
    //gStatusMessageStack.Dump();

    // Get the latest message and show it in the UI.
    var sm = gStatusMessageStack.Top();
    var messageWidget = document.getElementById('statusbar-message');
    if (sm) {
        gStatusLogger.debug("StatusBar: update message: top: msg='"+sm.msg+"', category='"+
             sm.category+"', timeout='"+sm.timeout+"', highlight='"+
             sm.highlight+"'");
        
        messageWidget.setAttribute("category", sm.category);
        messageWidget.setAttribute("label", sm.msg);
        messageWidget.setAttribute("tooltiptext", sm.msg);
        if (sm.highlight) {
            messageWidget.setAttribute("highlite","true");
        } else {
            messageWidget.removeAttribute("highlite");
        }
    } else {
        messageWidget.setAttribute("label","Ready");
        messageWidget.removeAttribute("tooltiptext");
        messageWidget.removeAttribute("highlite");
    }
}


function _StatusBar_Clear() {
    _StatusBar_ClearEncoding();
    _StatusBar_ClearLanguage();
    _StatusBar_ClearLineCol();
    _StatusBar_ClearCheck();
}

//---- local classes

function StatusBarObserver() {};
StatusBarObserver.prototype.observe = function(subject, topic, data)
{
    // Unless otherwise specified the 'subject' is the view,
    // and 'data' arguments are
    // expected to be empty for all notifications.
    gStatusLogger.debug("StatusBar observed '"+topic+"': ");
    var view = subject;
    if (!subject) {
        _StatusBar_Clear();
        return;
    }
    
    switch (topic) {
    case 'view_opened':
    case 'view_closed':
        _StatusBar_Clear();
        break;
    case 'current_view_changed':
        _StatusBar_UpdateEncoding(view);
        _StatusBar_UpdateLanguage(view);
        _StatusBar_UpdateLineCol(view);
        _StatusBar_UpdateCheck(view);
        break;
    case 'current_view_encoding_changed':
        _StatusBar_UpdateEncoding(gViewMgr.currentView);
        break;
    case 'current_view_language_changed':
        _StatusBar_UpdateLanguage(gViewMgr.currentView);
        break;
    case 'current_view_linecol_changed':
        _StatusBar_UpdateLineCol(gViewMgr.currentView);
        break;
    case 'current_view_check_status':
        _StatusBar_UpdateCheck(gViewMgr.currentView);
        break;

    case 'scc_status_message':
        // fall through

    case 'status_message':
        // "subject" is expected to be a koIStatusMessage object.
        _StatusBar_AddMessageObject(subject);
        break;
    }
}

//---- public functions

function StatusBar_AddMessage(msg, category, timeout, highlight)
{
    // Post a message to the status bar message area.
    // "msg" is the message string. An empty string or null indicates
    //      that the message (of the given category) should be cleared.
    // "category" is the message group to which the message belongs. It
    //      is an arbitrary string (it must be at least one character).
    // "timeout" is the amount of time, in milliseconds, that the message
    //      should appear. A value of 0 indicates that the message does
    //      not timeout.
    // "highlight" is a boolean indicating whether the message should be
    //      highlighted on the status bar.
    //
    // A structure similar to a stack of status messages is maintained.
    // The latest message is always shown. When/if it timesout then the
    // previous message is the stack is displayed. There can only be one
    // message per category, so reusing a category allows for removal of
    // status messages that are no longer appropriate.
    //
    // To add a message that does not timeout:
    //  StatusBar_AddMessage("hello there", "my_category", 0, false)
    // To remove that message:
    //  StatusBar_AddMessage(null, "my_category", 0, false)
    // To add a highlighted message for three seconds:
    //  StatusBar_AddMessage("hello there", "my_category", 3000, true)
    if (!gStatusBarInitialized) {
        gStatusLogger.warn("Adding status bar message before it is "+
                           "initialized. Dropping it.");
        return;
    }

    gStatusLogger.debug("StatusBar: add message: msg='"+msg+"', category='"+category+
        "', timeout='"+timeout+"', highlight='"+highlight+"'");

    // create a status message component and insert it into the stack
    var sm = Components.classes["@activestate.com/koStatusMessage;1"].
          createInstance(Components.interfaces.koIStatusMessage);
    sm.msg = msg;
    sm.category = category;
    sm.timeout = timeout;
    sm.highlight = highlight;
    
    _StatusBar_AddMessageObject(sm);
}

var _gStatusBarPrefObserver = null;
function StatusBarPrefObserver()
{
    this.prefSvc = Components.classes["@activestate.com/koPrefService;1"].
                  getService(Components.interfaces.koIPrefService);
    this.prefSvc.prefs.prefObserverService.addObserver(this,'editUseLinting',0);
};
StatusBarPrefObserver.prototype.destroy = function()
{
    this.prefSvc = Components.classes["@activestate.com/koPrefService;1"].
                  getService(Components.interfaces.koIPrefService);
    this.prefSvc.prefs.prefObserverService.removeObserver(this,'editUseLinting');
    this.prefSvc = null;
}
StatusBarPrefObserver.prototype.observe = function(prefSet, prefSetID, prefName)
{
    if (!gStatusBarInitialized) {
        gStatusLogger.warn("Adding status bar message before it is "+
                           "initialized. Dropping it.");
        return;
    }

    gStatusLogger.debug("StatusBar: observed pref '"
            + prefName + "' change (prefSet="
            + prefSet + ", prefSetID="
            + prefSetID)
    if (prefName == "editUseLinting") {
        var view = gViewMgr.currentView;
        if (view) {
            _StatusBar_UpdateCheck(view);
        }
    }
};


function StatusBar_Init()
{
    KomodoRegisterOnUnloadHandler(StatusBar_Fini);
    gStatusBarObserver = new StatusBarObserver();

    var obsSvc = Components.classes["@mozilla.org/observer-service;1"].
                       getService(Components.interfaces.nsIObserverService);
    obsSvc.addObserver(gStatusBarObserver, 'view_opened',false);
    obsSvc.addObserver(gStatusBarObserver, 'view_closed',false);
    obsSvc.addObserver(gStatusBarObserver, 'current_view_changed',false);
    obsSvc.addObserver(gStatusBarObserver, 'status_bar_reset',false);
    obsSvc.addObserver(gStatusBarObserver, 'current_view_encoding_changed',false);
    obsSvc.addObserver(gStatusBarObserver, 'current_view_language_changed',false);
    obsSvc.addObserver(gStatusBarObserver, 'current_view_linecol_changed',false);
    obsSvc.addObserver(gStatusBarObserver, 'current_view_check_status',false);
    obsSvc.addObserver(gStatusBarObserver, 'status_message',false);

    obsSvc.addObserver(gStatusBarObserver, 'scc_status_message',false);


    _gStatusBarPrefObserver = new StatusBarPrefObserver();

    gStatusMessageStack = Components.classes["@activestate.com/koStatusMessageStack;1"].
                          createInstance(Components.interfaces.koIStatusMessageStack);

    gStatusBarInitialized = true;
}


function StatusBar_Fini()
{
    var obsSvc = Components.classes["@mozilla.org/observer-service;1"].
                       getService(Components.interfaces.nsIObserverService);
    obsSvc.removeObserver(gStatusBarObserver, 'view_opened');
    obsSvc.removeObserver(gStatusBarObserver, 'view_closed');
    obsSvc.removeObserver(gStatusBarObserver, 'current_view_changed');
    obsSvc.removeObserver(gStatusBarObserver, 'status_bar_reset');
    obsSvc.removeObserver(gStatusBarObserver, 'current_view_encoding_changed');
    obsSvc.removeObserver(gStatusBarObserver, 'current_view_language_changed');
    obsSvc.removeObserver(gStatusBarObserver, 'current_view_linecol_changed');
    obsSvc.removeObserver(gStatusBarObserver, 'current_view_check_status');
    obsSvc.removeObserver(gStatusBarObserver, 'status_message');

    obsSvc.removeObserver(gStatusBarObserver, 'scc_status_message');


    _gStatusBarPrefObserver.destroy();
    _gStatusBarPrefObserver = null;
    gStatusBarObserver = null;
    gStatusMessageStack = null;
}

function Test_DumpStatusManager()
{
    // dumps to stdout
    gStatusMessageStack.Dump();
}
