# Copyright (c) 2000-2001 ActiveState Corporation.
# See the file LICENSE.txt for licensing information.

import exceptions
import os
from xpcom import components, nsError, ServerException, COMException, Exception, _xpcom
from xpcom.server.enumerator import SimpleEnumerator
import logging

log = logging.getLogger("koDebugger")

PROXY_SYNC    = 0x0001  # acts just like a function call.
PROXY_ASYNC   = 0x0002  # fire and forget.  This will return immediately and you will lose all return information.
PROXY_ALWAYS  = 0x0004  # ignore check to see if the eventQ is on the same thread as the caller, and alway return a proxied object.

DBG_STATE_STOPPED = components.interfaces.koIDebugger.DBG_STATE_STOPPED
DBG_STATE_BREAK = components.interfaces.koIDebugger.DBG_STATE_BREAK
DBG_STATE_RUNNING = components.interfaces.koIDebugger.DBG_STATE_RUNNING
DBG_STATE_STARTING = components.interfaces.koIDebugger.DBG_STATE_STARTING;
DBG_STATE_STOPPING = components.interfaces.koIDebugger.DBG_STATE_STOPPING;

class SimpleEnumeratorAggregator:
    _com_interfaces_ = [components.interfaces.nsISimpleEnumerator]

    def __init__(self, enumerators):
        self.enumerators = enumerators

    def _movenext(self):
        while len(self.enumerators)>0 and not self.enumerators[0].hasMoreElements():
            del self.enumerators[0]

    def hasMoreElements(self):
        self._movenext()
        return len(self.enumerators)>0

    def getNext(self):
        ret = self.enumerators[0].getNext()
        self._movenext()
        return ret


class koDebugger:
    _com_interfaces_ = [components.interfaces.koIDebugger,
                        components.interfaces.koIDebuggingFramework,
                        ]
    _reg_clsid_ = "{cdb2611a-3357-41e9-b62f-42bd3230e7e6}"
    _reg_contractid_ = "@activestate.com/debugger;1"
    def __init__(self):
        self.engines = {}
        self.break_engine = None
        self.listeners = {}
        self.listener_cookie = 0
        self.numEngines = 0
        self.isStarting = 0

    def addEngine(self, engineName):
        # Result: void (None)
        # In: param0: koIDebugEngine
        existing = self.engines.get(engineName, None)
        if existing is not None:
            return existing
        # Create a new one.
        klass = components.classes["@activestate.com/debugger_engine_"+engineName+";1"]
        engine = klass.createInstance(components.interfaces.koIDebugEngine)
        self.engines[engineName] = engine
        self.numEngines += 1 # public xpcom property we keep up to date
        return engine

    def get_state( self ):
        # XXX - get smarter with multi engines.
        for engine in self.engines.values():
            state = engine.state
            if state != DBG_STATE_STOPPED:
                return state
        if self.isStarting:
            return DBG_STATE_STARTING
        return DBG_STATE_STOPPED

    def set_state(self, new_state):
        # I guess this is a little hacky - we only allow the state to explicitly be set
        # from STOPPED to STARTING.
        if new_state != DBG_STATE_STARTING:
                raise COMException(nsError.NS_ERROR_UNEXPECTED, "Can only set the new state to STARTING")
            
        for engine in self.engines.values():
            state = engine.state
            if state != DBG_STATE_STOPPED:
                raise COMException(nsError.NS_ERROR_UNEXPECTED, "All engines must be stopped to set the state to STARTING")
        self.isStarting = 1
        self.onStateChange(None, new_state, 0)

    def startDebugging(self, stepType, preferences, stdin, stdout, stderr):
        language = preferences.getStringPref("debugger.language")
        engine = self._getDebugEngine(language)
        if engine is None:
            log.error("Attempt to start debugging with language engine '%s', but this engine does not exist" % (language,))
            raise COMException(nsError.NS_ERROR_NOT_AVAILABLE)

        try:
            engine.startDebugging(stepType, preferences, stdin, stdout, stderr)
        except COMException, err:
            raise ServerException(err.errno)

    def resume( self, how ):
        # Result: void (None)
        # In: param0: int32
        e = self.break_engine
        if e is None:
            raise ServerException(nsError.NS_ERROR_UNEXPECTED)
        return e.resume(how)

    def breakNow( self ):
        # Result: void (None)
        for engine in self.engines.values():
            engine.breakNow()

    def stop( self ):
        # Result: void (None)
        was_running = 0
        for engine in self.engines.values():
            if engine.state != DBG_STATE_STOPPED:
                was_running = 1
            engine.stop()
        # And drop all references to my engines
        # (they may still fire state event transitions, tho!)
        self.break_engine = None
        # self.engines = {}
        if not was_running and self.isStarting:
            self.isStarting = 0
            self.onStateChange(None, DBG_STATE_STOPPED, components.interfaces.koIDebugger.STOP_REASON_ERROR)

    def enumStackFrames( self ):
        # Result: nsISimpleEnumerator
        e = self.break_engine
        if e is None:
            log.warn("enumStackFrames - No debugger is in a break state") # mmm, steak ;-)
            raise ServerException(nsError.NS_ERROR_UNEXPECTED)
        return e.enumStackFrames()

    def getSourceCode( self, filename ):
        # Result: wstring
        e = self.break_engine
        if e is None:
            log.warn("No debugger is in a break steak")
            raise ServerException(nsError.NS_ERROR_UNEXPECTED)
        return e.getSourceCode(filename)

    # Breakpoints are annoying - they have to work in any engine state.
    def enumBreakpoints(self, url):
        # Result: nsISimpleEnumerator
        ret = []
        for e in self.engines.values():
            ret.append( e.enumBreakpoints( url ) )
        return SimpleEnumeratorAggregator(ret)

    def addBreakpoint(self, language, url, lineno):
        # Result: koIBreakpoint
        engine = self._getDebugEngine(language)
        if engine is None:
            log.warn("Attempt to add breakpoint for language '%s', but this language has no debugger" % (language,))
            return None
        # engines handle thread-state etc.
        return engine.addBreakpoint( url, lineno )


    def addListener(self, listener):
        # Result is a cookie, as a long.
        listener = _xpcom.getProxyForObject(0, components.interfaces.koIDebuggerEvents,
                                            listener, PROXY_ASYNC|PROXY_ALWAYS)
        # XXX - this is not thread-safe!
        cookie = self.listener_cookie = self.listener_cookie + 1
        self.listeners[cookie] = listener
        return cookie

    def removeListener(self, cookie):
        if not self.listeners.has_key(cookie):
            raise ServerException(nsError.NS_ERROR_UNEXPECTED)
        del self.listeners[cookie]

    def _fire_event(self, event_name, *args):
        for l in self.listeners.values():
            method = getattr(l, event_name)
            try:
                apply(method, args)
            except ServerException, details:
                log.warn("Firing event %s, failed, %s" %(method, details))

    def onStateChange(self, engine, state, stateExtra):
        if state != DBG_STATE_STARTING:
            self.isStarting = 0
            
        if state == DBG_STATE_BREAK:
            assert self.break_engine is None, "Moving to break state - should not have an existing break engine"
            self.break_engine = engine
        else:
            self.break_engine = None
        self._fire_event("onStateChange", state, stateExtra)

    def _getDebugEngine(self, language):
        # if we are in the debugger, then a debuggerLanguageService must exist
        registry = components \
                  .classes["@activestate.com/koLanguageRegistryService;1"] \
                  .getService(components.interfaces.koILanguageRegistryService)
        langService = registry.getLanguage(language) \
                  .getLanguageService(components.interfaces.koIDebuggerLanguageService)
        if langService is None:
            return None
        return langService.engine
