# Copyright (c) 2000-2003 ActiveState Corporation.
# See the file LICENSE.txt for licensing information.

# An implementation of a file object for Mozilla/xpcom.

from xpcom import components, nsError, ServerException, COMException, _xpcom
from xpcom.server import WrapObject, UnwrapObject
import xpcom
import logging
import eollib
import md5
import re
import sys
import cStringIO
import timeline
import stat, os, time

import koUnicodeEncoding, codecs

log = logging.getLogger('koDocument')

class koDocumentBase:
    _com_interfaces_ = [components.interfaces.koIDocument,
                        components.interfaces.nsIObserver]
    _reg_desc_ = "Komodo Document"
    _reg_contractid_ = "@activestate.com/koDocumentBase;1"
    _reg_clsid_ = "{A9F51FD2-CF82-4290-87B8-BD07CEBD1CD1}"

    re_firstline = re.compile(ur'(.*?)(?:\r|\n|$)')
    _indentWidth = None
    _tabWidth = None
    _useTabs = None

    def __init__(self):
        timeline.enter('koDocumentBase.__init__')
        # Grab a reference to the global preference service.
        self._globalPrefSvc = components.classes["@activestate.com/koPrefService;1"].\
                            getService(components.interfaces.koIPrefService)
        self._globalPrefs = self._globalPrefSvc.prefs

        self._buffer = None # string The contents of the document
        if sys.platform.startswith('win'):
            self._codePage = 65001 # long: the code page -- 65001 if unicode
        else:
            self._codePage = 0
        self.encoding = None # string The name of the Unicode encoding -- undefined if codepage is not 65001
        self._language = None # string what language is this document in?
        self._languageObj = None
        self.prefs = None # set after initWith() call
        self.file = None
        self._isDirty = 0     # boolean
        self.isUntitled = 1  # boolean
        self._views = [] # scintilla widget instances
        self._document = None # scimoz.docpointer
        #XXX should get eol from prefs and/or from document content
        self._eol = eollib.EOL_PLATFORM
        self._lastmd5 = None
        self._refcount = 0
        self.docSettingsMgr = None
        self.encodingServices = components.classes['@activestate.com/koEncodingServices;1'].\
                         getService(components.interfaces.koIEncodingServices);

        # yes, createInstance.  We want to provide our own observer services
        # for documents
        self.observerService = components.classes['@activestate.com/koObserverService;1'].\
                                createInstance(components.interfaces.nsIObserverService)

        self.lastErrorSvc = components.classes["@activestate.com/koLastErrorService;1"].\
                                getService(components.interfaces.koILastErrorService)
        
        self._obsSvc = components.classes['@mozilla.org/observer-service;1'].\
                                getService(components.interfaces.nsIObserverService);
        self._obsSvcProxy = _xpcom.getProxyForObject(1, components.interfaces.nsIObserverService,
                                          self._obsSvc, _xpcom.PROXY_SYNC | _xpcom.PROXY_ALWAYS)
        self.docSettingsMgr = components.classes['@activestate.com/koDocumentSettingsManager;1'].\
            createInstance(components.interfaces.koIDocumentSettingsManager);

        timeline.leave('koDocumentBase.__init__')
    
    def __del__(self):
        assert not self._views

    #void initWithFile(in koIFileEx file);
    def initWithFile(self, file, untitled):
        timeline.enter('koDocumentBase.initWithFile')
        log.info("initWithFile")
        self.isUntitled = untitled
        self.file = file
        self._setupPrefs()
        if self._language is None:
            self._guessLanguage()
        timeline.leave('koDocumentBase.initWithFile')

    def initWithURI(self, uri):
        timeline.enter('koDocumentBase.initWithURI')
        log.info("initWithURI")
        self.isUntitled = 0
        filesvc = components.classes["@activestate.com/koFileService;1"] \
                        .createInstance(components.interfaces.koIFileService)
        self.file = filesvc.getFileFromURI(uri)
        self._setupPrefs()
        if self._language is None:
            self._guessLanguage()
        timeline.leave('koDocumentBase.initWithURI')

    def initUntitled(self, name, encoding):
        timeline.enter('koDocumentBase.initUntitled')
        log.info("initUntitled")
        self.isUntitled = 1
        self._untitledName = name
        self._setupPrefs(encoding)
        self.set_buffer("", 0)  # make sure new buffer is not dirty
        if self._language is None:
            self._guessLanguage()
        timeline.leave('koDocumentBase.initWithURI0')

    def addReference(self):
        self._refcount += 1
        log.debug('refcount = %d', self._refcount)

    def releaseReference(self):
        if self._refcount <= 0:
            log.error("Trying to release a reference with a refcount < 1! for displayPath: %s",
                      self.get_displayPath())
        else:
            self._refcount -= 1
        log.debug('refcount = %d', self._refcount)
        
    def _setupPrefs(self, encoding_name=None):
        """ We can only setup the prefs on the document once we have a URI for it
        So after __init__, self.prefs is None, until we get an "initWith....()" call.
        Note that some things like encoding derive from prefs.
        """
        # Create a preference set to hold doc preferences
        docStateMRU = self._globalPrefSvc.getPrefs("docStateMRU");
        if not self.isUntitled and docStateMRU.hasPref(self.file.URI):
            url = self.file.URI
            docState = docStateMRU.getPref(url)
            self.prefs = docState
        else:
            self.prefs = components.classes['@activestate.com/koPreferenceSet;1'].\
                                     createInstance(components.interfaces.koIPreferenceSet)
            if self.isUntitled:
                self.prefs.id = self._untitledName
            else:
                self.prefs.id = self.file.URI
            docStateMRU.setPref(self.prefs)
        self.prefs.parent = self._globalPrefs

        # Set the default encoding for the file
        self.encoding = components.classes['@activestate.com/koEncoding;1'].\
                                 createInstance(components.interfaces.koIEncoding)
        if not encoding_name:
            encoding_name = self._getStringPref('encodingDefault')
        encoding = self.encodingServices.get_encoding_info(encoding_name)\
                       .python_encoding_name
        self.encoding.python_encoding_name =\
            self.encodingServices.get_canonical_python_encoding_name(encoding)

        if self.prefs.hasPref('language'):
            #print "found language in prefs: ", self.prefs.getStringPref('language')
            self.set_language(self.prefs.getStringPref('language'))

        # setup an observer on our own prefs since we provide access to some
        # through getters because e.g. indentWidth is computed on some cases, and
        # yet not stored in prefs except if set explicitely.
        self.prefs.addObserver(WrapObject(self, components.interfaces.nsIObserver))

    def _guessLanguage(self):
        #print "in _guessLanguage"
        timeline.enter('koDocumentBase._guessLanguage')
        if self.prefs.hasPrefHere('language'):
            self._language = self.prefs.getStringPref('language')
            #print "found from prefs:", self._language
            return
        registryService = components.classes['@activestate.com/koLanguageRegistryService;1'].\
                                     getService(components.interfaces.koILanguageRegistryService)
        fileNameLanguage = registryService.suggestLanguageForFile(self.get_baseName())
        contentLanguage = None
        contentOtherLanguages = []
        if (self._globalPrefs.getBooleanPref('langCharacteristics') or \
           (fileNameLanguage == '')) and (self.get_buffer() is not None):
            # optimization: only send the first thousand chars.  We need
            # more than the first line because for e.g. XHTML, the first line
            # doesn't provide enough information.
            buffer = self.get_buffer()[:1000]
            languages = registryService.guessLanguageFromContents(buffer, 'Text')
            if languages:
                contentLanguage = languages[0]
                if len(languages) > 1:
                    contentOtherLanguages = languages[1:]

        if contentLanguage is None:
            language = fileNameLanguage or 'Text'
        else:
            if fileNameLanguage in contentOtherLanguages:
                language = fileNameLanguage
            else:
                language = contentLanguage
        self._language = language
        #print "guessed:", language
        timeline.leave('koDocumentBase._guessLanguage')

    def load(self):
        if self.isUntitled or self.file.URI.startswith('chrome://'):
            return
        timeline.enter('koDocumentBase.load')
        self._loadfile(self.file)
        self._guessLanguage()
        eolpref = self.prefs.getStringPref('endOfLine')
        if self.prefs.hasPrefHere('endOfLine'):
            current_eol = eollib.eolPref2eol[eolpref]
        else:
            current_eol = self.get_existing_line_endings()
            if current_eol in (eollib.EOL_MIXED, eollib.EOL_NOEOL):
                current_eol = eollib.eolPref2eol[eolpref]
        self.set_new_line_endings(current_eol)
            
        timeline.leave('koDocumentBase.load')

    def _loadfile(self, file):
        timeline.enter('koDocumentBase._loadfile')
        if file:
            data = self._get_buffer_from_file(file)
        else:
            data = ''
        self._lastmd5 = md5.new(data).digest()
        self.set_buffer(data,0)
        self.setSavePoint()
        # if a file is in a project, then we have to check
        # and see if we need to update it's information
        if file.hasChanged:
            try:
                self._obsSvc.notifyObservers(self,'file_changed',self.file.URI)
            except:
                # ignore, noone listening
                pass
        self.set_isDirty(0)
        timeline.leave('koDocumentBase._loadfile')
        
    def _get_buffer_from_file(self, file):
        timeline.enter('koDocumentBase._get_buffer_from_file')
        try:
            file.open('rb')
            data = file.read(-1)
            file.close()
        except COMException, ex:
            # koFileEx.open(), .read(), and .close() will already
            # setLastError on failure so don't need to do it again. The
            # only reason we catch it to just re-raise it is because
            # PyXPCOM complains on stderr if a COMException passes out
            # of the Python run-time.
            raise ServerException(ex.errno, str(ex))
        timeline.leave('koDocumentBase._get_buffer_from_file')
        return data

    def get_isDirty(self):
        return self._isDirty
    
    def set_isDirty(self,isDirty):
        self._isDirty = isDirty
        try:
            self.observerService.notifyObservers(self,'buffer_dirty',str(isDirty))
        except COMException, e:
            pass # no one is listening!

    def differentOnDisk(self):
        if self.isUntitled or \
            not self.file.isLocal or \
            not self.file.exists or \
            self.file.URI.startswith('chrome://'):
            return 0
        # compare contents of the buffer with the contents of the file on disk
        newmd5 = None
        try:
            self.file.open('rb')
        except:
            # the file is gone.
            newmd5 = None
            return newmd5 != self._lastmd5
        try:
            ondisk = self.file.read(-1)
            newmd5 = md5.new(ondisk).digest()
        finally:
            self.file.close()
        return newmd5 != self._lastmd5

    def get_baseName(self):
        if self.isUntitled:
            return self._untitledName
        else:
            return self.file.baseName

    def set_baseName(self, val):
        if self.isUntitled:
            self._untitledName = val
        else:
            self.file.baseName = val

    def get_displayPath(self):
        if self.isUntitled:
            return self._untitledName
        else:
            return self.file.displayPath

    def get_language(self):
        return self._language
    
    def set_language(self,language):
        log.info("setting language to " + language);
        self._language = language
        
        if language == '':
            if self.prefs.hasPrefHere('language'):
                self.prefs.deletePref('language')
            self._guessLanguage()
        else:
            self.prefs.setStringPref('language', language)

        self._languageObj = None
        try:
            self.observerService.notifyObservers(self,'language_changed',language)
        except COMException, e:
            pass # no one is listening!

    def get_languageObj(self):
        if self._language is None:
            log.error('Asked to get language Object with no language')
        if self._languageObj == None:
            registryService = components.classes['@activestate.com/koLanguageRegistryService;1'].\
                getService(components.interfaces.koILanguageRegistryService)
            self._languageObj = registryService.getLanguage(self._language)
        return self._languageObj
    
    def get_codePage(self):
        return self._codePage
    
    def set_codePage(self,codePage):
        self._codePage = codePage
        for view in self._views:
            view.scimoz.codePage = codePage

    def get_buffer(self):
        if self._document:
            return self._views[0].scimoz.text
        return self._buffer
    
    def set_buffer(self, text, makeDirty=1):
        # detect encoding and set codePage, buffer
        if text:
            if not isinstance(text, unicode):
                encoded_buffer, encoding_name, bom = self._detectEncoding(text)
                self.encoding.python_encoding_name =\
                    self.encodingServices.get_canonical_python_encoding_name(encoding_name)
                self.encoding.use_byte_order_marker = bom
                try:
                    self.observerService.notifyObservers(self,'encoding_changed',self.encoding.python_encoding_name)
                except COMException, e:
                    pass # no one is listening!
                self._determineCodePage()
            else:
                encoded_buffer = text
            self._set_buffer_encoded(encoded_buffer, makeDirty)
        else:
            self._set_buffer_encoded(text, makeDirty)

    def _set_buffer_encoded(self,text,makeDirty=1):
        was_dirty = self.get_isDirty()
        if self._document:
            scimoz = self._views[0].scimoz
            cp = scimoz.currentPos
            an = scimoz.anchor
            fvl = scimoz.firstVisibleLine
            xoffset = scimoz.xOffset
            scimoz.beginUndoAction()
            scimoz.text = text
            textLength = scimoz.textLength # length in bytes. =(
            scimoz.endUndoAction()
            scimoz.currentPos = min(textLength, cp)
            scimoz.anchor = min(textLength, an)
            scimoz.lineScroll(0, min(fvl-scimoz.firstVisibleLine, scimoz.lineCount-scimoz.firstVisibleLine))
            scimoz.xOffset = xoffset
        else:
            self._buffer = text
        self.set_isDirty(was_dirty or makeDirty)
        try:
            self.observerService.notifyObservers(self,'buffer_changed','')
        except COMException, e:
            pass # no one is listening!

    def get_bufferLength(self):
        # XXX as we add more methods, we'll need a better system
        if self._document:
            return self._views[0].scimoz.textLength
        if self._buffer:
            return len(self._buffer)
        return 0

    def set_existing_line_endings(self, le):
        if le not in (eollib.EOL_LF, eollib.EOL_CR, eollib.EOL_CRLF):
            raise ServerException(nsError.NS_ERROR_FAILURE,
                                  "Invalid line ending: %s" % le)
        log.info("set_existing_line_endings to '%s'", eollib.eol2eolName[le])
        newBuffer = eollib.convertToEOLFormat(self.get_buffer(), le)
        self._set_buffer_encoded(newBuffer)

    def get_existing_line_endings(self):
        endings, recommended = eollib.detectEOLFormat(self.get_buffer())
        if endings == eollib.EOL_NOEOL:
            return self.get_new_line_endings()
        else:
            return endings

    def get_new_line_endings(self):
        return self._eol

    def set_new_line_endings(self, le):
        if le not in (eollib.EOL_LF, eollib.EOL_CR, eollib.EOL_CRLF):
            raise ServerException(nsError.NS_ERROR_FAILURE,
                                  "Invalid new line ending: %s" % le)
        log.info("set_new_line_endings to '%s'", eollib.eol2eolName[le])
        self._eol = le
        for view in self._views:
            if view.scimoz:
                view.scimoz.eOLMode = eollib.eol2scimozEOL[le]

    def _getStringPref(self, name, default=None):
        if self.prefs.hasPref(name):
            return self.prefs.getStringPref(name)
        return default

    def _getBooleanPref(self, name, default=0):
        if self.prefs.hasPref(name):
            return self.prefs.getBooleanPref(name)
        return default

    def _getEncodingFromName(self, name):
        try:
            registryService = components.classes['@activestate.com/koLanguageRegistryService;1'].\
                                getService(components.interfaces.koILanguageRegistryService)
            prefs = self._globalPrefs
            languagesPrefs = prefs.getPref('languages')
            language = registryService.suggestLanguageForFile(name)
            if not language:
                language = 'Text'
            encoding = 'Default Encoding'
            if languagesPrefs.hasPref('languages/'+language):
                langPref = languagesPrefs.getPref('languages/'+language)
                if langPref.hasStringPref(language+'/newEncoding'):
                    encoding = langPref.getStringPref(language+'/newEncoding')
            if encoding == 'Default Encoding':
                encoding = prefs.getStringPref('encodingDefault')
        except Exception, e:
            log.error("Error getting newEncoding for %s", language, exc_info=1)
            encoding = prefs.getStringPref('encodingDefault')
        return encoding

    def _detectEncoding(self, buffer):
        if not isinstance(buffer,str):
            errstr = 'buffer is not a string!'
            log.error(errstr)
            self.lastErrorSvc.setLastError(nsError.NS_ERROR_FAILURE, errstr)
            raise ServerException(nsError.NS_ERROR_FAILURE, errstr)
        
        if '\0' in buffer:
            self.lastErrorSvc.setLastError(nsError.NS_ERROR_FAILURE,
                                  'Komodo cannot open files containing null characters.')
            raise ServerException(nsError.NS_ERROR_FAILURE,
                                  'Komodo cannot open files containing null characters.')
            
        # returns encoded buffer, encoding name, and bom
        bom = 0
        encoding_name = self._getStringPref('encodingDefault')
        encoding = self.encodingServices.get_encoding_info(encoding_name)\
                       .python_encoding_name
        tryencoding = self._getStringPref('encoding')
        if tryencoding is None:
            tryencoding = self._getEncodingFromName(self.get_baseName())
        tryxmldecl = self._getBooleanPref('encodingXMLDec')
        trymeta = self._getBooleanPref('encodingHTML')
        trymodeline = self._getBooleanPref('encodingModeline')
        autodetect = self._getBooleanPref('encodingAutoDetect')
        # The job of this function is to, given a buffer (and knowing what
        # document we're loading),
        # - 1. Figure out what the encoding of the buffer is.
        # - 2. Convert it to Unicode
        # - 3. Store the encoding of the buffer in the document instance so
        #     that save operations can do the right thing and convert back
        #     if necessary
        
        # Detecting what encoding the buffer is in.  There are many sources
        # of information.
        # First, preferences -- if there is an encoding pref for this document, use it.
        # If there are no prefs, look for unambiguous markers of encoding. These
        # consist of:
        #   - BOM markers or XML declarations
        #   - lines like
        #            -*- coding: <encoding name> -*-
        #    as discussed in http://www.python.org/peps/pep-0263.html
        #    for Python, emacs


        #    tryencoding = None
        if autodetect or tryencoding is not None:
            unicodebuffer, encoding, bom =\
                koUnicodeEncoding.autoDetectEncoding(buffer, tryxmldecl,
                                                     trymeta, trymodeline,
                                                     tryencoding,
                                                     encoding)
            bom = len(bom) > 0
        else:
            try:
                if encoding.startswith('utf'):
                    unicodebuffer = unicode(buffer, encoding)
                    if (len(unicodebuffer) > 0):
                        encodingName, bomLength = koUnicodeEncoding.checkBOM(unicodebuffer)
                        if bomLength:
                            unicodebuffer = unicodebuffer[bomLength:]
                            bom = 1
                else:
                    # It's an unknown 8-bit encoding.  Assume it's the
                    # configured default, and convert to unicode
                    encoding = encoding_name
                    unicodebuffer = unicode(buffer, encoding_name)
            except UnicodeError:
                errstr = 'The file cannot be opened using the encoding "%s". '\
                    'You can change the encoding used to open files in the '\
                    '"File Settings" tab of the "Preferences" dialog.'\
                    % self.encodingServices.get_encoding_info(encoding)\
                          .friendly_encoding_name
                self.lastErrorSvc.setLastError(0, errstr)
                raise ServerException(nsError.NS_ERROR_FAILURE, errstr)
            
        # now we have a unicode buffer.

        if buffer is None:
            encName = self.encodingServices.get_encoding_info(encoding)\
                          .friendly_encoding_name
            raise ServerException(nsError.NS_ERROR_FAILURE, 'The file encoding "%s" is not supported.' % encName)

        # We've determined that the encoding of the file is whatever 'encoding' is.
        return unicodebuffer, encoding, bom

    def get_isEncodable(self):
        try:
            self.encode()
            return 1
        except UnicodeError:
            return 0

    def forceEncodingFromEncodingName(self, encoding_name):
        encoding = components.classes['@activestate.com/koEncoding;1'].\
                                 createInstance(components.interfaces.koIEncoding)
        encoding.python_encoding_name = encoding_name
        self.set_encoding(encoding, 'replace')

    def setEncodingFromEncodingName(self, encoding_name):
        encoding = components.classes['@activestate.com/koEncoding;1'].\
                                 createInstance(components.interfaces.koIEncoding)
        encoding.python_encoding_name = encoding_name
        self.set_encoding(encoding)

    def setBufferAndEncoding(self, buffer, encoding_name):
        self._set_buffer_encoded('')
        self.setEncodingFromEncodingName(encoding_name)
        self._set_buffer_encoded(buffer)

    def set_encoding(self, encoding, errors="strict"):
        """Convert the current buffer to the given encoding.
        
            "encoding" is a koIEncoding to convert to.
            "errors" is identical to the "errors" argument to the Python
                .encode() string method. The Python docs say:
                    The default for errors is 'strict', meaning that
                    encoding errors raise a ValueError. Other possible
                    values are 'ignore' and 'replace'. 

        The python buffer is *always* in unicode format, so we want to
        encode to the new encoding, then get the unicode buffer and reset
        the buffer.
        
        This will raise a ServerException and set an error on
        koLastErrorService if there is an exception in the
        conversion. If the error code is zero then the error is an
        unexpected internal error and should be handled as such.
        """
        log.info("set_encoding(encoding=%r, errors=%r)", encoding, errors)
        lastErrorSet = 0 # set iff exception and last error has been set properly
        was_dirty = self.get_isDirty()
        make_dirty = 0
        try:
            if (self.encoding.python_encoding_name == encoding.python_encoding_name and
                self.encoding.use_byte_order_marker == encoding.use_byte_order_marker):
                # no change necessary
                return
    
            # Need to update the styleapplier here so we can set a new
            # encoding to the file.  This will replace fonts, etc.  as a
            # user changes the encoding.  We also have to handle
            # changing back and forth between utf-8 and single byte
            # encodings
            if self._views:
                self._views[0].scimoz.undoCollection = 0
            try:
                if self.encoding.python_encoding_name != encoding.python_encoding_name:
                    buffer = self.get_buffer()
                    if buffer: # else no need to change anything.
                        unicodeBuffer = None
                        encodedString = None
                        try:
                            # 'buffer' is a unicode object.  If we
                            # started out with UTF-8, then we merely
                            # want to see if the characters are encodable
                            # to the new encoding, but we don't want
                            # to change the actual bytes in our buffer
                            #
                            # otherwise, if the new encoding is not UTF-8
                            # then we are going from one 8bit to another
                            # 8bit encoding.  We must encode to our current
                            # encoding, then convert back to unicode with
                            # the new encoding.
                            #
                            # Converting between UTF-8 and UTF-16 does not
                            # require any changes in here at all, and
                            # everything should succeed.
                            if self.encoding.python_encoding_name.startswith('utf'):
                                self.encodingServices.encode(
                                    buffer, encoding.python_encoding_name,
                                    errors)
                            elif not encoding.python_encoding_name.startswith('utf'):
                                encodedString = self.encodingServices.encode(
                                    buffer, self.encoding.python_encoding_name,
                                    errors)
                        except (UnicodeError, COMException), ex:
                            errno = 1  # arbitrary non-zero value, see docstring
                            errmsg = ("Unable to convert unicode buffer to '%s'.  "
                                      "This encoding cannot represent all "
                                      "characters in the current buffer."
                                      % self.encoding.python_encoding_name)
                            self.lastErrorSvc.setLastError(errno, errmsg)
                            lastErrorSet = 1
                            raise ServerException(nsError.NS_ERROR_FAILURE,
                                                  errmsg)
                        try:
                            # Get the unicode object from the raw string
                            # (if errors was not "strict" then the buffer
                            # may have changed).
                            if encodedString:
                                unicodeBuffer = self.encodingServices.unicode(
                                    encodedString, encoding.python_encoding_name,
                                    errors)
                        except (UnicodeError, COMException), ex:
                            errno = 1  # arbitrary non-zero value, see docstring
                            errmsg = ("Unable to convert '%s' to '%s'.  "
                                      "This encoding cannot represent all "
                                      "characters in the current buffer."
                                      % (self.get_baseName(),
                                         encoding.python_encoding_name))
                            self.lastErrorSvc.setLastError(errno, errmsg)
                            lastErrorSet = 1
                            raise ServerException(nsError.NS_ERROR_FAILURE,
                                                  errmsg)
                        if unicodeBuffer:
                            self._set_buffer_encoded(unicodeBuffer,0)
                    if self.encoding.python_encoding_name.startswith('utf') or \
                        encoding.python_encoding_name.startswith('utf'):
                        make_dirty = 1
                    self.encoding = encoding

                self.prefs.setStringPref("encoding",
                                         self.encoding.python_encoding_name)
                #XXX _determineCodePage() should specify its
                #    return/exception behaviour. For now we assume any
                #    exception from it is an internal error.
                self._determineCodePage()
                self.encoding.use_byte_order_marker = encoding.use_byte_order_marker

                try:
                    self.observerService.notifyObservers(self,
                        'encoding_changed',
                        self.encoding.python_encoding_name)
                except COMException, ex:
                    pass # no one is listening
            finally:
                self.set_isDirty(was_dirty or make_dirty)
                if self._views:
                    self._views[0].scimoz.undoCollection = 1
        except:
            if not lastErrorSet: # then this is an internal error
                exc_info = sys.exc_info()
                errno = 0
                errmsg = "internal error setting encoding of '%s' to '%s': %s" % \
                         (self.get_baseName(),
                          encoding.python_encoding_name, exc_info[1])
                log.exception(errmsg)
                self.lastErrorSvc.setLastError(errno, errmsg)
            raise
    
    def encode(self, mode='strict'):
        decodedText = self.get_buffer().encode(self.encoding.python_encoding_name,mode)
        if self.encoding.use_byte_order_marker:
            decodedText = self.encoding.encoding_info.byte_order_marker + decodedText
        return decodedText, self.get_codePage()

    def removeUnencodeable(self):
        encoding_name = self._getStringPref('encoding')
        decodedText = self.get_buffer().encode(encoding_name,'replace')
        if self.encoding.use_byte_order_marker:
            encoding_info = self.encodingServices.get_encoding_info(encoding_name)\
                       .python_encoding_name
            self.encoding.use_byte_order_marker = encoding_info.byte_order_marker != ''
        self._set_buffer_encoded(unicode(decodedText,encoding_name))

    def get_encodedText(self):
        return self.encode()[0]

    def _determineCodePage(self):
        # XXX NEED TO FIX THE VARIABLE NAMES HERE!
        if self.encoding:
            encoding = self.encoding.python_encoding_name
        else:
            encoding_name = self._getStringPref('encoding', self.prefs.getStringPref('encodingDefault'))
            encoding = self.encodingServices.get_encoding_info(encoding_name)\
                           .python_encoding_name
        if sys.platform.startswith('win') or encoding.startswith('utf'):
            self.set_codePage(65001) # Unicode
        elif encoding.startswith('cp'):
            # likely will only happen on windows, so not particularly likely given the above if block
            try:
                self.set_codePage(int(encoding[2:]))
                return
            except ValueError, ex: # not an int
                log.error("can't convert encoding: %r to a codepage, returning 0", encoding)
                self.set_codePage(0)
                return
        else:
            log.info("Don't know what codepage to use for encoding %r, returning 0", encoding)
            self.set_codePage(0)

    _cleanLineRe = re.compile("(.*?)([ \t]+?)?(\r\n|\n|\r)", re.MULTILINE)
    def _clean(self, ensureFinalEOL, cleanLineEnds):
        """ This function cleans the text of scintilla according to the three prefs,
        any combination of which is valid.  It is careful to preserve the selection
        markers (currentPos and anchor)."""
        # XXX this function expects scimoz to be available
        if not self._views:
            return
        # in the case of any failure at all, we don't change the text
        try:
            self.docSettingsMgr.applyViewSettingsToDocument(self._views[0])
            ANCHOR = '<<Komodo:Anchor>>'
            CURRENTPOS = '<<Komodo:CurrentPos>>'
            text = self.get_buffer()
            scimoz = self._views[0].scimoz
            eolMode = scimoz.eOLMode
            fvl = scimoz.firstVisibleLine
            xoffset = scimoz.xOffset
            currentPos = scimoz.currentPos
            anchor = scimoz.anchor
            if cleanLineEnds:
                # We don't want to leave the anchor and currentposition
                # in the middle of end-of-line whitespace because that
                # whitespace will dissappear
                prevCurrentPos = scimoz.charPosAtPosition(scimoz.positionBefore(scimoz.currentPos))
                while prevCurrentPos and text[prevCurrentPos] in ' \t':
                    currentPos = prevCurrentPos
                    prevCurrentPos = scimoz.charPosAtPosition(scimoz.positionBefore(prevCurrentPos))
                prevAnchor = scimoz.charPosAtPosition(scimoz.positionBefore(scimoz.anchor))
                while prevAnchor and text[prevAnchor] in ' \t':
                    anchor = prevAnchor
                    prevAnchor = scimoz.charPosAtPosition(scimoz.positionBefore(prevAnchor))
            if currentPos < anchor:
                text = text[:anchor] + ANCHOR + text[anchor:]
                text = text[:currentPos] + CURRENTPOS + text[currentPos:]
            else:
                text = text[:currentPos] + CURRENTPOS + text[currentPos:]
                text = text[:anchor] + ANCHOR + text[anchor:]
            #XXX Should use eollib.scimozEOL2eol (or whatever it is
            #    called) instead of 'eol' here.
            eol = {components.interfaces.ISciMoz.SC_EOL_CR: '\r',
                   components.interfaces.ISciMoz.SC_EOL_CRLF: '\r\n',
                   components.interfaces.ISciMoz.SC_EOL_LF: '\n'}[eolMode]
            if cleanLineEnds:
                text = self._cleanLineRe.sub(r"\1"+eol, text+eol)[:-len(eol)]
            if ensureFinalEOL:
                if not text.endswith(eol):
                    text += eol
            if currentPos < anchor:
                # do currentPos replacement first, to not screw up
                # position shifts do to the replacement of the markers.
                newCurrentPos = text.index(CURRENTPOS)
                text = text[:newCurrentPos] + text[newCurrentPos + len(CURRENTPOS):]
                newAnchor = text.index(ANCHOR)
                text = text[:newAnchor] + text[newAnchor + len(ANCHOR):]
            else:
                newAnchor = text.index(ANCHOR)
                text = text[:newAnchor] + text[newAnchor + len(ANCHOR):]
                newCurrentPos = text.index(CURRENTPOS)
                text = text[:newCurrentPos] + text[newCurrentPos + len(CURRENTPOS):]
            self._set_buffer_encoded(text)
            self.docSettingsMgr.applyDocumentSettingsToView(self._views[0])
            self._views[0].scimoz.anchor = newAnchor
            self._views[0].scimoz.currentPos = newCurrentPos
            scimoz.lineScroll(0, min(fvl-scimoz.firstVisibleLine,
                                     scimoz.lineCount-scimoz.firstVisibleLine))
            scimoz.xOffset = xoffset
        except Exception, e:
            #XXX This is poor error handling.
            log.exception(e)
        
    def save(self, forceSave):
        mode = None
        if self.file.exists and not self.file.isWriteable:
            if not forceSave:
                errmsg = "File is not writeable"
                self.lastErrorSvc.setLastError(0, errmsg)
                raise ServerException(nsError.NS_ERROR_FAILURE, errmsg)
                
            mode = self.file.permissions
            # if we're not the owner, we shouldn't be able to write anyway
            desired_mode = mode | stat.S_IWUSR
            try:
                os.chmod(self.file.path, desired_mode)
            except EnvironmentError, ex:
                errmsg = "Unable to set the file mode to writeable: %s" % ex
                self.lastErrorSvc.setLastError(0, errmsg)
                raise ServerException(nsError.NS_ERROR_FAILURE, errmsg)
                
            # .hasChanged has the side-effect of re-stat'ing, which is
            # required for the subsequent check. Psychotically, this is
            # the only way to initiate a re-stat.
            self.file.hasChanged 
            if not self.file.isWriteable:
                # reset the mode
                try:
                    os.chmod(self.file.path, mode)
                except EnvironmentError:
                    pass # skip because we are already in an error mode
                
                errmsg = "Unable to set the file mode to writeable: chmod "\
                         "succeeded but a subsequent stat says the file "\
                         "is still not writeable"
                self.lastErrorSvc.setLastError(0, errmsg)
                raise ServerException(nsError.NS_ERROR_FAILURE, errmsg)
                
        try:
            ensureFinalEOL = self._globalPrefs.getBooleanPref("ensureFinalEOL")
            cleanLineEnds = self._globalPrefs.getBooleanPref("cleanLineEnds")
            if ensureFinalEOL or cleanLineEnds:
                self._clean(ensureFinalEOL, cleanLineEnds)
    
            # translate the buffer before opening the file so if it
            # fails, we haven't truncated the file
            data = self.encode()[0]

            try:
                self.file.open('wb+')
                try:
                    self.file.write(data)
                finally:
                    self.file.close()
            except COMException, ex:
                # .open(), .write(), and .close() will setLastError on
                # failure so don't set it again. You will just override
                # better data.
                raise ServerException(nsError.NS_ERROR_FAILURE, str(ex))

            self._lastmd5 = md5.new(data).digest()
            self.set_isDirty(0)
            self.setSavePoint()

            try:
                self._obsSvc.notifyObservers(self, "file_changed",
                                             self.file.URI)
            except:
                pass # ignore, noone listening
        finally:
            # fix file mode
            if forceSave and mode:
                try:
                    os.chmod(self.file.path, mode)
                except:
                    errmsg = "Unable to reset the file mode."
                    self.lastErrorSvc.setLastError(0, errmsg)
                    raise ServerException(nsError.NS_ERROR_FAILURE, errmsg)
            self.removeAutoSaveFile()

    def revert(self):
        self._loadfile(self.file)
        self.setSavePoint()
        self.removeAutoSaveFile()

    def setSavePoint(self):
        if self._views:
            # All views share the same doc pointer.
            self._views[0].scimoz.setSavePoint()

    def saveState(self, scintilla):
        self.docSettingsMgr.applyViewSettingsToDocument(scintilla)

    # The document manages a reference count of the views onto that
    # document. When the last view 
    def addView(self, scintilla):
        timeline.enter('koDocumentBase.AddView')
        self._views.append(scintilla)
        scimoz = scintilla.scimoz
        xpself = WrapObject(self, components.interfaces.koIDocument)
        if not self._document:
            scimoz.docPointer = 0 # create document
            self.docSettingsMgr.register(xpself, scintilla)
            self._buffer = None # clear out any old buffer we may have had
            self._document = scimoz.docPointer
            scimoz.addRefDocument(self._document)
        else:
            scimoz.addRefDocument(self._document)
            scimoz.docPointer = self._document
            self.docSettingsMgr.register(xpself, scintilla)
        scimoz.codePage = self._codePage
        log.info("in AddView")
        timeline.leave('koDocumentBase.AddView')
    
    def releaseView(self, scintilla):
        #print "Releasing View"
        if scintilla not in self._views:
            raise ServerException(nsError.NS_ERROR_FAILURE,'SciMoz does not reference.')
        self.docSettingsMgr.unregister(scintilla)
        scimoz = scintilla.scimoz
        scimoz.releaseDocument(self._document);
        if len(self._views) == 1:
            buffer = self.get_buffer()
            self._document = None
            self._set_buffer_encoded(buffer, 0)

        self._views.remove(scintilla)
        #if not self._views:
        #    self.docSettingsMgr = None

    def get_numScintillas(self):
        return len(self._views)

    # we want to watch for changes in the prefs we derive from
    def observe(self, subject, topic, data):
        if topic == 'useTabs':
            self._useTabs = self.prefs.getBooleanPref('useTabs')
        elif topic == 'indentWidth':
            val = self._indentWidth = self.prefs.getLongPref('indentWidth')
            for view in self._views:
                view.scimoz.indent = val
        elif topic == 'tabWidth':
            val = self._tabWidth = self.prefs.getLongPref('tabWidth')
            for view in self._views:
                view.scimoz.tabWidth = val

    #attribute string URI; 
    def get_useTabs(self):
        if self._useTabs is None:
            if self.prefs.hasPrefHere('useTabs'):
                self._useTabs = self.prefs.getBooleanPref('useTabs')
            else:
                self._guessFileIndentation()
        return self._useTabs
  
    def set_useTabs(self, value):
        self.prefs.setBooleanPref('useTabs', value) # will affect _useTabs through prefs observers

    useTabs = property(get_useTabs, set_useTabs)

    def get_indentWidth(self):
        
        timeline.enter('koDocumentBase.get_indentWidth')
        try:
            if self._indentWidth is None:
                log.info("_indentWidth is None")
                if self.prefs.hasPrefHere('indentWidth'):
                    self._indentWidth = self.prefs.getLongPref('indentWidth')
                    log.info('got _indentWidth from prefs: %r' % self._indentWidth)
                else:
                    self._guessIndentWidth()
                    return self._indentWidth
            else:
                log.info("_indentWidth is not none, it's %s" % self._indentWidth)
        finally:
            timeline.leave('koDocumentBase.get_indentWidth')
        return self._indentWidth

    def set_indentWidth(self):
        self.prefs.setLongPref('indentWidth', value) # will affect _useTabs through prefs observers

    indentWidth = property(get_indentWidth, set_indentWidth)

    def get_tabWidth(self):
        if self._tabWidth is None:
            self._tabWidth = self.prefs.getLongPref('tabWidth')
        return self._tabWidth

    def set_tabWidth(self):
        self.prefs.setLongPref('tabWidth', value) # will affect _useTabs through prefs observers

    tabWidth = property(get_tabWidth, set_tabWidth)

    def _guessFileIndentation(self):
        timeline.enter('koDocumentBase._guessFileIndentation')
        # Heuristic to determine what file indentation settings the user
        # likely wants for this file.
        log.info("in _guessFileIndentation")
        useTabs = usesSpaces = linesChecked = 0
        buffer = self.get_buffer()
        if self._indentWidth is None:
            self._guessIndentWidth()
            log.warn("guessed indentWidth to be: %s" % self._indentWidth)

        # In the first 150 lines of the file, search for the non-blank
        # lines with leading white-space.  Searching farther takes too long.
        for line in buffer.splitlines()[:150]:
            if line[:1] == "\t":
                # If first char is a tab, recognize that and move on
                linesChecked += 1
                useTabs = 1
            elif line[:2] == "  ":
                # If first 2 chars are spaces, recognize that and move on
                # Require at least two spaces on the line to count it
                linesChecked += 1
                usesSpaces = 1
            if linesChecked == 25:
                # Only check up to 25 lines with indentation
                break

        if linesChecked:
            log.warn("guessed useTabs = %d and usesSpaces + %d",
                     useTabs, usesSpaces)
            # We found some lines with indentation
            if useTabs and usesSpaces:
                # If we found both space and tab indentation, leave the
                # indentWidth setting as default, but do use tabs where
                # appropriate
                self._useTabs = 1
                for v in self._views:
                    v.scimoz.indent = self._indentWidth
                    v.scimoz.useTabs = 1
            elif useTabs:
                # If only tab indentation was found, set the indentWidth
                # to the tabWidth, so we essentially always use tabs.
                self._useTabs = 1
                self._indentWidth = self._tabWidth
                for v in self._views:
                    v.scimoz.indent = self._indentWidth
                    v.scimoz.useTabs = 1
            else:
                # Otherwise we won't use tabs at all.
                self._useTabs = 0
                for v in self._views:
                    v.scimoz.useTabs = 0
        else:
            # Lacking better information, fallback to the pref values.
            if self._useTabs is None:
                self._useTabs = self.prefs.getBooleanPref("useTabs")
            if self._indentWidth is None:
                self._indentWidth = self.prefs.getLongPref("indentWidth")
            if self._tabWidth is None:
                self._tabWidth = self.prefs.getLongPref("tabWidth")
            for v in self._views:
                #XXX This is being a bit paranoid here. I am not sure if
                #    this method need worry about writing self._tabWith
                #    through to scimoz. David? --TM
                v.scimoz.useTabs = self._useTabs
                v.scimoz.indent = self._indentWidth
                v.scimoz.tabWidth = self._tabWidth
        timeline.leave('koDocumentBase._guessFileIndentation')

    # Guess indent-width from text content. (Taken from IDLE.)
    #
    # This should not be believed unless it's in a reasonable range
    # (e.g., it will be 0 if no indented blocks are found).
    def _guessIndentWidth(self):
        text = self.get_buffer()
        if text == '':
            indentWidth = self.prefs.getLongPref('indentWidth')
            self._indentWidth= indentWidth
            return
        # if we don't have a view yet, we can't do anything.
        if not self._views:
            log.error("Was asked to guess indent width before there's a view")
            self._indentWidth = self.prefs.getLongPref('indentWidth')
            return
        if not self._languageObj:
            self.get_languageObj()
        # The strategy for guessing the indentation is delegated to the
        # lexer language service, since different languages have very
        # different rules.
        indentWidth, useTabs = self._languageObj.guessIndentation(self._views[0].scimoz,
                                                                  self.get_tabWidth())
        if indentWidth == 0:  # still haven't found anything, so go with the prefs.
            indentWidth = self.prefs.getLongPref('indentWidth')
            useTabs = self.prefs.getBooleanPref('useTabs')

        log.info("_guessIndentWidth: indentWidth=%d, useTabs=%d",
                 indentWidth, useTabs)
        self._indentWidth = indentWidth
        self._useTabs = useTabs

    def _statusBarMessage(self, message):
        sm = components.classes["@activestate.com/koStatusMessage;1"]\
             .createInstance(components.interfaces.koIStatusMessage)
        sm.category = "Document"
        sm.msg = message
        sm.timeout = 5000 # 0 for no timeout, else a number of milliseconds
        sm.highlight = 1  # boolean, whether or not to highlight
        try:
            self._obsSvcProxy.notifyObservers(sm, 'status_message', None)
        except COMException, e:
            # do nothing: Notify sometimes raises an exception if (???)
            # receivers are not registered?
            pass
        
    def getUnsavedChanges(self):
        import difflib
        eolStr = eollib.eol2eolStr[self._eol]
        inmemory = self.get_buffer().splitlines(1)
        
        self.file.open('rb')
        try:
            ondisk = self.file.read(-1)
            (ondisk, encoding, bom) = self._detectEncoding(ondisk)
            ondisk = ondisk.splitlines(1)
        finally:
            self.file.close()

        difflines = difflib.ndiff(ondisk, inmemory)
        # difflib adds \n to its new lines, regardless of what was sent in
        fixed_difflines = []
        for line in difflines:
            if line.startswith('? '):
                fixed_difflines.append(line[:-1] + eolStr)
            else:
                fixed_difflines.append(line)
        return ''.join(fixed_difflines)

    def getAutoSaveFileName(self):
        koDirs = components.classes["@activestate.com/koDirs;1"].\
                 getService(components.interfaces.koIDirs)
        dname = os.path.join(koDirs.userDataDir, "autosave")
        if not os.path.exists(dname):
            os.mkdir(dname)
        # retain part of the readable name
        autoSaveFilename = "%s-%s" % (self.file.md5name,self.file.baseName)
        return os.path.join(dname, autoSaveFilename)
        
    def getAutoSaveFile(self):
        autoSaveFile = components.classes["@activestate.com/koFileEx;1"] \
                      .createInstance(components.interfaces.koIFileEx)
        autoSaveFile.URI = self.getAutoSaveFileName()
        return autoSaveFile

    def removeAutoSaveFile(self):
        if self.isUntitled: return
        autoSaveFile = self.getAutoSaveFile()
        if not autoSaveFile.exists:
            return
        os.remove(autoSaveFile.path)
        
    def haveAutoSave(self):
        if self.isUntitled: return 0
        # set the autosave file name
        autoSaveFile = self.getAutoSaveFile()
        
        if autoSaveFile.exists:
            if autoSaveFile.lastModifiedTime > self.file.lastModifiedTime:
                #print "tmpfile is newer than original"
                return 1;
            else:
                # we have an autosave file, but it's old, delete it
                #print "removing old temporary auto save file ",self._autoSaveFileName
                os.remove(autoSaveFile.path)
        return 0
        
    def doAutoSave(self):
        timeline.enter('koDocumentBase.doAutoSave')
        try:
            # no point in autosaving if we're not dirty
            if not self._isDirty or self.isUntitled: return
            
            savetime = 0
            if self._globalPrefs.hasPref("autoSaveMinutes"):
                savetime = self._globalPrefs.getLongPref("autoSaveMinutes") * 60
            if not savetime: return
    
            autoSaveFile = self.getAutoSaveFile()
            # if we've saved recently, just return
            if autoSaveFile.exists and \
                autoSaveFile.lastModifiedTime > time.time() + savetime:
                    return
            
            # if we're the current view, we cannot get the text from
            # scintilla in a thread, and must proxy.  since this shouldn't
            # get hit often, we'll just always proxy to be on the safe
            # side of things
            
            # translate the buffer before opening the file so if it
            # fails, we haven't truncated the file
            wrapSelf = xpcom.server.WrapObject(self, components.interfaces.koIDocument)
            doc = _xpcom.getProxyForObject(1, components.interfaces.koIDocument,
                                              wrapSelf, _xpcom.PROXY_SYNC | _xpcom.PROXY_ALWAYS)
            try:
                data = doc.encodedText
            except:
                self._statusBarMessage("Error getting encoded text for autosave of '%s'" %
                              doc.baseName)
                return
   
            try:
                autoSaveFile.open('wb+')
                try:
                    autoSaveFile.write(data)
                finally:
                    autoSaveFile.close()
            except:
                # .open(), .write(), and .close() will setLastError on
                # failure so don't set it again. You will just override
                # better data.
                self._statusBarMessage("Unable to autosave file '%s'" %
                              doc.baseName)
                return
        finally:
            timeline.leave('koDocumentBase.doAutoSave')
    
    def restoreAutoSave(self):
        if self.isUntitled: return
        timeline.enter('koDocumentBase.restoreAutoSave')
        try:
            autoSaveFile = self.getAutoSaveFile()
            self._loadfile(autoSaveFile)
            self.set_isDirty(1)
            # fix the file content md5
            data = self._get_buffer_from_file(self.file)
            self._lastmd5 = md5.new(data).digest()
        finally:
            timeline.leave('koDocumentBase.restoreAutoSave')
