#!python
# Copyright (c) 2000-2001 ActiveState Corporation.
# See the file LICENSE.txt for licensing information.
#
# Shane Caraveo
#
# Provide file status information
# runs background thread to periodicaly update status on a file
# has plugin capability to add new types of status retreivers (CVS, Perforce)
#
# the status service gets all file information from the
# file service, and updates it with scc information
#
# observer notifications:
#   file_status (url)
#   file_status_updated (None)
#
# when a file status changes, the service sends a notification file_status_update.
# any observer that listens for this will receive the url of the file that has
# been updated.  It must then call back to the service to get the status
# array.
#

from xpcom import components, nsError, ServerException, COMException
from xpcom._xpcom import PROXY_SYNC, PROXY_ALWAYS, PROXY_ASYNC
from xpcom.server import WrapObject, UnwrapObject

import logging
import timeline
import uriparse
import sys, os, time, stat
import threading, copy

log = logging.getLogger('koFileStatusService')

def sortFileStatus(a, b):
    if a.isDirectory and not b.isDirectory:
        return 1
    if b.isDirectory and not a.isDirectory:
        return -1
    return cmp(a.URI,b.URI)

class KoFileStatusService:
    _com_interfaces_ = [components.interfaces.koIFileStatusService, components.interfaces.nsIObserver]
    _reg_clsid_ = "{20732408-43DA-4ca2-BC9F-B82437A3CB2B}"
    _reg_contractid_ = "@activestate.com/koFileStatusService;1"
    _reg_desc_ = "Komodo File Status Service"

    def __init__(self):
        timeline.enter('koFileStatusService.__init__')
        self._proxyMgr = components.classes["@mozilla.org/xpcomproxy;1"].\
            getService(components.interfaces.nsIProxyObjectManager)
        self._observerSvc = components.classes["@mozilla.org/observer-service;1"].\
            getService(components.interfaces.nsIObserverService)
        self._observerProxy = self._proxyMgr.getProxyForObject(None,
            components.interfaces.nsIObserverService, self._observerSvc,
            PROXY_ALWAYS | PROXY_ASYNC)
        self._globalPrefs = components.classes["@activestate.com/koPrefService;1"].\
            getService(components.interfaces.koIPrefService).prefs
        self._fileSvc = \
            components.classes["@activestate.com/koFileService;1"] \
            .getService(components.interfaces.koIFileService)
        self._fileSvcProxy = self._proxyMgr.getProxyForObject(None,
            components.interfaces.koIFileService, self._fileSvc,
            PROXY_ALWAYS | PROXY_ASYNC)
        
        # set up data cache
        self._urllist = {}
        self._statusCheckers = []
        self.shutdown = 0
        self._continue = 0
        self._background = 0
        self._diskChecker = None
        
        # set up the background thread
        self._thread = None
        self._tlock = threading.Lock()
        self._cv = threading.Condition()
        self._shutdown = threading.Condition()
        timeline.leave('koFileStatusService.__init__')

    def init(self):
        timeline.enter('koFileStatusService.init')

        self._observerSvc.addObserver(self, "file_changed",0)
        self._observerSvc.addObserver(self, "file_status_now",0)
        self._observerSvc.addObserver(self, "file_update_now",0)
        
        self._fileSvc.observerService.addObserver(self,'',0)
        
        self._diskChecker = KoDiskFileChecker()
        self._statusCheckers.append(self._diskChecker)

        self._statusCheckers.append(KoCVSFileChecker())
        self._statusCheckers.append(KoP4FileChecker())

        
        for checker in self._statusCheckers:
            self._globalPrefs.prefObserverService.addObserver(self,checker.enabledPrefName,0)
            self._globalPrefs.prefObserverService.addObserver(self,checker.backgroundPrefName,0)

        self._tlock.acquire()
        self._thread = threading.Thread(target = KoFileStatusService._process,
                                     args = (self,))
        self._thread.start()
        self._tlock.release()
        timeline.leave('koFileStatusService.init')

    def _removeObserver(self, topic):
        try:
            self._observerSvc.removeObserver(self, topic)
        except:
            log.error("Unable to remove observer %s"%topic)

    def unload(self):
        self.shutdown = 1;
        if self._thread.isAlive():
            self._cv.acquire()
            self._cv.notify()
            self._cv.release()
            self._shutdown.acquire()
            self._shutdown.wait(5)
            self._shutdown.release()

        for checker in self._statusCheckers:
            try:
                self._globalPrefs.prefObserverService.removeObserver(self,checker.enabledPrefName)
                self._globalPrefs.prefObserverService.removeObserver(self,checker.backgroundPrefName)
            except:
                # prefs shutdown already?
                log.error("Unable to remove prefs observers")
    
        self._removeObserver("file_changed")
        self._removeObserver("file_status_now")
        self._removeObserver("file_update_now")

        try:
            self._fileSvc.observerService.removeObserver(self,'')
        except:
            # file service shutdown before us?
            log.error("Unable to remove file service observers")
    
    def notifyError(self, error, detail):
        # XXX 'scc_status_message' is evil (see bug 20568). In addition, it
        #     is poorly named because the code is structured such that the
        #     non-scc checkers here could set an error.
        if error:
            #print "attempting to send notification"
            try:
                self._observerProxy.notifyObservers(error, 'scc_status_message', detail)
            except COMException, e:
                pass
        
    def _getFileStatusNow(self, URI):
        url = self._fileSvc.getFileFromURI(URI)
        self._getFileStatusNowWithFile(url)
    
    def _getFileStatusNowWithFile(self, url):
        url = UnwrapObject(url)
        lasterror = None
        # force an update to the status
        urllist = self._fileSvc.getFilesInBaseURI(url.URI)
        urllist = [UnwrapObject(_url) for _url in urllist]
        for checker in self._statusCheckers:
            if not checker.isActive():
                continue
            if checker.executablePrefName:
                checker.setExecutable(self._globalPrefs.getStringPref(checker.executablePrefName))

            if checker.scc and (not url.isLocal or \
                (url.sccDirType and \
                ((url.sccType != checker.scc) or \
                (url.sccDirType != checker.scc)))):
                continue

            # invalidate all the url's if it's a directory
            checker.invalidateUrl(url,1)
                
            # only check for the primary url.  if it is a directory
            # getFileStatus will get status for all files in that
            # directory, but not recursivly
            checker.getFileStatus(url, not url.isDirectory)
            if checker.error and (not lasterror or checker.error.msg != lasterror.msg):
                lasterror = checker.error
                self.notifyError(checker.error,checker.error_detail)
                checker.error = None
            else:
                # reset the status indicators
                # we don't do error checking here 
                for _url in urllist:
                    checker.getFileStatus(_url, not _url.isDirectory)
        
        # send notifications of status updates
        for _url in urllist:
            try:
                self._observerSvc.notifyObservers(_url, 'file_status', _url.URI)
            except COMException, e:
                if e.errno == nsError.NS_ERROR_FAILURE:
                    # if noone is listening, we get an exception
                    pass
                else:
                    #print "caught unknown exception in notify"
                    raise            
        try:
            self._observerSvc.notifyObservers(self, 'file_status_updated',None)
        except COMException, e:
            if e.errno == nsError.NS_ERROR_FAILURE:
                # if noone is listening, we get an exception
                pass
            else:
                #print "caught unknown exception in notify"
                raise

    def getStatus(self, file):
        self._getFileStatusNowWithFile(file)
        return 1
        
    def isRepository(self, URI):
        url = self._fileSvc.getFileFromURI(URI)
        if not url or not url.isLocal or not url.exists:
            return ''
        
        if url.sccType:
            return url.sccType
        if url.sccDirType:
            return url.sccDirType
            
        for checker in self._statusCheckers:
            if not checker.isActive() or not checker.scc:
                continue
            if checker.isRepository(url.path):
                return checker.scc

        return ''
    
    # nsIObserver interface
    def observe(self, subject, topic, data):
        log.debug("status observer received %s:%s"%(topic,data))

        if topic == self._globalPrefs.id:
            # url is actually the pref name that was changed
            for checker in self._statusCheckers:
                if data == checker.enabledPrefName or \
                   data == checker.backgroundPrefName:
                    enabled = self._globalPrefs.getBooleanPref(checker.enabledPrefName)
                    background = self._globalPrefs.getBooleanPref(checker.backgroundPrefName)
                    if enabled != checker.enabled or background != checker.background:
                        checker.enabled = enabled
                        checker.background = background
                        self._cv.acquire()
                        self._cv.notify()
                        self._cv.release()
                        break
            # we don't continue in this observer for pref changes
            return
        
        elif topic == "file_status_now":
            #_url.update = 0
            self._getFileStatusNow(data)
            return
        elif topic == "file_changed":
            _url = self._fileSvc.getFileFromURI(data)
            _url = UnwrapObject(_url)
            _url.update = 0
            _url.dofileupdate = 1

        self._cv.acquire()
        # notify is an immediate release, but we want the thread
        # to continue until there is no more activity, so we flag
        # it to do so
        self._continue = 1
        self._cv.notify()
        self._cv.release()

    def _process(self):
        #print "starting file status background thread"
        urllist = []
        lasterror = None
        while 1:
            try:
                updated = []
                #print "file status waiting to do update..."
                
                # if we have more url's than we checked last
                # time, lets do them now so the user doesn't
                # have to wait 30 seconds
                if not self._continue:
                    self._cv.acquire()
                    self._cv.wait(30)
                    # if self._continue is set, this is not a timeout
                    self._background = not self._continue
                    self._cv.release()
                # reset flag, see Observer for more info
                self._continue = 0
                
                if self.shutdown:
                    self._shutdown.acquire()
                    self._shutdown.notify()
                    self._shutdown.release()
                    #print "status thread shutting down"
                    return
                
                
                #print "doing file status update"
                urllist = self._fileSvc.getAllFiles()
                urllist = [UnwrapObject(_url) for _url in urllist]
                
                start = time.time()
                
                # invalidate old status information
                # we do this seperately so that the scc modules can
                # catch full directories of urls
                for checker in self._statusCheckers:
                    
                    # do we need to run this checker?
                    if not checker.isActive():
                        #log.debug("Skipping %s"%checker.scc)
                        continue
                    if checker.executablePrefName:
                        checker.setExecutable(self._globalPrefs.getStringPref(checker.executablePrefName))
                    checker.background = self._globalPrefs.getBooleanPref(checker.backgroundPrefName)
                    checker.recursive = checker.recursivePrefName and self._globalPrefs.getBooleanPref(checker.recursivePrefName)
                    checker.duration = self._globalPrefs.getLongPref(checker.durationPrefName)
                    if self._background and not checker.background:
                        #log.debug("Skipping %s"%checker.scc)
                        continue
                    
                    urls = []
                    # invalidate paths first so they can be recached
                    # build a list of url's that we can actually update
                    for url in urllist:
                        if url.isLocal and (url.update == 0 or \
                           (checker.background and checker.invalidatePath(url))):
                            # if this file is marked to have scc, and it's scc does
                            # not match the checker, then skip passing the file to the checker
                            if not checker.scc or (checker.scc and \
                                ((url.sccType and (url.sccType == checker.scc)) or \
                                (not url.sccDirType or (url.sccDirType == checker.scc)))):
                                urls.append(url)
                            #else:
                            #    log.debug('skipping %s'%url.URI)

                    urls.sort(sortFileStatus)
                    for url in urls:
                        # if we're in process, but a shutdown is requested, do it
                        if self.shutdown:
                            self._shutdown.acquire()
                            self._shutdown.notify()
                            self._shutdown.release()
                            log.debug("status thread shutting down")
                            return
                        
                        if not checker.enabled:
                            #log.debug("Skipping %s"%checker.scc)
                            break
                        
                        # if we're in a big long list (large project)
                        # then it's possible for the url to have been
                        # removed before we get here
                        #if url.remove: continue
                        
                        # check the status of the url
                        done = 0
                        #print "url time is %d, now is %d" %(self._urllist[url].update, time.time())
                        
                        try:
                            #log.debug("%s checking %s %s" %(checker.scc,url.sccType,url.URI))
                            oldscc = copy.deepcopy(url.scc)
                            oldstat = copy.deepcopy(url.stats)
                            
                            if url.dofileupdate:
                                checker.invalidateUrl(url,force=1)
                                
                            if not checker.getFileStatus(url,url.dofileupdate):
                                if checker.error and (not lasterror or checker.error.msg != lasterror.msg):
                                    lasterror = checker.error
                                    self.notifyError(checker.error,checker.error_detail)
                                    checker.error = None
                                #log.debug("Check Failed %s %s"%(checker.scc,checker.error))
                                continue
                            # this avoids making the notification below if the status has
                            # not changed.  The notification interupts and updates the UI
                            # taking an enormous amount of cpu usage
                            if oldscc == url.scc and \
                                oldstat == url.stats: continue
                            updated.append(url)
    
                            try:
                                self._observerProxy.notifyObservers(url, 'file_status',url.URI)
                            except COMException, e:
                                if e.errno == nsError.NS_ERROR_FAILURE:
                                    # noone is listening, we get an exception
                                    pass
                        except Exception, e:
                            log.error("Unknown Exception %s" % repr(e))
                        
                    # on every checker we notify, since a particular checker
                    # can take some time
                    if updated:
                        try:
                            # this call is proxied to the UI thread.  The trees are invalidated
                            # in the observer for this call
                            self._observerProxy.notifyObservers(self, 'file_status_updated',None)
                        except COMException, e:
                            if e.errno == nsError.NS_ERROR_FAILURE:
                                # noone is listening, we get an exception
                                pass
                
                lasterror = None
                
                for url in urls:
                    url.dofileupdate = 0
                        
                end =  time.time()
                for url in updated:
                    url.update = end

            except:
                # we catch any exception so we can clear our thread variable
                # this allows us to restart the thread again later
                # we re raise the exception
                import traceback
                errmsg = ''.join(traceback.format_exception(*sys.exc_info()))
                log.debug('KoFileStatusService thread exception %s' % errmsg)

        log.debug("Status Thread ending")

class KoFileCheckerBase:
    def __init__(self):
        self.type = None
        self.scc = None
        self.duration = 15
        self.durationPrefName = None
        self.enabledPrefName = None
        self.enabled = 0
        self.backgroundPrefName = None
        self.background = 0
        self.recursivePrefName = None
        self.recursive = 0
        self.error = None
        self.error_detail = None
        self.executable = None
        self.executablePrefName = None
        self._globalPrefs = components.classes["@activestate.com/koPrefService;1"].\
            getService(components.interfaces.koIPrefService).prefs
    
    def setExecutable(self, executable):
        self.executable = executable
        
    def setError(self, msg='', detail=''):
        self.error = None
        if not msg: return
        
        self.error = components.classes["@activestate.com/koStatusMessage;1"]\
                .createInstance(components.interfaces.koIStatusMessage)
        self.error.category = "scc_error_message"
        self.error.msg = msg
        self.error.timeout = 5000
        self.error.highlight = 1
        self.error_detail = str(detail)
    
    def isRepository(self, path):
        return 0
    
    def isActive(self):
        self.enabled = self._globalPrefs.getBooleanPref(self.enabledPrefName)
        return self.enabled

    def invalidatePath(self, url,force=0):
        return 0

    def invalidateUrl(self, url,force=0):
        return 0

    def getFileStatus(self, url, dofile=1):
        return None
    
    
class KoDiskFileChecker(KoFileCheckerBase):
    def __init__(self):
        timeline.enter('KoDiskFileChecker.__init__')
        KoFileCheckerBase.__init__(self)
        self.type = 'disk'
        self.durationPrefName = 'diskBackgroundMinutes'
        self.enabledPrefName = 'diskStatusEnabled'
        self.backgroundPrefName = 'diskBackgroundCheck'
        timeline.leave('KoDiskFileChecker.__init__')

    def invalidatePath(self, url,force=0):
        return force or url.update < time.time() - (self.duration * 60)

    def invalidateUrl(self, url,force=0):
        return force or url.update < time.time() - (self.duration * 60)

    def getFileStatus(self, url, dofile=1):
        self.setError()
        changed = url.hasChanged
        # we mark dofileupdate to force
        # an scc update to occure
        if changed:
            url.dofileupdate = 1
        return changed


import p4lib
class KoP4FileChecker(KoFileCheckerBase):
    def __init__(self):
        timeline.enter('KoP4FileChecker.__init__')
        KoFileCheckerBase.__init__(self)
        self._have = {}
        self._open = {}
        self._files = {}
        self._not = {}
        #XXX rename _paths to _directories
        self._paths = {}
        self.p4 = p4lib.P4()
        self.type = 'scc'
        self.scc = 'p4'
        self.durationPrefName = 'p4BackgroundMinutes'
        self.enabledPrefName = 'p4Enabled'
        self.backgroundPrefName = 'p4BackgroundCheck'
        self.recursivePrefName = 'p4Recursive'
        self._p4Svc = components.classes["@activestate.com/koP4;1"].\
            getService(components.interfaces.koIP4)
        self.setExecutable(self._p4Svc.executable)
        self.executablePrefName = 'p4Executable'
        timeline.leave('KoP4FileChecker.__init__')

    def setExecutable(self, executable):
        self.executable = 'p4'
        if executable and os.path.exists(executable):
            self.executable = executable
        
    def notifyError(self, e):
        if str(e).find("is not under client's root") != -1:
            return
        log.debug(e)
        self.setError('P4 Status Error, see SCC Output pane for details',e)

    # we don't cache the where information into _path here, since that
    # would prevent getFileStatus from running correctly (ie retrieving
    # files/have/open data)
    def isRepository(self, path):
        npath = os.path.normcase(path)
        if npath in self._not: return 0
        if npath in self._paths: return 1
        try:
            self.p4.p4 = self.executable
            basename = path
            if not os.path.isdir(path):
                basename = os.path.dirname(path)
            if self.p4.where(path,cwd=basename): return 1
        except p4lib.P4LibError, e:
            self.notifyError(e)
            # XXX what is right? should we _not it?
            return 0
        self._not[npath] = time.time()
        return 0
    
    def isActive(self):
        self.enabled = self._globalPrefs.getBooleanPref(self.enabledPrefName)
        return self._p4Svc.isFunctional and self.enabled

    # returns true if the path is invalidated
    def invalidatePath(self, url, force=0):
        if not url.path:
            return 0
        if url.isDirectory:
            return self._invalidate(url.path, force)
        else:
            return self._invalidate(url.dirName, force)
    
    # returns true if the path is invalidated    
    def invalidateUrl(self, url, force=0):
        if not url.path:
            return 0
        return self._invalidate(url.path, force)
        
    def _invalidate(self, path, force):
        npath = os.path.normcase(path)
        if npath in self._paths:
            # this will force a refetch of the entire path
            # if the last stat on the path is older than
            # the duration
            lastcheck = self._paths[npath]
            if force or lastcheck < time.time() - self.duration *60:
                self._paths[npath] = 0
            else:
                return 0
        elif npath in self._not:
            lastcheck = self._not[npath]
            if force or lastcheck < time.time() - self.duration *60:
                self._not[npath] = 0
            else:
                return 0
        elif npath in self._have:
            # this will force a refetch of just the one file
            depotFile = self._have[npath]['depotFile']
            if self._files.has_key(depotFile):
                del self._files[depotFile]
            if self._open.has_key(depotFile):
                del self._open[depotFile]
            del self._have[npath]
        return 1
    
    def getFileStatus(self, url, dofile=0):
        # initialize all items.  This helps to clear the item properties
        # when it is removed from scc
        url.clearscc()
        _sccInfo = {}
        
        # we use nbasepath and npath for keys, we use
        # path and basepath for sending to P4 commands
        # this is because P4 is case sensitive
        basepath = path = url.path
        if not path:
            return None
        nbasepath = npath = os.path.normcase(path)
        #print "CHECKING p4 for ",basepath
        if not url.isDirectory:
            basepath = url.dirName
            nbasepath = os.path.normcase(basepath)

        try:
            if nbasepath in self._not:
                return None
            
            self.p4.p4 = self.executable
            
            if (nbasepath not in self._paths or \
                self._paths[nbasepath] == 0) and \
                os.path.isdir(basepath):
                # execute p4 on the path
                if self.recursive:
                    subpath = "/..."
                else:
                    subpath = "/*"
                    
                # empty string works if path is the base
                # directory of a client spec.
                where = self.p4.where('',cwd=basepath)
                if where:
                    # store the basepath to signal that this
                    # is a perforce directory
                    _sccInfo['sccDirType']=self.scc
                    _sccInfo['sccType']=self.scc
                else:
                    self._not[nbasepath]=time.time()
                    return None
                    
                #print "p4 checking ",basepath + subpath
                p4have = self.p4.have(basepath + subpath,
                                      cwd=basepath)
                for file in p4have:
                    filekey = os.path.normcase(file['localFile'])
                    path = os.path.normcase(os.path.dirname(filekey))
                    self._paths[path]=time.time()
                    file['ondisk'] = os.path.exists(filekey);
                    self._have[filekey] = file
                
                p4files = self.p4.files(basepath + subpath,cwd=basepath)
                for file in p4files:
                    self._files[file['depotFile']] = file

                p4files = self.p4.opened(basepath + subpath,cwd=basepath)
                for file in p4files:
                    self._open[file['depotFile']] = file
                    if not file['depotFile'] in self._files or \
                        self._files[file['depotFile']]['action'] == 'delete':
                        where = self.p4.where(file['depotFile'],cwd=basepath)
                        if len(where):
                            self._files[file['depotFile']] = file
                            filekey = os.path.normcase(where[0]['localFile'])
                            file['localFile'] = where[0]['localFile']
                            file['ondisk'] = os.path.exists(filekey);
                            self._have[filekey] = file
                
                self._paths[nbasepath]=time.time()

            if dofile and npath not in self._have and not os.path.isdir(npath):
                #print "checking p4 for ",path
                # file not in cache, get it now and stuff it into the cache
                p4have = self.p4.have(path,cwd=basepath)
                for file in p4have:
                    filekey = os.path.normcase(file['localFile'])
                    file['ondisk'] = os.path.exists(filekey);
                    self._have[filekey] = file
                
                p4files = self.p4.files(path,cwd=basepath)
                for file in p4files:
                    self._files[file['depotFile']] = file

                p4files = self.p4.opened(path,cwd=basepath)
                for file in p4files:
                    self._open[file['depotFile']] = file
                    if not file['depotFile'] in self._files or \
                        self._files[file['depotFile']]['action'] == 'delete':
                        where = self.p4.where(file['depotFile'],cwd=basepath)
                        if len(where):
                            self._files[file['depotFile']] = file
                            filekey = os.path.normcase(where[0]['localFile'])
                            file['localFile'] = where[0]['localFile']
                            file['ondisk'] = os.path.exists(filekey);
                            self._have[filekey] = file
                        
        
        except p4lib.P4LibError, e:
            self.notifyError(e)
            if nbasepath in self._paths:
                # this ensures that a path that IS in a repository, but contains
                # no files will be marked correctly
                _sccInfo['sccDirType']=self.scc
            return None

        if not self._have.has_key(npath):
            # still don't have it, not a perforce file
            if nbasepath in self._paths:
                # this ensures that a path that IS in a repository, but contains
                # no files will be marked correctly
                _sccInfo['sccDirType']=self.scc
            return None
        
        # get p4 info for the path, not the file
        
        have = self._have[npath]
        
        _sccInfo['sccType'] =  self.scc
        _sccInfo['sccDirType'] = self.scc
        _sccInfo['sccHaveOnDisk']= have['ondisk']
        _sccInfo['sccLocalRevision'] = str(have['rev'])
        if have['depotFile'] in self._files and 'rev' in self._files[have['depotFile']]:
            _sccInfo['sccDepotRevision'] = str(self._files[have['depotFile']]['rev'])
        _sccInfo['sccNeedSync'] = _sccInfo['sccLocalRevision'] != _sccInfo['sccDepotRevision']
        _sccInfo['sccConflict'] = '0'
        
        # necessary for status icons
        if _sccInfo['sccNeedSync']:
            _sccInfo['sccStatus'] = 'out of date'
            _sccInfo['sccSync'] = 1
        else:
            _sccInfo['sccStatus'] = 'up to date'
            _sccInfo['sccOk'] = 1
        
        if self._open.has_key(have['depotFile']):
            open = self._open[have['depotFile']]
            _sccInfo['sccAction'] = open['action']
            _sccInfo['sccChange'] = open['change']
        else:
            _sccInfo['sccAction'] = ''
            _sccInfo['sccChange'] = ''
        url.scc = _sccInfo
        return 1


import cvslib
class KoCVSFileChecker(KoFileCheckerBase):
    def __init__(self):
        timeline.enter('KoCVSFileChecker.__init__')
        KoFileCheckerBase.__init__(self)
        self._files = {}
        # XXX change _paths to _directories
        self._paths = {}
        self._not = {}
        self._cvsSvc = components.classes["@activestate.com/koCVS;1"].\
            getService(components.interfaces.koICVS)
        self.cvs = cvslib.CVS()
        self.type = 'scc'
        self.scc = 'cvs'
        self.executable = self._cvsSvc.executable
        self.executablePrefName = 'cvsExecutable'
        self.duration = 15
        self.durationPrefName = 'cvsBackgroundMinutes'
        self.enabledPrefName = 'cvsEnabled'
        self.backgroundPrefName = 'cvsBackgroundCheck'
        self.recursivePrefName = 'cvsRecursive'
        timeline.leave('KoCVSFileChecker.__init__')

    def setExecutable(self, executable):
        if executable:
            self.executable = executable
        else:
            self.executable = 'cvs'

    # we don't cache the where information into _path here, since that
    # would prevent getFileStatus from running correctly (ie retrieving
    # files/have/open data)
    def isRepository(self, path):
        basepath = os.path.dirname(path)
        npath = os.path.normcase(basepath)
        if npath in self._not: return 0
        if npath in self._paths: return 1
        cvsRepositoryFile = os.path.join(basepath,'CVS','Repository')
        if os.path.exists(cvsRepositoryFile): return 1
        self._not[npath]=time.time()
        return 0
        
    def isActive(self):
        self.enabled = self._globalPrefs.getBooleanPref(self.enabledPrefName)
        return self._cvsSvc.isFunctional and self.enabled

    def invalidatePath(self, url, force=0):
        if not url.path:
            return 0
        if url.isDirectory:
            return self._invalidate(url.path, force)
        else:
            return self._invalidate(url.dirName, force)
        
    def invalidateUrl(self, url, force=0):
        if not url.path:
            return 0
        return self._invalidate(url.path, force)
        
    def _invalidate(self, path, force):
        #print "invalidating ",path
        npath = os.path.normcase(path)
        if npath in self._paths:
            # this will force a refetch of the entire path
            # if the last stat on the path is older than
            # the duration
            lastcheck = self._paths[npath]
            if force or lastcheck < time.time() - self.duration *60:
                self._paths[npath] = 0
            else:
                return 0
        elif npath in self._not:
            lastcheck = self._not[npath]
            if force or lastcheck < time.time() - self.duration *60:
                self._not[npath] = 0
            else:
                return 0
        elif npath in self._files:
            del self._files[npath]
        return 1

    def _raiseCVSError(self, msg):
        # since we have no way to know if this is
        # informational or not, we have to parse for
        # what we want to know
        if msg.find('authorization failed') > -1:
            raise cvslib.CVSLibError(msg)
    
    def getFileStatus(self, url, dofile=0):
        # initialize all items.  This helps to clear the item properties
        # when it is removed from scc
        url.clearscc()
        _sccInfo = {}
        
        #print "getFileStatus CVS %d %s"%(dofile,url.url)
        basepath = path = url.path
        if not path:
            return None
        nbasepath = npath = os.path.normcase(path)
        
        if not url.isDirectory:
            basepath = url.dirName
            nbasepath = os.path.normcase(basepath)
        
        try:
            if nbasepath in self._not:
                return None

            self.cvs._cvs = self.executable
            
            # cvs is recursive for status, so lets make sure that both the file
            # and the path do not exist before we stat a path
            if (nbasepath not in self._paths or \
                self._paths[nbasepath] == 0) and \
                os.path.isdir(basepath):
                
                # cvs status does not return status on paths, so we
                # have to build a status for the path :(
                cvsRepositoryFile = os.path.join(basepath,'CVS','Repository')
                if not os.path.exists(cvsRepositoryFile):
                    self._not[nbasepath]=time.time()
                    return None
                
                _sccInfo['sccDirType']=self.scc
                
                self._files[nbasepath] = {
                    'filename':'',
                    'status':'Up-to-date',
                    'rev':'1',
                    'revdate':'',
                    'rrev':'1',
                    'rpath':'',
                    'path':basepath,
                    'stickytag':'',
                    'stickydate':'',
                    'stickyoptions':'',
                    'ondisk':1
                    }
                
                #print "checking CVS for path ",basepath
                # execute cvs on the path
                # XXX figure out how to parse showTags later
                result = self.cvs.status(cwd=basepath,recursive=self.recursive)
                if 'stderr' in result:
                    self._raiseCVSError(result['stderr'])
                        
                #print repr(result)
                #print "got %d files" % (len(result['files']))
                for file in result['files']:
                    filekey = os.path.normcase(file['path'])
                    path = os.path.dirname(filekey)
                    self._paths[path]=time.time()
                    #print "have ",filekey
                    file['ondisk'] = os.path.exists(filekey);
                    self._files[filekey] = file
                
                self._paths[nbasepath]=time.time()

            if dofile and not self._files.has_key(npath) and \
               not os.path.isdir(path):
                #print "checking cvs for file ",path
                # file not in cache, get it now and stuff it into the cache
                filename = os.path.basename(path)
                result = self.cvs.status(files=filename,cwd=basepath)
                if 'stderr' in result:
                    self._raiseCVSError(result['stderr'])
                    
                #print repr(result)
                for file in result['files']:
                    filekey = os.path.normcase(file['path'])
                    #print "have ",filekey
                    file['ondisk'] = os.path.exists(filekey);
                    self._files[filekey] = file
                
        except cvslib.CVSLibError, e:
            #we get an exception here if the cvs library call errors out
            log.debug(e)
            self.setError('CVS Status Error, see SCC Output pane for details',e)
            return None
        
        if not self._files.has_key(npath):
            # still don't have it, not a cvs file
            #print "cvs dont have ",npath
            return None
        
        # get cvs info for the path, not the file
        #print "setting info for ",npath
        file = self._files[npath]
        
        _sccInfo['sccType'] =  _sccInfo['sccDirType'] = 'cvs'
        _sccInfo['sccHaveOnDisk']= file['ondisk']
        _sccInfo['sccLocalRevision'] = file['rev']
        _sccInfo['sccRevdate'] = file['revdate']
        _sccInfo['sccDepotRevision'] = file['rrev']
        _sccInfo['sccNeedSync'] = _sccInfo['sccLocalRevision'] != _sccInfo['sccDepotRevision']
        _sccInfo['sccConflict'] = '0'
        _sccInfo['sccAction'] = ''
        _sccInfo['sccStatus'] = file['status']
        _sccInfo['sccChange'] = ''

        # this is independent of sync status
        if _sccInfo['sccStatus'].find('conflict') >= 0:
            _sccInfo['sccConflict'] = '1'

        # necessary for status icons
        if _sccInfo['sccNeedSync']:
            _sccInfo['sccSync'] = 1
        else:
            _sccInfo['sccOk'] = 1

        if (_sccInfo['sccLocalRevision'] and \
           (_sccInfo['sccLocalRevision'] == '0' or \
            _sccInfo['sccLocalRevision'][0] == '-')) or \
           _sccInfo['sccStatus'].find('Removed') >= 0:
            _sccInfo['sccAction'] = 'delete'
        elif (_sccInfo['sccRevdate'] and _sccInfo['sccRevdate'] == 'New File!') or \
            _sccInfo['sccStatus'].find('Added') >= 0:
            _sccInfo['sccAction'] = 'add'
        elif not _sccInfo['sccNeedSync'] and _sccInfo['sccStatus'] != 'Up-to-date':
            _sccInfo['sccAction'] = 'edit'
        url.scc = _sccInfo
        return 1

