#!python
# Copyright (c) 2000-2002 ActiveState SRL.
# See the file LICENSE.txt for licensing information.

# The implementation of the Komodo Run Output Window Terminal

import os
import sys
import re
import types
import threading
import process
import time
import logging
import traceback
import Queue

from xpcom import components, nsError, ServerException, COMException
from xpcom._xpcom import PROXY_SYNC, PROXY_ALWAYS, PROXY_ASYNC, getProxyForObject
from koTreeView import TreeView

#---- globals
log = logging.getLogger('koRunTerminal')

#---- internal support classes and routines


def _isurl(path):
    urlpattern = re.compile("^\w+://")
    if urlpattern.match(path):
        return 1
    else:
        return 0



#---- implementation

class _KoRunTerminalFile(process.IOBuffer):
    """A std handler for the Run output window.
    I.e. one each for <stdin>, <stdout>, and <stderr>.
    """
    _com_interfaces_ = [components.interfaces.koIFile]

    def __init__(self, name, terminal, scintilla, passBackLines=0):
        """Create a terminal file with the given name.
        "passBackLines" indicates whether this stream should use the
            terminal's .parseAndAddLine() method to have output lines parsed
            into its tree.
        """
        log.debug("[%s] _KoRunTerminalFile.__init__(name=%r)", name, name)
        self._terminal = terminal
        self._scintilla = scintilla
        self._passBackLines = passBackLines
        if self._passBackLines:
            self._pbbuf = ""  # pass-back buffer
        process.IOBuffer.__init__(self, self._terminal.mutex,
                                  self._terminal.stateChange, name=name)

    def _doWrite(self, s):
        log.debug("[%s] _KoRunTerminalFile._doWrite: write in: %r",
                  self._name, s)
        # anything we get in must be converted to unicode if it
        # isn't already
        try:
            s = unicode(s)
            l = len(s.encode('utf-8'))
        except:
            l = len(s)
        self._scintilla.readOnly = 0
        self._scintilla.addText(l, s)
        self._scintilla.gotoPos(self._scintilla.length)
        self._scintilla.readOnly = 1
        if self._passBackLines:
            self._pbbuf += s
            lines = self._pbbuf.splitlines(1)
            for line in lines:
                if line.endswith("\r\n") or line.endswith("\n") or line.endswith("\r"):
                    self._terminal.parseAndAddLine(line)
                else:
                    # This must be the last (incomplete line).
                    self._pbbuf = line
                    break
            else:
                self._pbbuf = ""
        
    def _doRead(self, n):
        """Pop 'n' bytes from the terminal in-buffer and return them."""
        log.debug("[%s] _KoRunTerminalFile._doRead(n=%r)", self._name, n)
        self._terminal.handlePendingEvents()
        if n < 0:
            idx = len(self._terminal.inBuffer)
        else:
            idx = min(n, len(self._terminal.inBuffer))
        retval = self._terminal.inBuffer[:idx]
        self._terminal.inBuffer = self._terminal.inBuffer[idx:]
        return retval

    def _doReadline(self, n):
        """Pop the front line (or n bytes of it, whichever is less) from
        the internal buffer and return it.
        """
        log.debug("[%s] _KoRunTerminalFile._doReadline(n=%r)", self._name, n)
        self._terminal.handlePendingEvents()
        idx = self._terminal.inBuffer.find('\n')
        if idx == -1:
            idx = len(self._terminal.inBuffer)
        else:
            idx += 1 # include the '\n'
        if n is not None:
            idx = min(idx, n) 
        retval = self._terminal.inBuffer[:idx]
        self._terminal.inBuffer = self._terminal.inBuffer[idx:]
        return retval

    def _haveLine(self):
        log.debug("[%s] _KoRunTerminalFile._haveLine()", self._name)
        self._terminal.handlePendingEvents()
        return len(self._terminal.inBuffer) > 0

    def _haveNumBytes(self, n):
        log.debug("[%s] _KoRunTerminalFile._haveNumBytes(n=%r)",
                  self._name, n)
        self._terminal.handlePendingEvents()
        # Note: Checking for >0 rather than >=n because don't want to
        #       do any buffering.
        return len(self._terminal.inBuffer) > 0

    def _doClose(self):
        log.debug("[%s] _KoRunTerminalFile._doClose", self._name)
        if self._passBackLines:
            if self._pbbuf:
                self._terminal.parseAndAddLine(self._pbbuf)
                self._pbbuf = ""
            self._terminal.parsedLastLine()

_keyCodeToLabel = None

def _buildKeyLabelMap():
    """ This function builds a map between event 'keyCodes' and their names,
    for dispatching to methods in the TerminalHandler
    """
    global _keyCodeToLabel
    _keyCodeToLabel = {}
    
    labels = """    
    DOM_VK_BACK_SPACE
    DOM_VK_TAB
    DOM_VK_CLEAR 
    DOM_VK_RETURN
    DOM_VK_ENTER
    DOM_VK_SHIFT
    DOM_VK_CONTROL 
    DOM_VK_ALT
    DOM_VK_PAUSE
    DOM_VK_CAPS_LOCK
    DOM_VK_ESCAPE 
    DOM_VK_SPACE
    DOM_VK_PAGE_UP
    DOM_VK_PAGE_DOWN
    DOM_VK_END 
    DOM_VK_HOME
    DOM_VK_LEFT
    DOM_VK_UP
    DOM_VK_RIGHT 
    DOM_VK_DOWN
    DOM_VK_PRINTSCREEN
    DOM_VK_INSERT
    DOM_VK_DELETE 
    DOM_VK_0
    DOM_VK_1
    DOM_VK_2
    DOM_VK_3 
    DOM_VK_4
    DOM_VK_5
    DOM_VK_6
    DOM_VK_7
    DOM_VK_8
    DOM_VK_9
    DOM_VK_SEMICOLON
    DOM_VK_EQUALS
    DOM_VK_A
    DOM_VK_B
    DOM_VK_C
    DOM_VK_D
    DOM_VK_E
    DOM_VK_F
    DOM_VK_G
    DOM_VK_H
    DOM_VK_I
    DOM_VK_J
    DOM_VK_K
    DOM_VK_L
    DOM_VK_M
    DOM_VK_N
    DOM_VK_O
    DOM_VK_P
    DOM_VK_Q
    DOM_VK_R
    DOM_VK_S
    DOM_VK_T
    DOM_VK_U
    DOM_VK_V
    DOM_VK_W
    DOM_VK_X
    DOM_VK_Y
    DOM_VK_Z
    DOM_VK_NUMPAD0
    DOM_VK_NUMPAD1
    DOM_VK_NUMPAD2
    DOM_VK_NUMPAD3
    DOM_VK_NUMPAD4
    DOM_VK_NUMPAD5
    DOM_VK_NUMPAD6
    DOM_VK_NUMPAD7
    DOM_VK_NUMPAD8
    DOM_VK_NUMPAD9
    DOM_VK_MULTIPLY
    DOM_VK_ADD
    DOM_VK_SEPARATOR
    DOM_VK_SUBTRACT
    DOM_VK_DECIMAL
    DOM_VK_DIVIDE
    DOM_VK_F1
    DOM_VK_F2
    DOM_VK_F3
    DOM_VK_F4
    DOM_VK_F5
    DOM_VK_F6
    DOM_VK_F7
    DOM_VK_F8
    DOM_VK_F9
    DOM_VK_F10
    DOM_VK_F11
    DOM_VK_F12
    DOM_VK_F13
    DOM_VK_F14
    DOM_VK_F15
    DOM_VK_F16
    DOM_VK_F17
    DOM_VK_F18
    DOM_VK_F19
    DOM_VK_F20
    DOM_VK_F21
    DOM_VK_F22
    DOM_VK_F23
    DOM_VK_F24
    DOM_VK_NUM_LOCK
    DOM_VK_SCROLL_LOCK
    DOM_VK_COMMA
    DOM_VK_PERIOD
    DOM_VK_SLASH
    DOM_VK_BACK_QUOTE
    DOM_VK_OPEN_BRACKET
    DOM_VK_BACK_SLASH
    DOM_VK_CLOSE_BRACKET
    DOM_VK_QUOTE
    DOM_VK_HELP
    """
    
    for name in labels.split():
        label = getattr(components.interfaces.nsIDOMKeyEvent, name, None)
        if label:
            _keyCodeToLabel[label] = name
        else:
            log.warn("Unknown DOM key event name: "+ name)

class koTerminalHandler:
    """
    This class is to be associated with views of type 'terminal'.
    It is the argument to "initWithTerminal" for that view.
    
    It is easy to add new handlers for special keystrokes.  Look at the way
    _handle_keypress_DOM_VK_BACK_SPACE is being defined for example.  See above
    for the list of special key codes.
    """
    _com_interfaces_ = [components.interfaces.koITerminalHandler]
    _reg_clsid_ = "36C5DDAA-7A5D-444D-80CA-17D52EBDBA9A"
    _reg_contractid_ = "@activestate.com/koTerminalHandler;1"
    _reg_desc_ = "Terminal View Handler"

    def __init__(self):
        self._lastErrorSvc = components.classes["@activestate.com/koLastErrorService;1"]\
            .getService(components.interfaces.koILastErrorService)

        self.passBackLines = 0
        self.active = 0 # Initially NOT interacting with a child process.
        self.stdin = None
        self.stdout = None
        self.stderr = None
        if not _keyCodeToLabel:
            _buildKeyLabelMap()
        self.lineBuffer = ''    # Current line input.
        self.inBuffer = ''      # All unread input except the current line.
        self._history = ['']
        self._historyIndex = 0
        # This is to avoid a lockup which can happen if e.g.
        # a keystroke is being received while the output thread is
        # writing to scintilla, which in turn can cause SciMoz events
        # to wait on the main thread.
        self.pendingEvents = Queue.Queue(0) # infinite

    #---- koIRunTerminal methods
    def setScintilla(self, scintilla):
        log.debug("[%s] KoRunTerminal.setScintilla(%s)", self, scintilla)
        self._scintilla = getProxyForObject(None, components.interfaces.ISciMozLite,
                                            scintilla,
                                            PROXY_ALWAYS | PROXY_SYNC)
        self._scintilla.styleSetFore(1, 0x0000bb);

        # Synchronization between the terminal and its _KoRunTerminalFile's.
        self.mutex = threading.RLock()
        self.stateChange = threading.Condition()

        self.lineBuffer = ''    # Current line input.
        self.inBuffer = ''      # All unread input except the current line.

    def startSession(self):
        # Move insertion point to the end of the output window (this is
        # necessary if we are not clearing the window's contents and the
        # cursor has been moved).
        self._scintilla.selectionStart = self._scintilla.length-1;
        self._scintilla.selectionEnd = self._scintilla.length-1;

        self.stdin = _KoRunTerminalFile("<stdin>", self, self._scintilla)
        self.stdout = _KoRunTerminalFile("<stdout>", self, self._scintilla, self.passBackLines)
        self.stderr = _KoRunTerminalFile("<stderr>", self, self._scintilla, self.passBackLines)
        self.active = 1
        self._history = ['']
        self._historyIndex = 0

    def endSession(self):
        self.active = 0
        self.stdin = None
        self.stdout = None
        self.stderr = None

    def clear(self):
        log.debug("koTerminalHandler.clear()")
        self.mutex.acquire()
        self._scintilla.readOnly = 0
        self._scintilla.text = ''
        self.lineBuffer = ''    # Current line input.
        self.inBuffer = ''      # All unread input except the current line.
        self._scintilla.gotoPos(self._scintilla.length)
        self._scintilla.readOnly = 1
        self.mutex.release()

        self._parseRegex = None
        self._cwd = None
        self._currentFile = None

    def handleKeyEvent(self, event):
        if not self.active: return
        # Build method name for a specific key
        log.debug("_handlePendingEvent started")
        modifiers = [event.type]
        if event.ctrlKey:
            modifiers.append('Ctrl')
        if event.altKey:
            modifiers.append('Alt')
        if event.charCode:
            if len(modifiers) > 1: # there's at least a ctrl, alt or shift
                modifiers.append(unichr(event.charCode))
            else:
                if event.type == 'keypress':
                    event.stopPropagation()
                    self.pendingEvents.put((self._handle_plainKey, (event.charCode,)))
                    self.stateChange.acquire()
                    self.stateChange.notify()
                    self.stateChange.release()
                return
        if event.shiftKey:
            modifiers.append('Shift')
        if event.keyCode:
            label = _keyCodeToLabel.get(event.keyCode, None)
            if label:
                modifiers.append(label)
            else:
                log.warn("No handler for keyCode: %r" % event.keyCode)
        name = '_handle_' + '_'.join(modifiers)
        method = getattr(self, name, None)
        if method:
            event.stopPropagation()
            self.pendingEvents.put((method, ()))
        else:
            log.warn("No handler for event: %r" % name)

        self.stateChange.acquire()
        self.stateChange.notify()
        self.stateChange.release()
        
    def handlePendingEvents(self):
        while 1:
            log.debug("handlePendingEvents: Looking for an event")
            try:
                method, args = self.pendingEvents.get_nowait()
                log.debug("handlePendingEvents: got event")
            except Queue.Empty:
                log.debug("handlePendingEvents: queue is empty")
                break
            log.debug("handlePendingEvents: calling method for event")
            method(*args) # calls the method to handle a particular event
        log.debug("handlePendingEvents: done")
    
    def _handle_plainKey(self, charCode):
        log.debug("[%s] _KoRunTerminal._handle_plainKey()" % threading.currentThread().getName())
        ch = unichr(charCode)
        self.mutex.acquire()
        self.lineBuffer += ch
        self._scintilla.readOnly = 0
        # Are we not in the input buffer?  If not, move us to the end of the document
        firstAllowed = self._scintilla.length - len(self.lineBuffer) + 1
        if (self._scintilla.currentPos < firstAllowed and
            self._scintilla.anchor < firstAllowed):
            self._scintilla.anchor = self._scintilla.currentPos = self._scintilla.length
        else:
            # at least part of the selection is 'legal'.
            # trim the selection to be inside the allowed region
            if self._scintilla.currentPos < firstAllowed:
                self._scintilla.currentPos = firstAllowed
            if self._scintilla.anchor < firstAllowed:
                self._scintilla.anchor = firstAllowed
        self._scintilla.addChar(charCode)
        self._scintilla.readOnly = 1
        self.mutex.release()

    def _handle_keypress_DOM_VK_RETURN(self):
        log.debug("XXX [%s] _KoRunTerminal._handle_keypress_DOM_VK_RETURN()" % id(self))
        if self.active:
            self.mutex.acquire()
            if self.lineBuffer:
                self._history.insert(self._historyIndex, self.lineBuffer)
                self._historyIndex += 1
            self.inBuffer = self.lineBuffer + '\n'
            self._scintilla.readOnly = 0
            self._scintilla.selectionEnd = self._scintilla.selectionStart = self._scintilla.length
            self._scintilla.replaceSel('\r\n')
            self._scintilla.readOnly = 1
            self.lineBuffer = ''
            # Notify the terminal in-file of the new input.
            self.mutex.release()
        log.debug("Exiting _handle_keypress_DOM_VK_RETURN")
    def _handle_keypress_DOM_VK_BACK_SPACE(self):
        log.debug("XXX [%s] _KoRunTerminalCommandHandler._handle_keypress_DOM_VK_BACK_SPACE()" % id(self))
        if self.active and self.lineBuffer:
            self.mutex.acquire()
            self._scintilla.readOnly = 0
            self._scintilla.deleteBack()
            self._scintilla.readOnly = 1
            self.lineBuffer = self.lineBuffer[:-1]
            self.mutex.release()

    # this causes a crash -- don't know why yet.  Will investigate w/ a debug build later.
    def _handle_keypress_Shift_DOM_VK_UP(self):
        if self.active:
            self.mutex.acquire()
            self._scintilla.readOnly = 0
            self._scintilla.lineUp()
            self._scintilla.readOnly = 1
            self.mutex.release()


    def _handle_keypress_DOM_VK_LEFT(self):
        firstAllowed = self._scintilla.length - len(self.lineBuffer)
        if self.active and \
           self.lineBuffer and \
           self._scintilla.currentPos > firstAllowed:
            self._scintilla.currentPos = self._scintilla.anchor = self._scintilla.currentPos - 1

    def _handle_keypress_DOM_VK_RIGHT(self):
        if self.active and self._scintilla.currentPos < self._scintilla.length:
            self._scintilla.currentPos = self._scintilla.anchor = self._scintilla.currentPos + 1

    def _handle_keypress_DOM_VK_UP(self):
        if self.active and self._historyIndex > 0:
            self._remember(self._historyIndex - 1)

    def _handle_keypress_DOM_VK_DOWN(self):
        if self.active and self._historyIndex < len(self._history) - 1:
            self._remember(self._historyIndex + 1)

    def _remember(self, historyIndex):
        self.mutex.acquire()
        self._scintilla.readOnly = 0
        # clear current line
        self._scintilla.selectionStart = self._scintilla.length - len(self.lineBuffer)
        self._scintilla.selectionEnd = self._scintilla.length
        event = self._history[historyIndex]
        self._historyIndex = historyIndex
        self._scintilla.replaceSel(event)
        self._scintilla.readOnly = 1
        self.lineBuffer = event
        self.mutex.release()

class KoRunTerminal(koTerminalHandler, TreeView):
    """This is the interface between run sub-processes and Komodo's command
    output window (or _one_ of the command output windows if there are ever
    more than one). This acts as the handler for events on the output
    window's <scintilla> widget. It acts as the tree view for the output
    window's <tree> widget.

    The embedded Scintilla is kept in a 'readOnly' state unless
    currently interacting with the child process (self.active is True).
    XXX Perhaps there is a better API for that (perhaps call it
    'interactive')?

    Input is buffered in line chunks, i.e. user input is forwarded to
    the child a line at a time when <Enter> is pressed. This means that
    one can use backspace on the current line only.

    If enabled, output lines are parsed with a given regular expression
    as results into its tree.
    """
    _com_interfaces_ = [components.interfaces.koITreeOutputHandler,
                        components.interfaces.nsITreeView]
    _reg_clsid_ = "{A077FDDF-DBAD-4e49-9090-4390C084893B}"
    _reg_contractid_ = "@activestate.com/koRunTerminal;1"
    _reg_desc_ = "Run Output Window Terminal Handler"

    def __init__(self):
        #print "XXX [%s] KoRunTerminal.__init__()" % id(self)
        # Initialization of tree and parsing attributes.
        TreeView.__init__(self, debug=0)
        koTerminalHandler.__init__(self)
        self.passBackLines = 1
        self._data = []
        self._sortedBy = None
        self._parseRegex = None
        self._cwd = None
        self._currentFile = None
        # These two vars are used to group update calls to the outliner,
        # .rowCountChanged(), .invalidate(), for performance reasons.
        self._updateGroupSize = 20
        self._updateGroupCounter = 0

    def setParseRegex(self, parseRegex):
        try:
            self._parseRegex = re.compile(parseRegex)
        except re.error, ex:
            self._lastErrorSvc.setLastError(0, str(ex))
            raise ServerException(nsError.NS_ERROR_INVALID_ARG, str(ex))

    def setCwd(self, cwd):
        if cwd:
            self._cwd = cwd
        else:
            self._cwd = os.getcwd()

    def setCurrentFile(self, filename):
        self._currentFile = filename

    def clear(self):
        koTerminalHandler.clear(self)
        self._parseRegex = None
        self._cwd = None
        self._currentFile = None
        
        self._data = []
        self._sortedBy = None
        self._tree.rowCountChanged(0, 0)
        self._tree.invalidate()

    def getFile(self, index):
        return self._data[index]["file"]
    def getLine(self, index):
        return self._data[index]["line"]
    def getColumn(self, index):
        return self._data[index]["column"]

    def parseLine(self, line):
        """Parse the given line with the current regex and return:
            (file, line, column)

        Raise an exception and set an error on the last error service if
        there is no current regex or the line does not match.
        
        (This is used to handle double-clicks in the command output window's
        scintilla.)
        """
        if not self._parseRegex:
            errmsg = "There was no parse regex specified for this command.";
            self._lastErrorSvc.setLastError(0, errmsg);
            raise ServerException(nsError.NS_ERROR_NOT_AVAILABLE, errmsg)

        match = self._parseRegex.search(line)
        if not match:
            errmsg = "This line does not match the current pattern: '%s'"\
                     % self._parseRegex.pattern
            self._lastErrorSvc.setLastError(0, errmsg);
            raise ServerException(nsError.NS_ERROR_NOT_AVAILABLE, errmsg)

        datum = match.groupdict("")
        if "file" not in datum:
            if self._currentFile:
                datum["file"] = self._currentFile
            else:
                datum["file"] = ""
        elif not _isurl(datum["file"]) \
                and not os.path.isabs(datum["file"]) \
                and self._cwd:
            # If this is a relative path name then prepend the cwd.
            datum["file"] = os.path.join(self._cwd, datum["file"])
        if "line" not in datum:
            datum["line"] = ""
        if "column" not in datum:
            datum["column"] = ""
        return datum["file"], datum["line"], datum["column"]

    def sort(self, sortBy):
        """Sort the current data by the given key. If already sorted by this
        key then reverse the sorting order."""
        if self._sortedBy == sortBy:
            self._data.reverse()
        else:
            try:
                if sortBy in ('runoutput-tree-line',
                              'runoutput-tree-column'):
                    # compare the integer values (line and column string
                    # can possibly be ranges, so pick first part of range)
                    try:
                        self._data.sort(lambda dict1,dict2,sortBy=sortBy:
                                            cmp(int( dict1[sortBy].split('-')[0] ),
                                                int( dict2[sortBy].split('-')[0] ))
                                       )
                    except ValueError:
                        self._data.sort(lambda dict1,dict2,sortBy=sortBy:
                                            cmp(dict1[sortBy], dict2[sortBy])
                                       )
                elif sortBy == 'runoutput-tree-content':
                    # strip leading whitespace for content sort order
                    self._data.sort(lambda dict1,dict2,sortBy=sortBy:
                                        cmp(dict1[sortBy].lstrip(), dict2[sortBy].lstrip())
                                   )
                else:
                    self._data.sort(lambda dict1,dict2,sortBy=sortBy:
                                        cmp(dict1[sortBy], dict2[sortBy])
                                   )
            except KeyError:
                log.error("Cannot sort find results by: '%s'" % sortBy)
                raise
        self._sortedBy = sortBy
        self._tree.invalidate()

    #---- nsITree methods
    def get_rowCount(self):
        return len(self._data)

    def getCellText(self, row, col):
        try:
            datum = self._data[row][col]
        except IndexError:
            # Silence this, it is too annoying.
            # c.f. http://bugs.activestate.com/Komodo/show_bug.cgi?id=27487
            #log.error("no %sth result" % row)
            return ""
        except KeyError:
            log.error("unknown result column id: '%s'" % col)
            return ""
        #if type(datum) not in (types.StringType, types.UnicodeType):
        #    datum = str(datum)
        return datum

    #---- non-XPCOM methods
    def parseAndAddLine(self, line):
        #print "XXX KoRunTerminal.parseAndAddLine(line=%r)" % line
        if not self._parseRegex:
            return
        try:
            match = self._parseRegex.search(line)
        except RuntimeError:
            # Can get an error like the following:
            #   RuntimeError: maximum recursion limit exceeded
            # from the regex engine for rediculously long lines with
            # potential null byte content. Just silently drop those.
            return
        if match:
            datum = match.groupdict("")
            # Ensure that there are reasonable defaults for the std
            # tree column names.
            if "file" not in datum:
                if self._currentFile:
                    datum["file"] = self._currentFile
                else:
                    datum["file"] = ""
            elif not _isurl(datum["file"]) \
                 and not os.path.isabs(datum["file"]) \
                 and self._cwd:
                # If this is a relative path name then prepend the cwd.
                datum["file"] = os.path.join(self._cwd, datum["file"])
            if "line" not in datum:
                datum["line"] = ""
            if "column" not in datum:
                datum["column"] = ""
            if "content" not in datum:
                datum["content"] = line # default "content" to the whole line
            if "runoutput-tree-file" not in datum:
                datum["runoutput-tree-file"] = datum["file"]
            if "runoutput-tree-line" not in datum:
                datum["runoutput-tree-line"] = datum["line"]
            if "runoutput-tree-column" not in datum:
                datum["runoutput-tree-column"] = datum["column"]
            if "runoutput-tree-content" not in datum:
                datum["runoutput-tree-content"] = datum["content"]

            self._data.append(datum)
            self._sortedBy = None

            # Group update calls to the tree for performance.
            self._updateGroupCounter += 1
            if self._updateGroupCounter == self._updateGroupSize:
                self._tree.rowCountChanged(
                    max(0, len(self._data) - 1 - self._updateGroupCounter),
                    self._updateGroupCounter)
                self._updateGroupCounter = 0

    def parsedLastLine(self):
        """Indicates that the last line of data to parse has been send.
        
        This lets the terminal do any necessary finalization.
        """
        if self._updateGroupCounter:
            self._tree.rowCountChanged(
                max(0, len(self._data) - 1 - self._updateGroupCounter),
                self._updateGroupCounter)
            self._updateGroupCounter = 0
