#!python
# Copyright (c) 2000-2003 ActiveState Corporation.
# See the file LICENSE.txt for licensing information.

from xpcom import components, nsError, ServerException
from koLintResult import *
from koLintResults import koLintResults
import os, sys, re, string
import process
import koprocessutils
import logging


log = logging.getLogger("koTclLinter")
endWordSet = ";} \t\n"


class KoTclCompileLinter:
    _com_interfaces_ = [components.interfaces.koILinter]
    _reg_desc_ = "Komodo Tcl Linter"
    _reg_clsid_ = "{842EC3B2-B79B-44a2-9417-04CEA5C02BF1}"
    _reg_contractid_ = "@activestate.com/koTclLinter;1"

    def __init__(self):
        self._linterArgvCache = None
        self._tclProRe = re.compile(r'^[^:]+:(\d+) \((\w+)\) (.*)$')

        self._lastErrorSvc = components.classes["@activestate.com/koLastErrorService;1"].\
            getService(components.interfaces.koILastErrorService)
        koDirs = components.classes["@activestate.com/koDirs;1"].\
                 getService(components.interfaces.koIDirs)
        self._installDir = koDirs.installDir
        self.internalLinter = os.path.join(self._installDir, "tcl",
                                           "komodo_tcl_lint.tcl")

    def _getLinterArgv(self):
        if self._linterArgvCache:
            return self._linterArgvCache

        if sys.platform.startswith('win'):
            executable = 'kotcllint.exe'
        else:
            executable = 'kotcllint'
        linter = os.path.join(self._installDir, 'tcl', executable)
        linterArgv = [linter, '-onepass']

        self._linterArgvCache = linterArgv
        return linterArgv

    def lint(self, request):
        """Lint the given Tcl content.
        
        Raise an exception and set an error on koLastErrorService if there
        is a problem.
        """
        argv = self._getLinterArgv()
        env = koprocessutils.getUserEnv()
        cwd = request.cwd or None
        text = request.content.encode(request.encoding.python_encoding_name)
        p = process.ProcessOpen(argv, cwd=cwd, env=env, mode='b')
        p.stdin.write(text)
        p.stdin.close()
        lOutput = p.stdout.readlines()
        lErrOut = p.stderr.read(-1)
        p.close()
        #print '======'
        #print lOutput
        #print '======'
        #print lErrOut
        #print '<<<<<<'

        if not lOutput and lErrOut:
            log.error(lErrOut)
            self._lastErrorSvc.setLastError(1, lErrOut)
            raise ServerException(nsError.NS_ERROR_UNEXPECTED)
            
        if lOutput[0].startswith("TclPro") or \
           lOutput[0].startswith("scanning"):
            # The open source TclPro has the header stripped,
            # so we need to check for both "TclPro" and "scanning"
            lInput = text.split('\n')
            return self._TclPro_lint(lInput, lOutput)
        else:
            errmsg = "unrecognized lint output format:\n%s" % lOutput[0]
            log.error(errmsg)
            self._lastErrorSvc.setLastError(1, errmsg)
            raise ServerException(nsError.NS_ERROR_UNEXPECTED)

    def _TclPro_lint(self, lInput, lOutput):
        # TclPro output has the following header info:
        ###################
        #TclPro -- Version 1.4.1
        #Copyright (C) Ajuba Solutions 1998-2001. All rights reserved.
        #This product is registered to: <user name>
        #
        #scanning: <filename>
        #checking: <filename>
        #<filename>:<lineNo> (<errorName>) <errorText>
        #<commandWithError>
        #<caret at col in command where error is>
        #[repeat above three until no more errors]
        ###################
        # The 'checking' line exists when the -onepass option isn't used
        i = 1
        lastCol = -1
        lastLine = -1
        results = koLintResults()
        numLines = len(lOutput)
        while i < numLines:
            match = re.search(self._tclProRe, lOutput[i])
            if not match:
                # We expect the first few lines to not match, while we
                # skip the header lines
                i += 1
                # But if too many lines aren't matching, then we must
                # unrecognizable input
                if i > 8:
                    log.warn("couldn't recognized Tcl linter output format")
                    break
                continue
            if i+2 >= numLines:
                log.error("unexpected return data format: short data")
                return                    
            cmd = lOutput[i+1].rstrip()
            col = lOutput[i+2].find('^')
            if col == -1:
                log.error("unexpected return data format: no caret")
                return
            lineNo = match.group(1)
            # lInput is zero-based, whereas text display lines start at one,
            # so we adjust lineNo down one here.
            lineNo = int(lineNo) - 1
            if lineNo != lastLine:
                lastLine = lineNo
                lastCol = 0

            startCol = lInput[lineNo].find(cmd, lastCol)
            if startCol == -1:
                # It's possible that we are throwing multiple warnings on
                # the same command.  Let's check again, starting at col 0.
                # If something was found, ignore multiple warnings in the
                # same line, otherwise log a problem
                if lastCol == 0 or \
                   lInput[lineNo].find(cmd, 0) == -1:
                    # Looks like we can't figure something out...
                    log.debug('undecipherable lint output: %s,%s "%s" in "%s" last %s' \
                              % (lineNo+1, col, cmd, lInput[lineNo], lastCol))
                i += 3
                continue

            #print 'TclLint: %s,%s "%s" in "%s"' \
            #      % (lineNo+1, startCol, cmd, lInput[lineNo])

            result = KoLintResult()
            msg = match.group(2)
            result.description = match.group(3).strip()
            result.lineStart = lineNo+1
            # All our warnings stay on one line
            result.lineEnd = result.lineStart
            cmdLen = len(cmd)
            if msg.startswith("warn") or msg.startswith("nonPort"):
                result.severity = result.SEV_WARNING
                # In the warning case, go to the next char that we think
                # ends the word (not precise), starting at the caret.
                endCol = col
                while endCol < cmdLen and cmd[endCol] not in endWordSet:
                    # might want to handle \ escapes
                    endCol += 1
                # readjust for where the command started
                endCol += startCol
                # and bump up the startCol to where the caret indicated
                startCol += col
            else:
                result.severity = result.SEV_ERROR
                # In the error case, highlight the entire command
                # The linter should only give one error per command
                endCol = startCol + cmdLen
            # Adjust by 1 as columns start at 1 for the text display
            result.columnStart = startCol + 1
            result.columnEnd = endCol + 1
            results.addResult(result)

            # Set lastCol to properly handle multiple similar code-blocks
            # with errors on the same line.
            lastCol = startCol+1
            i += 3
            # end while
        return results

