#!/usr/bin/env python
# Copyright (c) 2003 ActiveState Corp.
# See the file LICENSE.txt for licensing information.

"""Komodo's GUI Test Suite service"""

import sys
import os
import threading
import logging
import unittest
import time

from xpcom import components, nsError, ServerException, COMException

#---- globals

log = logging.getLogger('testsvc')

#---- the service and support stuff

class KoTestSvc:
    _com_interfaces_ = [components.interfaces.koITestSvc]
    _reg_clsid_ = "7084EBB9-1C71-403B-AB8F-CDD3C9AB9BE0"
    _reg_contractid_ = "@activestate.com/koTestSvc;1"
    _reg_desc_ = "Service for interacting with the unit test system"

    def __init__(self):
        self.useSmoke = 0
        self.server = None
        self.projectId = None # 0 or None means "use default"
        self.buildId = None # 0 or None means "use default"

        self._runner = None # runner thread

    def setSmokeInfo(self, useSmoke, server, projectId, buildId):
        self.useSmoke = useSmoke
        self.server = server
        self.projectId = projectId
        self.buildId = buildId

    def runAllTests(self):
        #XXX Eventually, getting this suite will come from gathering
        #    suites from all found testgui_*.py modules (a la test_*.py
        #    suite gathering done in test/test.py).
        suite = unittest.makeSuite(TestUI)

        self._runner = TestRunningThread(suite, self.useSmoke,
                                         self.server, self.projectId,
                                         self.buildId)
        self._runner.start()

    def join(self):
        if self._runner:
            self._runner.join()


class TestRunningThread(threading.Thread):
    def __init__(self, suite, useSmoke, server, projectId, buildId):
        threading.Thread.__init__(self)
        self.suite = suite
        self.useSmoke = useSmoke
        self.server = server
        self.projectId = projectId
        self.buildId = buildId

    def run(self):
        if self.useSmoke:
            from smokeclient.unittestutils import SmokeTestRunner
            #XXX Might want to eventually subclass SmokeTestRunner to
            #    our own one that provides some GUI feedback as well.
            #    Actually, we would want our TestRunner to use a
            #    subclass of what is now called _QuietSmokeTestResult.
            runner = SmokeTestRunner(server=self.server,
                                     project_id=self.projectId,
                                     build_id=self.buildId)
            
        else:
            #XXX Should write a GUITestRunner that displays results in a
            #    XUL Window for the user to see.
            runner = unittest.TextTestRunner(sys.stdout, verbosity=2)
        result = runner.run(self.suite)

        #XXX Eventually this should notify whoever cares when the test
        #    suite is complete. Presumably "bk test --gui" will, at some
        #    point, want to be able to synchronize on this.

        


class TestUI(unittest.TestCase):
    def __init__(self, *args, **kw):
        unittest.TestCase.__init__(self, *args, **kw)
        self._windowMediator = components.classes["@mozilla.org/appshell/window-mediator;1"].\
                        getService(components.interfaces.nsIWindowMediator)
        self._focused_window = self._windowMediator.getMostRecentWindow(None)
        self._document = self._focused_window.document
        self._dialogSvc = components.classes['@activestate.com/asDialogProxy;1'].\
                          getService(components.interfaces.asIDialogProxy)
        self.komodo = None
        self._viewSvc = components.classes["@activestate.com/koViewService;1"].\
                          getService(components.interfaces.koIViewService)
        self._commandmentSvc = components.classes['@activestate.com/koCommandmentService;1'].\
                              getService()
        
    def _doBroadcaster(self, broadcasterId):
        # find a broadcaster and do its oncommand
        broadcaster = self._document.getElementById(broadcasterId)
        self.failUnless(broadcaster)
        oncommand = broadcaster.getAttribute('oncommand')
        self.failUnless(oncommand)
        self._dialogSvc.eval_(self._focused_window, oncommand)

    def _test_toolbar_visibility(self, toolbarId, menuId):
        # This test makes sure that by toggling the hidden element of the
        # toolbars, and updating the right broadcasters, the menu items
        # reflect the state of things.

        toolbar = self._document.getElementById(toolbarId)
        self.failUnless(toolbar, "couldn't find toolbar: %r" % toolbarId)
        
        # save old state 
        if toolbar.hasAttribute('hidden'):
            oldstate = toolbar.getAttribute('hidden')
        else:
            oldstate = None

        broadcasterId = toolbar.getAttribute('broadcaster')
        self.failUnless(broadcasterId, "No 'broadcaster' attribute defined for toolbar: %r" % toolbarId)

        broadcaster = self._document.getElementById(broadcasterId)
        self.failUnless(broadcaster, "couldn't find broadcaster: %r" % broadcasterId)

        menuitem = self._document.getElementById(menuId)
        self.failUnless(menuitem, "Couldn't find menuitem: %s" % menuId)

        # 1) We hide the toolbar
        toolbar.setAttribute('hidden', 'true')
        # 2 We update the toolbar broadcaster based on the toolbar
        self._dialogSvc.eval_(self._focused_window,
                              "uilayout_updateToolbarBroadcaster('%s')" % toolbarId)
        self.failIf(menuitem.hasAttribute('checked') and
                    menuitem.getAttribute('checked') == 'true',
                    "%r shouldn't be checked" % menuId)

        toolbar.removeAttribute('hidden')  # make the pane visible
        self._dialogSvc.eval_(self._focused_window,
                              "uilayout_updateToolbarBroadcaster('%s')" % toolbarId)
        self.failUnless(menuitem.hasAttribute('checked'), 'menuitem should have a checked attribute')
        checked = menuitem.getAttribute('checked')
        self.failUnless(checked == u'true', "menuitem's checked attribute should be true")

        toolbar.setAttribute('hidden', 'true')
        self._dialogSvc.eval_(self._focused_window,
                              "uilayout_updateToolbarBroadcaster('%s')" % toolbarId)
        self.failIf(menuitem.hasAttribute('checked') and
                    menuitem.getAttribute('checked') == 'true')

        # now try using the command to change the visibility
        self._doBroadcaster(broadcasterId)
        self.failIf(getAttribute(toolbar, 'hidden', 'false') != 'false')

        # reset old state
        if oldstate:
            toolbar.setAttribute('hidden', oldstate)
        else:
            if toolbar.hasAttribute('hidden'):
                toolbar.removeAttribute('hidden')
        self._dialogSvc.eval_(self._focused_window,
                              "uilayout_updateToolbarBroadcaster('%s')" % toolbarId)

    def test_toolbar_visibility(self):
        data = [('standardToolbar', 'menu_viewedittoolbar'),
                ('debuggerToolbar', 'menu_viewdebugtoolbar'),
                ('sccToolbar', 'menu_viewscctoolbar'),
                ]
        for (toolbarId, menuId) in data:
            self._test_toolbar_visibility(toolbarId, menuId)
    
    def test_toolbar_text_visibility(self):
        menuitemId = 'menu_toggleButtonText'

        menuitem = self._document.getElementById(menuitemId)
        assert menuitem

        oldstate = getAttribute(menuitem, 'checked', 'false')

        self._doBroadcaster('cmd_toggleButtonText')
        assert menuitem.getAttribute('checked') != oldstate
        
        self._doBroadcaster('cmd_toggleButtonText')
        assert menuitem.getAttribute('checked') == oldstate

    def failUnlessDisabled(self, menuId):
        menuitem = self._document.getElementById(menuId)
        disabled = 'false'
        if menuitem.hasAttribute('disabled'):
            disabled = menuitem.getAttribute('disabled')
        self.failUnless(disabled == 'true', "Menu item: %s was not disabled as it should have been" % menuId)
    failIfEnabled = failUnlessDisabled

    def failIfDisabled(self, menuId):
        menuitem = self._document.getElementById(menuId)
        disabled = 'false'
        if menuitem.hasAttribute('disabled'):
            disabled = menuitem.getAttribute('disabled')
        self.failIf(disabled == 'true', "Menu item: %s was disabled and should not have been" % menuId)
    failUnlessEnabled = failIfDisabled
    
    def sendKey(self, ch):
        if self.komodo is None:
            import win32com.client    
            # - get a WScript handle on it, for SendKeys playing
            self.komodo = win32com.client.Dispatch("WScript.Shell")
            self.komodo.AppActivate(os.getpid())
        if ch == '\r':
            return
        if ch == '\n':
            self.komodo.SendKeys("{ENTER}")
            # defeat komodo's autoindent
            self.komodo.SendKeys("{HOME}")
            self.komodo.SendKeys("{HOME}")
        elif ch == '\t':
            self.komodo.SendKeys("{TAB}")
        elif ch in ["(",")","{","}","[","]","+","^","%","~"]:
            ch = "{%s}"%(ch)
            self.komodo.SendKeys(ch)
        else:
            self.komodo.SendKeys(ch)

    def doCommand(self, commandId):
        "Evaluate a command"
        self._dialogSvc.eval_(self._focused_window,
                              "command_doCommand('" + commandId + "')")

    def test_typing(self):
        koDirs = components.classes["@activestate.com/koDirs;1"].\
                    getService(components.interfaces.koIDirs)
        test_file = os.path.join(koDirs.installDir, 'test', 'test_files', "Makefile.PL")
        commandment = "open\t%s" % test_file
        self.failUnlessDisabled('menu_bufferClose')
        self._commandmentSvc.handleCommandment(commandment)
        view = self._viewSvc.currentView
        doc = view.document
        self.failUnless(doc.bufferLength == 1849,
                        "bufferLength == %d instead of 1849" % doc.bufferLength)
        test_string = "Hello World! {\nasdasd}"
        for char in test_string:
            self.sendKey(char)
        time.sleep(1) # let the keys get there.
        l = 1849 + len(test_string) + 1 # we're on windows, so \n will get converted to two chars.
        self.failUnless(doc.bufferLength == l,
                        "bufferLength == %d instead of %d" % (doc.bufferLength, l))
        # this would fail when we fix close to not close dirty buffers, but not sure
        # how to detect this failure since it'll bring up a dialog =(
        self.doCommand('cmd_bufferClose') 
        self.failUnlessDisabled('menu_bufferClose')

    def test_open(self):
        "Tests whether files get opened properly"
        koDirs = components.classes["@activestate.com/koDirs;1"].\
                    getService(components.interfaces.koIDirs)
        test_file = os.path.join(koDirs.installDir, 'test', 'test_files', "Makefile.PL")
        commandment = "open\t%s" % test_file
        self.failUnlessDisabled('menu_bufferClose')
        print "Doing commandment:", commandment
        self._commandmentSvc.handleCommandment(commandment)
        self.failIfDisabled('menu_bufferClose')
        self.failIfDisabled('menu_cut')
        self.failIfDisabled('menu_copy')
        self.failIfDisabled('menu_paste')
        self.failUnlessDisabled('menu_undo')
        # Check a few things about the opened document
        currentLanguage = self._viewSvc.currentView.document.language
        self.failUnless(currentLanguage == 'Perl')
        view = self._viewSvc.currentView
        doc = view.document
        self.failUnless(doc.bufferLength == 1849,
                        "bufferLength == %d instead of 1849" % doc.bufferLength)
        self.doCommand('cmd_editSelectAll')
        #return
        self.doCommand('cmd_cut')
        #self.failUnlessEnabled('menu_undo')
        # after cutting all selection, there should be nothing left in the buffer
        self.failUnless(doc.bufferLength == 0,
                        "bufferLength == %d instead of 0" % doc.bufferLength)
        self.doCommand('cmd_paste')
        # after pasting, we should be back to full length
        self.failUnless(doc.bufferLength == 1849,
                        "bufferLength == %d instead of 1849" % doc.bufferLength)
        #self.failUnlessEnabled('menu_undo')
        self.doCommand('cmd_undo')
        # after undoing, we should be back to empty
        self.failUnless(doc.bufferLength == 0,
                        "bufferLength == %d instead of 0" % doc.bufferLength)
        self.doCommand('cmd_undo')
        # after undoing again, we should be back to full and an empty undo stack
        self.failUnless(doc.bufferLength == 1849,
                        "bufferLength == %d instead of 1849" % doc.bufferLength)
        #self.failUnlessDisabled('menu_undo')
        #self.failUnlessEnabled('menu_redo')
        self.doCommand('cmd_bufferClose')
        self.failUnlessDisabled('menu_bufferClose')

def getAttribute(obj, attr, default):
    if obj.hasAttribute(attr):
        return obj.getAttribute(attr)
    else:
        return default


