#!/usr/bin/env python
# Copyright (c) 2000-2001 ActiveState SRL.
# See the file LICENSE.txt for licensing information.

import sys, os, re
from xpcom import components, ServerException, nsError
import process
import koprocessutils

#---- internal support functions
# XXX _mkdir and _copy are duplicated in Komodo code.

def _mkdir(newdir):
    """works the way a good mkdir should :)
        - already exists, silently complete
        - regular file in the way, raise an exception
        - parent directory(ies) does not exist, make them as well
    """
    if os.path.isdir(newdir):
        pass
    elif os.path.isfile(newdir):
        raise OSError("a file with the same name as the desired " \
                      "dir, '%s', already exists." % newdir)
    else:
        head, tail = os.path.split(newdir)
        if head and not os.path.isdir(head):
            _mkdir(head)
        #print "_mkdir %s" % repr(newdir)
        if tail:
            os.mkdir(newdir)


def _copy(src, dst):
    """works the way a good copy should :)
        - no source, raise an exception
        - destination directory, make a file in that dir named after src
        - source directory, recursively copy the directory to the destination
        - filename wildcarding allowed
        - make the files writeable
    """
    import string, glob, shutil

    # determine if filename wildcarding is being used
    # (only raise error if non-wildcarded source file does not exist)
    if string.find(src, '*') != -1 or \
       string.find(src, '?') != -1 or \
       string.find(src, '[') != -1:
        usingWildcards = 1
        srcFiles = glob.glob(src)
    else:
        usingWildcards = 0
        srcFiles = [src]

    for srcFile in srcFiles:
        if os.path.isfile(srcFile):
            if usingWildcards:
                srcFileHead, srcFileTail = os.path.split(srcFile)
                srcHead, srcTail = os.path.split(src)
                dstHead, dstTail = os.path.split(dst)
                if dstTail == srcTail:
                    dstFile = os.path.join(dstHead, srcFileTail)
                else:
                    dstFile = os.path.join(dst, srcFileTail)
            else:
                dstFile = dst
            dstFileHead, dstFileTail = os.path.split(dstFile)
            if dstFileHead and not os.path.isdir(dstFileHead):
                _mkdir(dstFileHead)
            #print "copy %s to %s" % (repr(srcFile), repr(dstFile))
            if os.path.isfile(dstFile):
                # make sure 'dstFile' is writeable
                os.chmod(dstFile, 0755)
            shutil.copy(srcFile, dstFile)
            # make the new 'dstFile' writeable
            os.chmod(dstFile, 0755)
        elif os.path.isdir(srcFile):
            srcFiles = os.listdir(srcFile)
            if not os.path.exists(dst):
                _mkdir(dst)
            for f in srcFiles:
                s = os.path.join(srcFile, f)
                d = os.path.join(dst, f)
                try:
                    _copy(s, d)
                except (IOError, os.error), why:
                    raise OSError("Can't copy %s to %s: %s"\
                          % (repr(s), repr(d), str(why)))
        elif not usingWildcards:
            raise OSError("Source file %s does not exist" % repr(srcFile))



#---- components

class KoPerlWSSetupService:
    """Service for setting up a Perl installation for Web Services."""
    _com_interfaces_ = [components.interfaces.koIWSSetupService]
    _reg_clsid_ = "2506c712-0e23-4a50-bce6-ed12b1a23aea"
    _reg_contractid_ = "@activestate.com/koPerlWSSetupService;1"
    _reg_desc_ = "Perl Web Service Setup Service"

    def __init__(self):
        self.installationPath = None
        self._globalPrefSvc = components.classes["@activestate.com/koPrefService;1"]\
                              .getService(components.interfaces.koIPrefService)

    def SetInstallationPath(self, installationPath):
        self.installationPath = installationPath
        # determine which ppm to invoke
        if sys.platform.startswith("win"):
            self.ppm = os.path.join(self.installationPath,
                                    "bin", "ppm.bat")
        else:
            self.ppm = os.path.join(self.installationPath,
                                    "bin", "ppm")

    def _getPerlInfo(self):
        # iii. Ensure the installation path is valid.
        if not self.installationPath:
            return 0,"No Perl installation was specified."
        if not os.path.isdir(self.installationPath):
            return 0,"The given Perl installation dir does not exist: "\
                   "'%s'. You must specify an existing Perl "\
                   "installation."\
                   % self.installationPath

        # Trap a common error: choosing <prefix>/bin instead of just
        # <prefix>.
        if sys.platform.startswith("win"):
            perlExe = os.path.join(self.installationPath, "perl.exe")
        else:
            perlExe = os.path.join(self.installationPath, "perl")
        if os.path.isfile(perlExe):
            return 0,"You selected the Perl bin directory. The installation "\
                   "path is one directory up, '%s'. Please select that "\
                   "directory." % os.path.dirname(self.installationPath)

        # Ensure the perl executable exists in the expected place.
        if sys.platform.startswith("win"):
            perlExe = os.path.join(self.installationPath, "bin", "perl.exe")
        else:
            perlExe = os.path.join(self.installationPath, "bin", "perl")
        if not os.path.isfile(perlExe):
            return 0,"There is no Perl installation at '%s': could not find "\
                   "the perl executable: '%s'."\
                   % (self.installationPath, perlExe)

        # At this point we have a perl installation. Now we make sure
        # that it is a sufficient installation (version, distro, etc.).
        perlInfoEx = components.classes["@activestate.com/koPerlInfoEx;1"]\
           .createInstance(components.interfaces.koIAppInfoEx)
        perlInfoEx.installationPath = self.installationPath
        try:
            revision, version = map(int, perlInfoEx.version.split(".")[:2])
        except ValueError: # Old format perl versions contained garbage other than integers in their
                           # version numbers
            revision, version = (0, 0) # Too old for us
        return revision * 10 + version, ''
    
    def ReasonCannotSetup_GivenInstallationPath(self):
        """Return the reason the given Perl installation cannot
        be setup for WS, if there is an anticipated reason."""
        version, error = self._getPerlInfo()
        if error:
            return error

        if version < 56:
            return 'The is no Web Services support for Perl '\
                   'versions below 5.6. For information on how '\
                   'to upgrade to ActivePerl 5.6 or later, see '\
                   '"http://www.activestate.com/Products/ActivePerl/".'
        
        # ii. Ensure this is a distribution type that we know how to
        #     setup. For Perl, this means that there is a PPM.
        if not os.path.isfile(self.ppm):
            return "No PPM could be found in your Perl "\
                   "installation (%s). This probably means that you "\
                   "are using a Perl installation other than "\
                   "ActivePerl. PPM, which comes with "\
                   "ActivePerl, is required by this Setup to "\
                   "install the required Web Services modules. "\
                   "You can manually install these components from "\
                   '"http://aspn.activestate.com/ASPN/WebServices/SWSAPI/perltut".'\
                   % (self.installationPath)
        if not os.access(self.ppm, os.X_OK):
            return "The PPM script (%s) in your ActivePerl "\
                   "installation (%s) is not executable."\
                   % (self.ppm, installationPath)

        # i. Ensure have write access to the site-packages dir.
        #       <installationPath>/site/lib/ppm.xml   (windows)
        #       <installationPath>/lib/site_perl/<ver>/ppm.xml (linux)
        #    where <ver> is "5.6.0" or "5.6.1" or something like that,
        #    is used as a landmark for this.

        if sys.platform.startswith("win"):
            ppmPath = os.path.join(self.installationPath,
                                  "site", "lib")
            ppmXml = os.path.join(ppmPath, "ppm.xml")
        else:
            perlInfoEx = components.classes["@activestate.com/koPerlInfoEx;1"]\
               .createInstance(components.interfaces.koIAppInfoEx)
            perlInfoEx.installationPath = self.installationPath
            ppmPath = os.path.join(self.installationPath,
                                  "lib", "site_perl", perlInfoEx.version)
            ppmXml = os.path.join(ppmPath, "ppm.xml")
        
        # ppm.xml is only created after the first run of ppm, if the
        # user has never run ppm, bummer, see if we have write access
        # to the directory it will end up in.
        if not os.path.exists(ppmXml):
            ppmXml = ppmPath
        
        if not os.access(ppmXml, os.W_OK):
            return "Komodo does not have the necessary write "\
                   "permissions to your Perl installation (%s) to "\
                   "install modules. This probably means that your "\
                   "Perl installation was installed by the "\
                   "'root/Administrator' user." % self.installationPath
        
        return None

    def DescribeSetupProcedure(self):
        perlWSModule = self._globalPrefSvc.prefs\
                       .getStringPref("perlWSModule")
        return "Komodo will install or upgrade the %s "\
               "module (using PPM, the Programmer's "\
               "Package Manager) into the Perl installation you "\
               "selected (%s)."\
               % (perlWSModule, self.installationPath)

    def _InstallOrUpgradeModule(self, module):
        version, error = self._getPerlInfo()
        if error:
            return 0, error

        # build the command
        if sys.platform.startswith("win"):
            # Note: Because of a poor implementation of ppm.bat this must
            # be run in the "ppm.bat" directory on Windows to ensure
            # that the right Perl interpreter is called.
            queryCommand = "%s query \"^%s$\""\
                           % (self.ppm, module)
        else:
            queryCommand = "%s query %s" % (self.ppm, module)
        installCommand = "%s install %s" % (self.ppm, module)
        if version < 58:
            upgradeCommand = "%s verify --upgrade %s" % (self.ppm,
                                                     module)
        else:
            upgradeCommand = "%s upgrade %s"\
                         % (self.ppm, module)
        
        log = ""
        # Determine if the module module is already installed.
        #   empty output => not installed
        log += "> %s\n" % queryCommand
        env = koprocessutils.getUserEnv()
        p = process.ProcessOpen(queryCommand, env=env, cwd=os.path.dirname(self.ppm))
        output = p.stdout.read() + '\n' + p.stderr.read()
        p.close()
        retval = p.wait()
        log += output + "\n\n"
        if retval:
            return 0, log # failed
        
        if version < 58:
            doUpgrade = output.strip()
        else:
            doUpgrade = output.find('No matches') == -1
        if doUpgrade:
            command = upgradeCommand
        else:
            command = installCommand
        
        # Install or upgrade as appropriate. (XXX Perhaps user doesn't
        # always want to upgrade: should update dialog language just say
        # "install or upgrade"?)
        log += "> %s\n" % command
        p = process.ProcessOpen(command, env=env, cwd=os.path.dirname(self.ppm))
        output = p.stdout.read() + '\n' + p.stderr.read()
        p.close()
        retval = p.wait()
        log += output + "\n\n"
        if retval:
            return 0, log # failed
        # Scrape output for "^ERROR:..."
        errorRe = re.compile("^ERROR: (?P<msg>.*)$", re.MULTILINE)
        errorMatch = errorRe.search(output)
        if errorMatch:
            return 0, log
        
        return 1, log

    def Setup(self):
        """Install the necessary modules for WebService support.
        Required modules: whatever the pref 'perlWSModule' is set to
        Return whether successful (boolean) and a log string.
        """
        # note that user has run this setup
        globalPrefSvc = components.classes["@activestate.com/koPrefService;1"]\
                        .getService(components.interfaces.koIPrefService)
        globalPrefSvc.prefs.setBooleanPref(
            "ranWebServicesSetupWizard_ForPerl", 1)
        
        # install/upgrade each of the required modules
        perlWSModule = self._globalPrefSvc.prefs\
                       .getStringPref("perlWSModule")
        completeLog = ""
        success, log = self._InstallOrUpgradeModule(perlWSModule)
        completeLog += log
        if not success:
            return 0, completeLog

        return 1, completeLog

    def DescribeManualPostSetupSteps(self):
        return ""


class KoPHPWSSetupService:
    """Service for setting up a PHP installation for Web Services."""
    _com_interfaces_ = [components.interfaces.koIPHPWSSetupService]
    _reg_clsid_ = "cd4aee03-b1cf-48d6-aa52-4f15bbd32e65"
    _reg_contractid_ = "@activestate.com/koPHPWSSetupService;1"
    _reg_desc_ = "PHP Web Service Setup Service"

    def __init__(self):
        self.installationPath = None
        self.moduleDir = None

    def SetInstallationPath(self, installationPath):
        self.installationPath = installationPath

    def ReasonCannotSetup_GivenInstallationPath(self):
        """Return the reason the given PHP installation cannot
        be setup for WS, if there is an anticipated reason.
        """
        # Ensure the installation path is valid.
        if not self.installationPath:
            return "No PHP installation was specified."
        if not os.path.isdir(self.installationPath):
            return "The given PHP installation does not exist: "\
                   "'%s'. You must specify and existing PHP "\
                   "installation."\
                   % self.installationPath
        if not sys.platform.startswith("win"):
            if not os.path.isfile(os.path.join(self.installationPath,
                                               "bin", "php")):
                return "There is no PHP installation at the "\
                       "given location: '%s'." % self.installationPath

        phpInfoEx = components.classes["@activestate.com/koPHPInfoEx;1"]\
           .createInstance(components.interfaces.koIAppInfoEx)

        phpInfoEx.installationPath = self.installationPath
        if not phpInfoEx.valid_version:
            return 'Komodo cannot automatically install the '\
                   'Web Services modules for PHP versions below '\
                   '4.0.5. For information on how to upgrade to '\
                   'PHP 4.0.5 or later, see "http://aspn.activestate.com/ASPN/PHP".'

        return None
    
    def SetModuleDir(self, moduleDir):
        self.moduleDir = moduleDir

    def ReasonCannotSetup_GivenModuleDir(self):
        """Return the reason that the required PHP Web Services modules
        cannot be installed to the given module dir, if there is an
        anticipated reason.
        """
        # Ensure that a module dir was specified.
        if not self.moduleDir:
            return "No module directory was specified."
        # Ensure that the module dir either does not exist or is an existing
        # DIRECTORY.        
        if os.path.exists(self.moduleDir) \
           and not os.path.isdir(self.moduleDir):
            return "'%s' is not a directory. You must specify a module "\
                   "directory." % self.moduleDir
        # Ensure that the user has write access to this directory (or to
        # whatever parent directory actually exists).
        deepestDirThatExists = self.moduleDir
        previousCycle = None
        while 1:
            if not deepestDirThatExists \
               or deepestDirThatExists == previousCycle:
                return "'%s' is an invalid module directory. No parent "\
                       "directory exists." % self.moduleDir
            if os.path.isdir(deepestDirThatExists):
                break
            else:
                previousCycle = deepestDirThatExists
                deepestDirThatExists = os.path.dirname(deepestDirThatExists)
        if not os.access(deepestDirThatExists, os.W_OK):
            return "You do not have write access to '%s'. Please select a "\
                   "different module directory." % deepestDirThatExists
        return None

    def DescribeSetupProcedure(self):
        if os.path.isdir(self.moduleDir):
            return "Komodo will install the 'webservice' and 'soapx4' "\
                   "modules into '%s' for use by "\
                   "the PHP installation you selected (%s)."\
                   % (self.moduleDir, self.installationPath)
        else:
            return "Komodo will create the '%s' directory and install the "\
                   "'webservice' and 'soapx4' modules into it for use by "\
                   "the PHP installation you selected (%s)."\
                   % (self.moduleDir, self.installationPath)

    def Setup(self):
        # create the module dir, if necessary
        log = ""
        if not os.path.isdir(self.moduleDir):
            log += "Created '%s'.\n" % self.moduleDir
        try:
            _mkdir(self.moduleDir)
        except OSError, e:
            return 0, str(e) + "\n"
        
        # copy over the files in <komodoInstallDir>/php/WebService/...
        koDirs = components.classes["@activestate.com/koDirs;1"]\
                 .getService(components.interfaces.koIDirs)
        srcDir = os.path.join(koDirs.installDir, "php", "WebService")
        try:
            _copy(srcDir, self.moduleDir)
        except OSError, e:
            return 0, log + str(e) + "\n"
            
        log += "The Web service support modules from '%s' were "\
               "installed in '%s'." % (srcDir, self.moduleDir)
        return 1, log
        
    def DescribeManualPostSetupSteps(self):
        phpInfo = components.classes["@activestate.com/koPHPInfoEx;1"]\
                  .createInstance(components.interfaces.koIPHPInfoEx)
        phpInfo.installationPath = self.installationPath
        includePath = phpInfo.GetIncludePathArray()
        if self.moduleDir not in includePath:
            return "To complete the Setup you must manually add '%s' to "\
                   "your PHP include path. This can be done by editting "\
                   "the 'include_path' variable in '%s'."\
                   % (self.moduleDir, phpInfo.cfg_file_path)


class KoPythonWSSetupService:
    """Service for setting up a Python installation for Web Services."""
    _com_interfaces_ = [components.interfaces.koIWSSetupService]
    _reg_clsid_ = "bc98bc35-2b2b-4301-8545-4ef24fa89d2c"
    _reg_contractid_ = "@activestate.com/koPythonWSSetupService;1"
    _reg_desc_ = "Python Web Service Setup Service"

    def SetInstallationPath(self, installationPath):
        self.installationPath = installationPath
        # determine which pyppm to invoke
        if sys.platform.startswith("win"):
            self.pyppm = os.path.join(self.installationPath,
                                      "pyppm.bat")
        else:
            self.pyppm = os.path.join(self.installationPath, "bin",
                                      "pyppm")

    def ReasonCannotSetup_GivenInstallationPath(self):
        """Return the reason the given Python installation cannot
        be setup for WS, if there is an anticipated reason."""
        # iii. Ensure the installation path is valid.
        if not self.installationPath:
            return "No Python installation was specified."
        if not os.path.isdir(self.installationPath):
            return "The given Python installation does not exist: "\
                   "'%s'. You must specify and existing Python "\
                   "installation."\
                   % self.installationPath
        if not sys.platform.startswith("win"):
            if not os.path.isfile(os.path.join(self.installationPath,
                                               "bin", "python")):
                return "There is no Python installation at the "\
                       "given location: '%s'." % self.installationPath
        
        # ii. Ensure this is a distribution type that we know how to
        #     setup. For Python, this means that there is a PyPPM.

        pythonInfoEx = components.classes["@activestate.com/koPythonInfoEx;1"]\
           .createInstance(components.interfaces.koIAppInfoEx)

        pythonInfoEx.installationPath = self.installationPath
        self.executablePath = pythonInfoEx.executablePath

        major, minor = map(int, pythonInfoEx.version.split(".")[:2])
        if major * 10 + minor < 21:
            return 'Komodo cannot automatically install the '\
                   'Web Services modules for Python versions '\
                   '2.0 or lower. You can either upgrade to '\
                   'ActivePython 2.1 or later ("http://www.activestate.com/Products/ActivePython/") '\
                   'or manually install these components from '\
                   '"http://aspn.activestate.com/ASPN/WebServices/SWSAPI/pytut".'
        if not os.path.isfile(self.pyppm):
            return "No PyPPM could be found in your Python "\
                   "installation (%s). This probably means that you "\
                   "are using a Python installation other than "\
                   "ActivePython. PyPPM, which comes with "\
                   "ActivePython, is required by this Setup to "\
                   "install the required Web Services modules. "\
                   "You can manually install these components from "\
                   '"http://aspn.activestate.com/ASPN/WebServices/SWSAPI/pytut".'\
                   % (self.installationPath)
        if not os.access(self.pyppm, os.X_OK):
            return "The PyPPM script (%s) in your ActivePython "\
                   "installation (%s) is not executable."\
                   % (self.pyppm, installationPath)

        # i. Ensure have write access to the site-packages dir.
        #       <installationPath>/Lib/os.py   (windows)
        #       <installationPath>/lib/pythonN.N/os.py   (linux)
        #    is used as a landmark for this.
        if sys.platform.startswith("win"):
            osDotPy = os.path.join(self.installationPath, "Lib", "os.py")
        else:
            versionedDir = "python%d.%d" % (major, minor)
            osDotPy = os.path.join(self.installationPath, "lib", versionedDir, "os.py")
        if not os.access(osDotPy, os.W_OK):
            return "Komodo does not have the necessary write "\
                   "permissions to your Python installation (%s) to "\
                   "install modules. This probably means that your "\
                   "Python installation was installed by the "\
                   "'root/Administrator' user." % self.installationPath
        
        return None

    def DescribeSetupProcedure(self):
        return "Komodo will install or upgrade the PyXML and "\
               "WebService modules (using PyPPM, the Programmer's "\
               "Package Manager) into the Python installation you "\
               "selected (%s)."\
               % self.installationPath

    def Setup(self):
        """Install the necessary modules for WebService support.
        Required modules: PyXML, WebService
        Return whether successful (boolean) and a log string.
        """
        # note that user has run this setup
        globalPrefSvc = components.classes["@activestate.com/koPrefService;1"]\
                        .getService(components.interfaces.koIPrefService)
        globalPrefSvc.prefs.setBooleanPref(
            "ranWebServicesSetupWizard_ForPython", 1)
        
        # install/upgrade each of the required modules
        completeLog = ""
        success, log = self._InstallOrUpgradeModule("PyXML")
        completeLog += log
        if not success:
            return 0, completeLog
        success, log = self._InstallOrUpgradeModule("WebService")
        completeLog += log
        if not success:
            return 0, completeLog

        return 1, completeLog

    def _InstallOrUpgradeModule(self, module):
        # NOTE: Must remove PYTHONPATH and PYTHONHOME from Mozilla.exe's
        #       environment else it will conflict with the PyPPM call.
        # build the command
        if sys.platform.startswith("win"):
            # Note: Because of a poor implementation in APy20 this must
            # be run in the "pyppm.bat" directory on Windows to ensure
            # that the right Python interpreter is called.
            queryCommand = "%s query %s" % (self.pyppm, module)
            installCommand = "%s install %s" % (self.pyppm, module)
            upgradeCommand = "%s verify --upgrade %s" % (self.pyppm, module)
        else:
            pyppmCmd = "%s %s" % (self.executablePath, self.pyppm)
            queryCommand = "%s query %s" % (pyppmCmd, module)
            installCommand = "%s install %s" % (pyppmCmd, module)
            upgradeCommand = "%s verify --upgrade %s" % (self.pyppm, module)
        
        log = ""
        # Determine if the module module is already installed.
        #   empty output => not installed
        log += "> %s\n" % queryCommand
        env = koprocessutils.getUserEnv()
        if 'PYTHONPATH' in env:
            del env['PYTHONPATH']
        if 'PYTHONHOME' in env:
            del env['PYTHONHOME']
        p = process.ProcessOpen(queryCommand, env=env, cwd=os.path.dirname(self.pyppm))
        output = p.stdout.read() + '\n' + p.stderr.read()
        p.close()
        retval = p.wait()
        log += output + "\n\n"
        if retval:
            return 0, log # failed
        if output.strip():
            command = upgradeCommand
        else:
            command = installCommand
        
        # Install or upgrade as appropriate. (XXX Perhaps user doesn't
        # always want to upgrade: should update dialog language just say
        # "install or upgrade"?)
        log += "> %s\n" % command
        env = koprocessutils.getUserEnv()
        p = process.ProcessOpen(command, env=env, cwd=os.path.dirname(self.pyppm))
        output = p.stdout.read() + '\n' + p.stderr.read()
        p.close()
        retval = p.wait()
        log += output + "\n\n"
        if retval:
            return 0, log # failed
        # Scrape output for "^ERROR:..."
        errorRe = re.compile("^ERROR: (?P<msg>.*)$", re.MULTILINE)
        errorMatch = errorRe.search(output)
        if errorMatch:
            return 0, log
        
        return 1, log

    def DescribeManualPostSetupSteps(self):
        return ""



#---- self test code

if __name__ == "__main__":
    for lang in ["Python", "Perl", "PHP"]:
        langSvc = components.classes["@activestate.com/ko"+lang+"InfoEx;1"]\
                  .getService(components.interfaces.koIAppInfoEx)
        wsSetupSvc = components.classes["@activestate.com/ko" +
                                        lang + "WSSetupService;1"]\
                     .getService()
        
        # select the installation to setup
        installationPath = langSvc.installationPath
        wsSetupSvc.SetInstallationPath(installationPath)
        reason = wsSetupSvc.ReasonCannotSetup_GivenInstallationPath()
        if reason:
            raise "Cannot setup %s for WS because: %s" % (lang, reason)
        
        # select the PHP module dir
        if lang == "PHP":
            import tempfile
            wsSetupSvc.SetModuleDir(tempfile.mktemp())
            reason = wsSetupSvc.ReasonCannotSetup_GivenModuleDir()
            if reason:
                raise "Cannot setup %s for WS because: %s" % (lang, reason)
            
        # set it up
        desc = wsSetupSvc.DescribeSetupProcedure()
        print "To setup your %s installation (at %s) for WS:\n"\
              % (lang, installationPath)
        print desc
        answer = raw_input("Do you want to do this? [Y/n] ")
        if answer in ("", "y", "Y", "Yes", "yes", "YES"):
            succeeded = wsSetupSvc.Setup()
            if succeeded:
                print "Setup succeeded!"
                manualSteps = wsSetupSvc.DescribeManualPostSetupSteps()
                if manualSteps:
                    print "Manual Steps:", manualSteps
            else:
                print "Setup failed!"
        else:
            print "Skipping setup for %s." % lang

