"""Base classes for languages and language services"""

from xpcom import components
import cgi, re, types, os, uriparse, eollib
import string

import timeline
import logging
log = logging.getLogger('koLanguageServiceBase')
indentlog = logging.getLogger('koLanguageServiceBase.indenting')
#indentlog.setLevel(logging.DEBUG)
sci_constants = components.interfaces.ISciMoz
numWordLists=6 # from ScintillaBase.h

class KoDebuggerLanguageService:
    _com_interfaces_ = [components.interfaces.koIDebuggerLanguageService]

    _language = None
    _engine = None

    def __init__(self, language):
        self._language = language

        
    def get_engine(self):
        if self._engine is None:
            debugger = components.classes["@activestate.com/debugger;1"] \
                       .getService(components.interfaces.koIDebugger)
            self._engine = debugger.addEngine(self._language)
            
            handler = components \
                  .classes["@mozilla.org/network/protocol;1?name=kodebugger"] \
                  .getService(components.interfaces.koIDebuggerProtocolHandler)
            handler.registerEngine(self._engine, self._language)
        return self._engine

class KoCompletionLanguageService:
    _com_interfaces_ = [components.interfaces.koICompletionLanguageService]
    triggers = ''
    completionSeparator = ord('\n')
    triggersCallTip = None
    
    def __init__(self):
        pass
    
    def scanFileCompletion(self, filename):
        pass
    def scanBufferCompletion(self, scimoz, filename):
        pass

class KoLexerLanguageService:
    _com_interfaces_ = [components.interfaces.koILexerLanguageService]

    def __init__(self):
        self._lexer = None
        self._keywords = []
        for i in range(numWordLists):
            self._keywords.append([])
        self._properties = {}
        self.supportsFolding = 0

    def setLexer(self, sciLexer):
        self._lexer = sciLexer

    def setKeywords(self, keywordSet, keywords):
        assert keywordSet >= 0 and keywordSet < numWordLists
        self._keywords[keywordSet] = keywords

    def setProperty(self, property, value):
        self._properties[property] = value

    def setCurrent(self, scimoz):
        if self._lexer is None:
            scimoz.lexer = scimoz.SCLEX_NULL
        else:
            scimoz.lexer = self._lexer

        for i in range(len(self._keywords)):
            scimoz.setKeyWords(i, " ".join(self._keywords[i]))

        for prop in self._properties.keys():
            scimoz.setProperty(prop, self._properties[prop])
    
class KoLinterLanguageService:
    """Just a way to get the contract IDs used by koLintService"""
    _com_interfaces_ = [components.interfaces.koILinterLanguageService]

    def __init__(self, linterCID):
        # this argument-laden constructor should go away --
        # not portable to other languages.
        self._linterCID = linterCID

    def get_linterCID(self):
        return self._linterCID

#---- base commenting language service classes

class KoCommenterLanguageService:
    _com_interfaces_ = [components.interfaces.koICommenterLanguageService]

    def __init__(self, delimiterInfo):
        """'delimiterInfo' is a dictionary describing how comments are
        specified in the given language. The used keys are as follows:
            line:   an ordered list of prefixes to remove for line 
                    uncommenting, the first one is prefixed for line
                    commenting
            block:  an ordered list of (prefix, suffix) pair tuples
                    for block commenting and uncommenting
        If 'line' is defined then the language supports line commenting --
        Python and Perl. If 'block' is defined then the language supports
        block commenting -- XML. A language may support both -- C++. Which
        style is applied depends on the current selection in scimoz.
        """
        self.delimiterInfo = delimiterInfo

    def _commentLines(self, scimoz, startIndex, endIndex):
        """Comment the indexed lines."""
        xOffset = scimoz.xOffset
        selStart = scimoz.selectionStart
        selEnd = scimoz.selectionEnd
        if selStart != selEnd and scimoz.getColumn(selEnd) == 0:
            # if at the start of a line, work from the end of the previous
            # line
            workingEndIndex = scimoz.getLineEndPosition(
                scimoz.lineFromPosition(endIndex) - 1)
        else:
            workingEndIndex = endIndex
        original = scimoz.getTextRange(startIndex, workingEndIndex)
        prefix = self.delimiterInfo["line"][0]
        if type(prefix) == types.TupleType:
            prefix, suffix = prefix
        else:
            suffix = ''

        # For easier line terminator handling turn this:
        #      original = 'asdf\nqwer\r\nzxcv\rasdf'
        # in this:
        #      originalLines = [('asdf', '\n'), ('qwer', '\r\n'),
        #                       ('zxcv', '\r'), ('asdf', '')]
        lines = re.split("(\r\n|\n|\r)", original)
        originalLines = []
        for i in range(0, len(lines), 2):
            try:
                originalLines.append( (lines[i], lines[i+1]) )
            except IndexError:
                originalLines.append( (lines[i], '') )

        # Find the maximum common whitespace offset of all the lines (comment
        # delimiters will be indented to this level). Skip blank lines
        # when determining this.
        linesToGetIndent = [line[0] for line in originalLines if line[0].strip()]
        indentRe = re.compile("^(?P<indent>\s+).*$")
        if len(linesToGetIndent) > 0:
            indentMatch = indentRe.search(linesToGetIndent[0])
            if indentMatch:
                commonIndent = indentMatch.group("indent")
            else:
                commonIndent = ""
        else:
            commonIndent = ""
        for line in linesToGetIndent:
            indentMatch = indentRe.search(line)
            if indentMatch:
                indent = indentMatch.group("indent")
                while commonIndent:
                    if indent.startswith(commonIndent):
                        break
                    commonIndent = commonIndent[:-1]
                else:
                    break
            else:
                commonIndent = ""
                break

        # make the replacments
        replacementLines = []
        for line in originalLines:
            r = commonIndent + prefix + line[0][len(commonIndent):] \
                + suffix + line[1]
            replacementLines.append(r)
        replacement = "".join(replacementLines)
        #log.debug("line comment: '%s' -> '%s'" % (original, replacement))

        if original != replacement:
            scimoz.hideSelection(1)
            startIndexLine = scimoz.lineFromPosition(startIndex)
            endIndexLine = scimoz.lineFromPosition(endIndex)
            selStartColumn = scimoz.getColumn(selStart)

            # apply the commenting change
            scimoz.targetStart = startIndex
            scimoz.targetEnd = workingEndIndex
            scimoz.replaceTarget(len(replacement), replacement)

            # restore the selection and cursor position
            if selStart != selEnd:
                scimoz.selectionStart = scimoz.positionFromLine(startIndexLine)
                if endIndex == workingEndIndex:
                    scimoz.selectionEnd = scimoz.getLineEndPosition(
                        endIndexLine)
                else:
                    scimoz.selectionEnd = scimoz.positionFromLine(
                        endIndexLine)
            else:
                if selStartColumn <= len(commonIndent):
                    scimoz.selectionStart = selStart
                    scimoz.selectionEnd = selEnd
                else:
                    scimoz.selectionStart = selStart + len(prefix)
                    scimoz.selectionEnd = selEnd + len(prefix)
            scimoz.hideSelection(0)
            scimoz.xOffset = xOffset

    def _commentBlock(self, scimoz, startIndex, endIndex):
        """Comment the indexed block."""
        xOffset = scimoz.xOffset
        original = scimoz.getTextRange(startIndex, endIndex)
        prefix, suffix = self.delimiterInfo["block"][0]

        replacement = prefix + original + suffix
        #log.debug("block comment: '%s' -> '%s'" % (original, replacement))

        if original != replacement:
            scimoz.hideSelection(1)
            selStart = scimoz.selectionStart
            selEnd = scimoz.selectionEnd

            # apply the commenting change
            scimoz.targetStart = startIndex
            scimoz.targetEnd = endIndex
            scimoz.replaceTarget(len(replacement), replacement)

            # restore the selection and cursor position
            scimoz.selectionStart = selStart + len(prefix)
            scimoz.selectionEnd = selEnd + len(prefix)
            scimoz.hideSelection(0)
            scimoz.xOffset = xOffset

    def _uncommentLines(self, scimoz, startIndex, endIndex):
        xOffset = scimoz.xOffset
        selStart = scimoz.selectionStart
        selEnd = scimoz.selectionEnd
        if selStart != selEnd and scimoz.getColumn(selEnd) == 0:
            # if at the start of a line, work from the end of the previous
            # line
            workingEndIndex = scimoz.getLineEndPosition(
                scimoz.lineFromPosition(endIndex) - 1)
        else:
            workingEndIndex = endIndex
        original = scimoz.getTextRange(startIndex, workingEndIndex)
        prefixes = self.delimiterInfo["line"]

        # For easier line terminator handling turn this:
        #      original = 'asdf\nqwer\r\nzxcv\rasdf'
        # in this:
        #      originalLines = [('asdf', '\n'), ('qwer', '\r\n'),
        #                       ('zxcv', '\r'), ('asdf', '')]
        lines = re.split("(\r\n|\n|\r)", original)
        originalLines = []
        for i in range(0, len(lines), 2):
            try:
                originalLines.append( (lines[i], lines[i+1]) )
            except IndexError:
                originalLines.append( (lines[i], '') )

        replacementLines = []
        lastIndent = lastPrefix = ""  # for restoring the position below
        for line in originalLines:
            for prefix in prefixes:
                if type(prefix) == types.TupleType: # block comments
                    prefix, suffix = prefix
                    commentRe = re.compile("^(\s*)(%s)(.*)(%s)(\s*)$"
                                           % (re.escape(prefix),
                                              re.escape(suffix)))
                else: # line comments
                    suffix = None
                    commentRe = re.compile("^(\s*)(%s)(.*)()()$" % re.escape(prefix))
                commentMatch = commentRe.search(line[0])
                if commentMatch:
                    if not (commentMatch.group(3) + commentMatch.group(5)):
                        # If there is no line content after the comment
                        # delimiters then just replace the whole line with
                        # an empty line, i.e. strip out the indent as well.
                        # This allows a comment/uncomment cycle on a blank
                        # line (when part of a group of lines) to not change
                        # the line.
                        replacementLines.append(line[1])
                    else:
                        replacementLines.append(commentMatch.group(1) +
                                                commentMatch.group(3) +
                                                commentMatch.group(5) +
                                                line[1])
                    lastIndent = commentMatch.group(1)
                    lastPrefix = commentMatch.group(2)
                    break
            else:
                replacementLines.append(line[0] + line[1])
        replacement = "".join(replacementLines)
        #log.debug("line uncomment: '%s' -> '%s'" % (original, replacement))

        if original != replacement:
            scimoz.hideSelection(1)
            startIndexLine = scimoz.lineFromPosition(startIndex)
            endIndexLine = scimoz.lineFromPosition(endIndex)
            selStartColumn = scimoz.getColumn(selStart)

            # apply the commenting change
            scimoz.targetStart = startIndex
            scimoz.targetEnd = workingEndIndex
            scimoz.replaceTarget(len(replacement), replacement)

            # restore the selection and cursor position
            if selStart != selEnd:
                scimoz.selectionStart = scimoz.positionFromLine(startIndexLine)
                if endIndex == workingEndIndex:
                    scimoz.selectionEnd = scimoz.getLineEndPosition(
                        endIndexLine)
                else:
                    scimoz.selectionEnd = scimoz.positionFromLine(
                        endIndexLine)
            else:
                if selStartColumn <= len(lastIndent):
                    scimoz.selectionStart = selStart
                    scimoz.selectionEnd = selEnd
                else:
                    scimoz.selectionStart = selStart - len(lastPrefix)
                    scimoz.selectionEnd = selEnd - len(lastPrefix)
            scimoz.hideSelection(0)
            scimoz.xOffset = xOffset

    def _uncommentBlock(self, scimoz, startIndex, endIndex):
        xOffset = scimoz.xOffset
        selStart = scimoz.selectionStart
        selEnd = scimoz.selectionEnd
        delimiters = self.delimiterInfo["block"]

        for prefix, suffix in delimiters:
            pStart, pEnd = startIndex-len(prefix), startIndex
            sStart, sEnd = endIndex, endIndex+len(suffix)
            p = scimoz.getTextRange(pStart, pEnd)
            s = scimoz.getTextRange(sStart, sEnd)
            #log.debug("block uncomment delimiter match if (%s==%s  "\
            #         "and  %s==%s)" % (p, prefix, s, suffix))
            if (p == prefix and s == suffix):
                scimoz.hideSelection(1)
                # remove the existing prefix and suffix (suffix first to get
                # indeces correct)
                scimoz.beginUndoAction()
                try:
                    scimoz.targetStart = sStart
                    scimoz.targetEnd = sEnd
                    scimoz.replaceTarget(0, "")
                    scimoz.targetStart = pStart
                    scimoz.targetEnd = pEnd
                    scimoz.replaceTarget(0, "")
                finally:
                    scimoz.endUndoAction()
                # restore the selection and cursor position
                scimoz.selectionStart = selStart - len(prefix)
                scimoz.selectionEnd = selEnd - len(prefix)
                scimoz.hideSelection(0)
        scimoz.xOffset = xOffset

    def _determineMethodAndDispatch(self, scimoz, workers):
        selStart = scimoz.selectionStart
        selEnd = scimoz.selectionEnd
        selStartColumn = scimoz.getColumn(selStart)
        selStartLine = scimoz.lineFromPosition(selStart)
        selStartLineEndPosition = scimoz.getLineEndPosition(selStartLine)
        selEndColumn = scimoz.getColumn(selEnd)
        selEndLine = scimoz.lineFromPosition(selEnd)
        selEndLineEndPosition = scimoz.getLineEndPosition(selEndLine)

        # determine preferred commenting method (if the selection is _within_
        # a line then block commenting is preferred)
        if selStart != selEnd:
            if (selStartColumn == 0 or selStart == selStartLineEndPosition) \
               and (selEndColumn == 0 or selEnd == selEndLineEndPosition):
                preferBlockCommenting = 0
            else:
                preferBlockCommenting = 1
        else:
            preferBlockCommenting = 0

        # do the commenting/uncommenting
        if preferBlockCommenting and self.delimiterInfo.get("block", None):
            workers["block"](scimoz, selStart, selEnd)
        elif self.delimiterInfo.get("line", None):
            textStart = scimoz.positionFromLine(selStartLine)
            if selStart != selEnd and scimoz.getColumn(selEnd) == 0:
                # if at the start of a line, then do NOT include that line
                textEnd = selEnd
            else:
                textEnd = scimoz.getLineEndPosition(selEndLine)
            workers["line"](scimoz, textStart, textEnd)
        elif self.delimiterInfo.get("block", None):
            # This lang only has block delimiters but the selection is such
            # that it looks like line commenting should be done. Sneak
            # the block delimiter tuple in as line delimiters.
            self.delimiterInfo["line"] = self.delimiterInfo["block"]
            textStart = scimoz.positionFromLine(selStartLine)
            if selStart != selEnd and scimoz.getColumn(selEnd) == 0:
                # if at the start of a line, then do NOT include that line
                textEnd = selEnd
            else:
                textEnd = scimoz.getLineEndPosition(selEndLine)
            workers["line"](scimoz, textStart, textEnd)
        else:
            log.warn("Could not comment code because no appropriate "\
                       "comment delimiter info exists for the current "\
                       "language.\n")

    def comment(self, scimoz):
        """Comment the current line or current selection."""
        commenters = {
            "line"  : self._commentLines,
            "block" : self._commentBlock
        }
        self._determineMethodAndDispatch(scimoz, commenters)

    def uncomment(self, scimoz):
        """Uncomment the current line or current selection."""
        uncommenters = {
            "line"  : self._uncommentLines,
            "block" : self._uncommentBlock
        }
        self._determineMethodAndDispatch(scimoz, uncommenters)



class KoWSInsertLanguageService:
    _com_interfaces_ = [components.interfaces.koIWSInsertLanguageService]

    def __init__(self, inserter, creationTemplate, module, useMD5):
        self._inserter = inserter
        self._creationTemplate = creationTemplate
        self._module = module
        self._useMD5 = useMD5

    def getScintillaLine(self, scin, lineno):
        (length, text) = scin.getLine(lineno)
        return text[:length]
    
    def insert(self, scin, position, objectName, url):
        lineno = scin.lineFromPosition(position)
        eolformat = eollib.scimozEOL2eol[scin.eOLMode]

        if self._useMD5:
            import WShelper
            md5hash = WShelper.Get_MD5_Hash(url)
            statement = self._creationTemplate % (objectName, url, md5hash)
        else:
            statement = self._creationTemplate % (objectName, url)
            
        insertions = self._inserter.autoInsert(self._splitlines_with_trailing(scin.text),
                                               lineno,
                                               [statement],
                                               [self._module])
        linesInserted = 0
        for (insert_lineno, insert_lines) in insertions:
            scin_lineno = insert_lineno + linesInserted
            pos = scin.getLineEndPosition(scin_lineno)
            insert_text = eollib.eolMappings[eolformat].join(insert_lines)

            lineText = self.getScintillaLine(scin, scin_lineno)
            insertedLineCount = len(insert_lines)
            insert_text = eollib.eolMappings[eolformat] + insert_text
            scin.targetStart = pos
            scin.targetEnd = pos
            scin.replaceTarget(len(insert_text), insert_text)

            # set insertion point to end of the new content
            newPos = scin.getLineEndPosition(scin_lineno + insertedLineCount)
            scin.gotoPos(newPos)
            linesInserted += insertedLineCount
            
        line = scin.lineFromPosition(scin.currentPos)
        newPos = scin.positionFromLine(line + 1)
        scin.gotoPos(newPos)

    def _splitlines_with_trailing(self, s):

        """Same as splitlines() except we treat a trailing newline as
        an empty line.

        """
        lines = s.splitlines()
        if not s or s[-1] in ("\n", "\r"):
            lines.append('')
        return lines


class KoLanguageBase:
    _com_interfaces_ = [components.interfaces.koILanguage]

    _lexer = None
    _linter = None
    _debugger = None
    _style = None
    _methodcompletion = None
    _commenter = None
    _inserter = None
    _interpreter = None
    styleBits = 5
    stylingBitsMask = None # set by setter below
    indicatorBits = 2
    supportsBraceHighlighting = 1
    sample = '' # used in the fonts & colors dialog
    downloadURL = '' # location to download the language
    remoteDebuggerSupport = 0 # does this language support remote debugging?
    variableIndicators = ''
    supportsSmartIndent = "text"
    
    # the following four are used by the getBraceIndentStyle call
    _indent_chars = "{}" # may be overriden by the language implementation
    _indent_styles = [0,32,10]  # may be overriden by the language implementation
    _indent_open_chars = "{:" # may be overriden by the language implementation
    _indent_close_chars = "}" # may be overriden by the language implementation
    _indent_open_styles = [10]  # may be overriden by the language implementation
    _lineup_chars = "()[]{}" # may be overriden by the language implementation
    _indent_close_chars = "}" # may be overriden by the language implementation
    _indent_close_styles = [10] # may be overriden by the language implementation
    _lineup_open_chars = "([" # may be overriden by the language implementation
    _lineup_close_chars = ")]" # may be overriden by the language implementation
    _lineup_close_styles = [10]  # may be overriden by the language implementation
    _lineup_styles = [10]  # may be overriden by the language implementation
    _comment_styles = [] # should be overriden by the language implementation
    _dedenting_statements = [] # should be overriden by the language implementation
    _indenting_statements = [] # should be overriden by the language implementation
    _line_comment_styles = []
    _block_comment_styles = []
    commentDelimiterInfo = {}
    
    _svcdict = {
        components.interfaces.koILexerLanguageService: 'get_lexer',
        components.interfaces.koILinterLanguageService: 'get_linter',
        components.interfaces.koIDebuggerLanguageService: 'get_debugger',
        components.interfaces.koICompletionLanguageService: 'get_methodcompletion',
        components.interfaces.koICommenterLanguageService: 'get_commenter',
        components.interfaces.koIAppInfoEx: 'get_interpreter',
        components.interfaces.koIWSInsertLanguageService: 'get_inserter'
        }
    namedBlockRE = ''
    namedBlockDescription = ''

    def __init__(self):
        self.stylingBitsMask = 0    
        for bit in range(self.styleBits):
            self.stylingBitsMask <<= 1
            self.stylingBitsMask |= 1

    def getLanguageService(self, iid):
        if self._svcdict.has_key(iid):
            return getattr(self, self._svcdict[iid])()
        return None

    # override these getters if you need to override the language services

    def get_lexer(self):
        if self._lexer is None:
            self._lexer = KoLexerLanguageService()
            self._lexer.setLexer(components.interfaces.ISciMoz.SCLEX_NULL)
        return self._lexer

    def get_linter(self):
        return self._linter

    def get_debugger(self):
        return self._debugger

    def get_debuggable(self):
        if self.get_debugger():
            return 1
        else:
            return 0
        
    def get_foldable(self):
        lexer = self.get_lexer()
        return lexer.supportsFolding

    def get_style(self):
        if self._style is None:
            self._style = KoStyleLanguageService()
        return self._style

    def get_methodcompletion(self):
        return self._methodcompletion

    def get_interpreter(self):
        return self._interpreter
    
    def get_commenter(self):
        return self._commenter

    def get_inserter(self):
        return self._inserter
    
    def getEncodingWarning(self, encoding):
        if not encoding.encoding_info.ascii_superset:
            return '"%s" is not a recommended encoding for %s. Using ASCII or an ASCII superset is recommended.'\
                    % (encoding.friendly_encoding_name, self.name)
        elif encoding.use_byte_order_marker:
            return 'Including a signature (BOM) is not recommended for %s.' % self.name
        else: # It's all good
            return ''

    def getBraceIndentStyle(self, ch, style):
        style = style & self.stylingBitsMask
        if ch in self._indent_chars and style in self._indent_styles:
            return components.interfaces.koILanguage.INDENT_BRACE
        
        if ch in self._lineup_chars and style in self._lineup_styles:
            return components.interfaces.koILanguage.LINEUP_BRACE
        return 0        

    def guessIndentation(self, scimoz, tabWidth):
        indent, usesTabs = _findIndent(scimoz,
                                       self.stylingBitsMask,
                                       self._indent_open_chars,
                                       self._indent_open_styles, self._comment_styles,
                                       tabWidth)
        return indent, usesTabs


    def _shouldIndent(self, scimoz, pos):
        # The first phase is to look for indenting statements.
        # these can be followed on the current line by comments.
        #
        # look past any comments to the left until hit either beginning
        # of line or a character.
        # if that character is an indenting brace, then indent.
        timeline.enter('_shouldIndent')
        curLineNo = scimoz.lineFromPosition(pos)
        lineStart = scimoz.positionFromLine(curLineNo)
        data = scimoz.getStyledText(lineStart, pos+1)
        for p in range(pos-1, lineStart-1, -1):
            char = data[(p-lineStart)*2]
            style = ord(data[(p-lineStart)*2+1]) & self.stylingBitsMask
            #indentlog.debug("char = %s, style = %d", char, style)
            #indentlog.debug("indent_open_chars = %r, indent_open_styles = %r", self._indent_open_chars,
            #          self._indent_open_styles)
            if style in self._comment_styles:
                indentlog.debug("skipping comment character")
                continue
            elif char in ' \t':
                continue
            elif (char in self._indent_open_chars and
                  style in self._indent_open_styles):
                # need to find the beginning of the logical statement.
                lineNo = self._statementStartingLineFromPos(scimoz, p)
                indentlog.debug("we've decided that the statement starting from line %d actually starts at line %d" % (curLineNo, lineNo))
                currentIndentWidth = self._getIndentWidthForLine(scimoz, lineNo)
                indent = scimoz.indent
                if indent == 0:
                    indentlog.error('indent was 0, defaulting to 8')
                    indent = 8
                nextIndentWidth = (divmod(currentIndentWidth, indent)[0] + 1) * indent
                return self._makeIndentFromWidth(scimoz, nextIndentWidth)
            break
        timeline.leave('_shouldIndent')
        return None

    def _shouldLineUp(self, scimoz, pos):
        # first we look for an unmatched 'line-up' brace, going back until
        # either we've gone back 100 lines, or we get past the indent
        # of the current line

        timeline.enter('_shouldLineUp')
        try:
            curLineNo = scimoz.lineFromPosition(pos)
            minIndent = self._getIndentWidthForLine(scimoz, curLineNo)
            startLineNo = max(curLineNo-100, 0)
            startPos = scimoz.positionFromLine(startLineNo)
            p = pos
            timeline.enter('getStyledText')
            data = scimoz.getStyledText(startPos, pos+1)
            timeline.leave('getStyledText')
            while p > startPos:
                p = p - 1
                char = data[(p-startPos)*2]
                indentlog.info("looking at char %r at position %d", char, p)
                style = ord(data[(p-startPos)*2+1]) & self.stylingBitsMask
                if char in self._lineup_close_chars:
                    # skip to the match
                    braceMatch = scimoz.braceMatch(p)
                    if braceMatch == -1:
                        break
                    p = braceMatch
                    char = data[(p-startPos)*2]
                    style = ord(data[(p-startPos)*2+1]) & self.stylingBitsMask
                    lineNo = scimoz.lineFromPosition(p)
                    minIndent = self._getIndentWidthForLine(scimoz, lineNo)
                    continue # keep looking
                if char in self._lineup_open_chars and style in self._lineup_close_styles:
                    # aha!
                    # is there something to our right?
                    indentlog.info("We've got a line up open char!: %r in %r", char, self._lineup_open_chars)
                    lineNo = scimoz.lineFromPosition(p)
                    lineEnd = scimoz.getLineEndPosition(lineNo)
                    line = scimoz.getTextRange(p+1, lineEnd)
                    indentlog.info("Looking to what's to the right of us: %r", line)
                    if line.strip():
                        indentlog.info("look forward until the first non-comment non-whitespace char until EOL")
                        for p2 in range(p, min(pos, lineEnd)):
                            ch2 = data[(p2-startPos)*2]
                            if ch2 in '  \t':
                                continue
                            style = ord(data[(p2-startPos)*2+1]) & self.stylingBitsMask
                            if style in self._comment_styles:
                                continue
                            indentlog.info("we're here -- this is the column that we should use for indenting")
                            column = scimoz.getColumn(p2)+1
                            return self._makeIndentFromWidth(scimoz, column)
                    else:
                        indentlog.info("it's a whitespace line, use the indent of the opened brace + 1")
                        column = scimoz.getColumn(p)+1
                        return self._makeIndentFromWidth(scimoz, column)
                col = scimoz.getColumn(p)
                if col <= minIndent:
                    timeline.mark('breaking out')
                    indentlog.debug("breaking out of looking for line-up characters" +
                                    "because we've moved too far to the left at column %d < %d", col, minIndent)
                    return None
            timeline.mark("looked over %d/%d characters" % (pos-p, pos-startPos))
            return None
        finally:
            timeline.leave('_shouldLineUp')

    def _skipOverParentheticalStatement(self, scimoz, pos):
        # we may just have finished a multi-line statement -- those involve
        # braces:
        #  if (x and y \
        #      and z)!
        # then we need to indent with the indent of the line with
        # the matching brace

        curLineNo = scimoz.lineFromPosition(pos)
        startOfLine = scimoz.positionFromLine(curLineNo)
        p = pos
        data = scimoz.getStyledText(startOfLine, pos+2)
        while p > startOfLine:
            p = p - 1
            char = data[(p-startOfLine)*2]
            style = ord(data[(p-startOfLine)*2+1]) & self.stylingBitsMask
            # if we encounter an _opening_ brace, then we should stop.
            if char in self._lineup_open_chars and \
               style in self._lineup_close_styles:
                return pos
            # if we encounter a closing brace, we skip to its match
            # and move on.
            if char in self._lineup_close_chars and \
               style in self._lineup_close_styles:
                # skip to the match
                braceMatch = scimoz.braceMatch(p)
                if braceMatch == -1:
                    return pos
                else: # we'll use the indent of this line
                    return braceMatch
        return pos
        
    def _statementStartingLineFromPos(self, scimoz, position):
        """ This function looks for the line that starts
        the statement which encompasses the position 'position'
        """
        curLineNo = scimoz.lineFromPosition(position)

        indentlog.info("looking for statement starting line from line %d", curLineNo)

        # we'll look at most 40 lines back # should be adjustable?
        startLineNo = max(curLineNo-40, 0) # curLineNo
        startPos = scimoz.positionFromLine(startLineNo)
        p = position
        lineNo = scimoz.lineFromPosition(p)
        data = scimoz.getStyledText(startPos, p)
        minIndent = self._getIndentWidthForLine(scimoz, curLineNo)
        indentlog.info("startPos = %d, p = %d", startPos, p)
        # moving back through the buffer, look for and 'end-condition', i.e.
        # something which indicates the start of a statement.
        #  - we run out of characters (40 lines back)
        #  - we find an indenting statement
        # If we encounter a line-up brace, we skip to the matching brace
        # and keep going.

        while p > startPos:
            p = p - 1
            char = data[(p-startPos)*2]
            style = ord(data[(p-startPos)*2+1]) & self.stylingBitsMask
            #indentlog.debug("in _statementStartingLineFromPos: char = %s, style = %d", char, style)
            #indentlog.debug("indent_close_chars = %r, indent_clos_styles = %r",
            #                self._indent_close_chars, self._indent_close_styles)
            lineNo = scimoz.lineFromPosition(p)
            indentlog.debug("\np = %d (%r) - going until %d" % (p, char, startPos))
            if char in self._lineup_close_chars and \
               style in self._lineup_close_styles:
                # we found an closing brace, we want to skip back to the beginning.
                indentlog.info("found closing brace, skipping back")
                next_p = self._skipOverParentheticalStatement(scimoz, p+1)
                lineNo = scimoz.lineFromPosition(next_p)
                minIndent = self._getIndentWidthForLine(scimoz, lineNo)
                indentlog.debug("SKIPPED from %d to %d" %(p+1, next_p))
                if p == next_p:
                    indentlog.debug("SKIPPED nowhere!")
                    return lineNo
                p = next_p
                lineNo = scimoz.lineFromPosition(p)
                minIndent = self._getIndentWidthForLine(scimoz, lineNo)
                continue # we've skipped over parens, keep going
                #indentlog.info("jumped to line %d" % lineNo)
                #return lineNo
            if scimoz.getColumn(p) <= minIndent:
                # if we've gone past the indent of the current position, it
                # may be time to call it quits. The only exception to that
                # is if the line _before_ ours ends with a continuation line.
                indentlog.debug("Got to column %d and minIndent is %d", scimoz.getColumn(p), minIndent)
                if lineNo == 0: return lineNo  # there can't be a continuation line before us
                prevLineNo = lineNo - 1
                lineStart = scimoz.positionFromLine(prevLineNo)
                lineEnd = scimoz.getLineEndPosition(prevLineNo)
                prevLine = scimoz.getTextRange(lineStart, lineEnd)
                indentlog.debug("previous line is: %r"% prevLine)
                if prevLine.rstrip().endswith('\\'):
                    indentlog.debug("skipping back")
                    p = lineEnd
                    lineNo = scimoz.lineFromPosition(p)
                    minIndent = self._getIndentWidthForLine(scimoz, lineNo)
                    continue
                indentlog.debug("we stop here")
                return lineNo
        # if we find nothing, we just give up and assume the statement
        # starts on the line containing 'position'.
        indentlog.info("Found nothing, returning the line we ended up at.")
        return lineNo
    
    def _analyzeIndentNeededAtPos(self, scimoz, pos, continueComments):
        """ This function returns:
        
          - None if a 'plain' style indent should be done
          
          - a string which should be used as the indent otherwise.
          
        """
        timeline.enter('_analyzeIndentNeededAtPos')
        try:
            if scimoz.getColumn(pos) == 0:
                # If we're in column zero, do me no favors
                return None
            curLineNo = scimoz.lineFromPosition(pos)
            lineStart = scimoz.positionFromLine(curLineNo)
            lineEnd = scimoz.getLineEndPosition(curLineNo)
            curLine = scimoz.getTextRange(lineStart, lineEnd)
            
            inBlockCommentIndent, inLineCommentIndent = self._inCommentIndent(scimoz, pos, continueComments)

            # see if we're in a block comment, and do 'inside-comment' indents
            if inBlockCommentIndent is not None:
                indentlog.info("we're in a block comment")
                return inBlockCommentIndent

            shouldIndent = self._shouldIndent(scimoz, pos)
            if shouldIndent is not None:
                indentlog.info("detected indent")
                return shouldIndent
            indentlog.info("did not detect indentation")

            while 1:
                jumped = self._skipOverParentheticalStatement(scimoz, pos)
                log.debug("pos = %d, jumped = %d" % (pos, jumped))
                if pos == jumped:
                    log.debug('breaking out of the jumping code')
                    break # nothing to go back to.
                # recurse from the beginning of the statement
                return self._analyzeIndentNeededAtPos(scimoz, jumped, continueComments)
    
            # We're not indenting.  We may be doing a line-up,
            # having just finished a line-up
            # closing a comment, or doing a comment prefix
    
            shouldLineUp = self._shouldLineUp(scimoz, pos)
            if shouldLineUp is not None:
                indentlog.info("detected line-up")
                return shouldLineUp
    
            indentlog.info("did not detect line-up")
    
                
            ## We may just have finished a multi-line statement.  We need
            ## to align with the line that started the statement.
            #finishedMultiLineStatement = self._finishedMultiLineStatement(scimoz, pos)
            #if finishedMultiLineStatement is not None:
            #    indentlog.info("detected multi line statement")
            #    return finishedMultiLineStatement
            #
            #indentlog.info("did not detect end of multi-line statement")

            # See if we just closed a block comment, in which case we need to
            # dedent by a variable amount depending on the commenting style.

            finishedBlockCommentIndents = self._finishedBlockComment(scimoz, pos)
            if finishedBlockCommentIndents is not None:
                indentlog.info("detected block comment close")
                return finishedBlockCommentIndents[0]
            indentlog.info("did not detect block comment close")

            # See if we just closed a block comment, in which case we need to
            # dedent by a variable amount depending on the commenting style.

            continuationLineIndent = self._continuationLineIndent(scimoz, pos)
            if continuationLineIndent is not None:
                indentlog.info("detected continuation line indent")
                return continuationLineIndent
            indentlog.info("did not detect continuationLineIndent")

            # see if we're in a comment, and do 'inside-comment' indents
            if inLineCommentIndent is not None:
                indentlog.info("we're in a comment")
                return inLineCommentIndent

            indentingOrDedentingStatementIndent = self._indentingOrDedentingStatement(scimoz, pos)
            if indentingOrDedentingStatementIndent is not None:
                indentlog.info("detected indenting/dedenting statement")
                return indentingOrDedentingStatementIndent
            indentlog.info("did not detect indenting/dedenting statement")

            indentlog.info("not in comment, doing plain")
            return None
        finally:
            timeline.leave('_analyzeIndentNeededAtPos')

    def _continuationLineIndent(self, scimoz, pos):
        """ This function looks to see if the line we just ended ends
        with a continuation symbol ('\' in all languages AFAIK).
        
        If yes, it looks to see if the line _before_ ours ends in a \
        as well, and if so, returns that line's indent.
        
        If not, it looks at the beginning of the current _logical_ line
        for an assignment (<>, =, or any of [|&<>]= ).  If it finds one, it returns
        an indent corresponding to the first non-whitespace character
        after the = sign.  If not, it returns an indent corresponding
        to the first non-whitespace characted in that line.
        """
        lineNo = scimoz.lineFromPosition(pos)
        lineStart = scimoz.positionFromLine(lineNo)
        lineEnd = scimoz.getLineEndPosition(lineNo)
        curLine = scimoz.getTextRange(lineStart, lineEnd)
        if not curLine.rstrip().endswith('\\'):
            #print "last line did NOT end with a \\"
            return None  # easy answer
        if lineNo:
            prevLineNo = lineNo - 1
            lineStart = scimoz.positionFromLine(prevLineNo)
            lineEnd = scimoz.getLineEndPosition(prevLineNo)
            prevLine = scimoz.getTextRange(lineStart, lineEnd)
            log.info("prevLine = %r", prevLine)
            if prevLine.rstrip().endswith('\\'):
                #print "last line is %r, ends with \\" % prevLine
                return self._getIndentForLine(scimoz, lineNo)
        m = self.assignmentRE.match(curLine)
        if m is None:
            #print "match failed, line = %r" % curLine
            return self._getIndentForLine(scimoz, lineNo)
        return self._makeIndentFromWidth(scimoz,
                                         len(m.group(1)) +
                                         len(m.group(2)) +
                                         len(m.group(3)))
    assignmentRE = re.compile(r"(.+?)(!=|<=|>=|<>|=|&=)(\s*)(.*)")

    firstWordRE = re.compile(r"\s*(\w*)");
    def _indentingOrDedentingStatement(self, scimoz, pos):
        """ This function looks to see if we just finished a dedenting
        statement (e.g. a pass, return, raise, yield in Python)
        or an indenting statement (e.g. an if or else that is not
        followed by a brace). 
        """
        lineNo = self._statementStartingLineFromPos(scimoz, pos-1)
        lineStart = scimoz.positionFromLine(lineNo)
        lineEnd = scimoz.getLineEndPosition(lineNo)
        curLine = scimoz.getTextRange(lineStart, lineEnd)
        indentlog.debug('got curLine = %r', curLine)
        wordMatch = self.firstWordRE.match(unicode(curLine))
        if wordMatch and wordMatch.group(1) in self._indenting_statements:
            return 1
        if wordMatch and wordMatch.group(1) in self._dedenting_statements:
            lineNo = scimoz.lineFromPosition(pos)
            currentIndentWidth = self._getIndentWidthForLine(scimoz, lineNo)
            indent = scimoz.indent
            if indent == 0:
                log.error('indent was 0, defaulting to 8')
                indent = 8 # XXX
            indentLevel, extras = divmod(currentIndentWidth, indent)
            if indentLevel and not extras:
                indentLevel -= 1
            nextIndentWidth = indentLevel * scimoz.indent
            return self._makeIndentFromWidth(scimoz, nextIndentWidth)
        return None

    def _finishedBlockComment(self, scimoz, pos):
        """ This function looks to see if we just ended a block comment.
        If not, it returns (None, None).
        If yes, it returns two indents -- the indent which is appropriate
        for the next line, and the indent of the beginning of the block
        statement, e.g.:
        
            ....x = 3 /* this is the start
                         and so
            ..........*/
                
        the first is four spaces, the second ten.  the first is used
        to process the newline, the second to shift the end of the
        block comment if is not preceded by non-whitespace chars on
        the same line.
        
        """
        lineNo = scimoz.lineFromPosition(pos)
        startOfLine = scimoz.positionFromLine(lineNo)
        p = pos
        data = scimoz.getStyledText(startOfLine, pos)
        while p >= startOfLine:
            p -= 1
            # move back until we hit a character.
            char = data[(p-startOfLine)*2]
            if char in ' \t': continue
            style = ord(data[(p-startOfLine)*2+1]) & self.stylingBitsMask
            if style not in self._block_comment_styles:
                # we hit a non-comment character -- it can't be the
                # end of a block comment
                indentlog.info("we hit a non-comment character -- it can't be the end of a block comment")
                return None
            # see if our current position matches the ends of block comments
            blockCommentPairs = self.commentDelimiterInfo['block']
            for blockCommentStart, blockCommentEnd in blockCommentPairs:
                for i in range(len(blockCommentEnd)):
                    c = blockCommentEnd[len(blockCommentEnd)-i-1]
                    if c != data[(p-i-startOfLine)*2]:
                        indentlog.info("looking for %s, got %s",
                                       data[(p-i-startOfLine)*2],
                                       c)
                        return None
                # we have a comment end!
                # find the matching comment start
                text = scimoz.getTextRange(0, p)
                startOfComment = text.rfind(blockCommentStart)
                if startOfComment == -1:
                    indentlog.info("could not find the beginning of the block comment")
                    return None
                nextLineIndent = self._getIndentForLine(scimoz,
                                                scimoz.lineFromPosition(startOfComment))
                adjustIndent = self._makeIndentFromWidth(scimoz,
                                                   scimoz.getColumn(startOfComment))
                return nextLineIndent, adjustIndent
        return None
    
    def shiftRegionByDelta(self, scimoz, startLineNo, endLineNo, delta):
        for lineNo in range(startLineNo, endLineNo+1):
            indentWidth = self._getIndentWidthForLine(scimoz, lineNo)
            indentWidth = max(0, indentWidth+delta)
            indent = self._makeIndentFromWidth(scimoz, indentWidth)
            start = scimoz.positionFromLine(lineNo)
            end = scimoz.getLineEndPosition(lineNo)
            line = scimoz.getTextRange(start, end)
            for i in range(len(line)):
                if line[i] not in ' \t': break
            line = indent + line[i:]
            scimoz.anchor = start
            scimoz.currentPos = end
            scimoz.replaceSel(line) # optimize
        
    def _inCommentIndent(self, scimoz, pos, continueComments=0):
        # this function returns two arguments -- if the first
        # is non-None, then we're in a block comment and should
        # indent the next line w/ the returned value.  If the
        # second is non None, then we're in a line comment and
        # _if no other indentation takes precendence_, we should
        # indent w/ the returned value.  If both values are None,
        # then we're in no kind of comment.
        style = scimoz.getStyleAt(pos) & self.stylingBitsMask
        if style not in self._comment_styles:
            return None, None
        
        # determine if we're in a line-style comment or a block-style comment
        if style in self._block_comment_styles:
            commentType = 'block'
            indentlog.debug("in block comment style")
            commentStartMarker = self.commentDelimiterInfo[commentType][0][0]
        else:
            commentType = 'line'
            indentlog.debug("in line comment style")
            if commentType not in self.commentDelimiterInfo:
                log.error("The style %d is not recognized as a block comment, but there are no line-style comments for this language", style)
                return None, None
            commentStartMarker = self.commentDelimiterInfo[commentType][0]

        curLineNo = scimoz.lineFromPosition(pos)
        lineStart = scimoz.positionFromLine(curLineNo)
        lineEnd = scimoz.getLineEndPosition(curLineNo)
        curLine = scimoz.getTextRange(lineStart, lineEnd)
        # Does the comment start on this line?
        commentStart = curLine.find(commentStartMarker)
        if commentType == 'block' and commentStart == -1:
            # The block marker we looked for didn't start on this line
            # 
            # It could be that we're in a line comment type, but the lexers
            # aren't rich enough to tell the difference.
            #
            # We'll look for the beginning of a line marker and treat
            # the comment as a line
            if 'line' in self.commentDelimiterInfo:
                commentType = 'line'
                commentStartMarker = self.commentDelimiterInfo[commentType][0]
                if commentStartMarker:
                    commentStart = curLine.find(commentStartMarker)

        # We want to find out what part of the beginning of the line
        # we want to duplicate -- whitespace, as well as any "markup"
        # characters, as in:
        #  /*<-- that's the comment start marker
        #   *<-- that's markup
        #
        markup = self.commentDelimiterInfo.get('markup', '')
        duplicate = ' \t' + markup
        
        # did the comment start on this line?  If no, then just
        # do an indent that duplicates whatever markup there is.
        if commentStart == -1: # look for 'markup' that should be duplicated
            if markup == '': # just do a standard indent
                return None, None
            indent = ''
            curLineUntilPos = scimoz.getTextRange(lineStart, pos)
            for ch in curLineUntilPos:
                if ch in duplicate:
                    indent += ch
                else:
                    break
            if commentType == 'line':
                return None, indent.expandtabs(scimoz.tabWidth)
            else:
                return indent.expandtabs(scimoz.tabWidth), None

        rest = curLine[commentStart+len(commentStartMarker):]
        indentWidth =len(curLine[:commentStart].expandtabs(scimoz.tabWidth))
        if commentType == 'line':
            # the comment started on this line -- if the line consisted of nothing but
            # comments, then continue it (except if continueComments is false)
            # If there was code on the line, then don't do anything special.
            fromStartToCommentStart = curLine[0:commentStart]
            if not continueComments or fromStartToCommentStart.strip():
                return None, None
            indent = self._makeIndentFromWidth(scimoz, indentWidth) + commentStartMarker
        else:
            indent = self._makeIndentFromWidth(scimoz, indentWidth) + len(commentStartMarker)*' '

        while rest[0:len(commentStartMarker)] and rest[0:len(commentStartMarker)] == commentStartMarker:
            if commentType == 'line':
                indent += commentStartMarker
            else:
                indent += len(commentStartMarker)*' '
            rest = rest[len(commentStartMarker):]
        while rest[0:1] and rest[0:1] in duplicate:
            indent += rest[0:1]
            rest = rest[1:]
        if commentType == 'line':
            return None, indent
        else:
            return indent, None

    def computeIndent(self, scimoz, indentStyle, continueComments):
        try:
            if indentStyle == 'none':
                return ''
            if self.supportsSmartIndent == 'text' or indentStyle == 'plain':
                return self._getPlainIndent(scimoz)
            if self.supportsSmartIndent in ('brace', 'python', 'XML'):
                return self._getSmartBraceIndent(scimoz, continueComments)
        except Exception, e:
            log.warn("Got exception computing indent", exc_info=1)
            return ''

    def _getSmartBraceIndent(self, scimoz, continueComments):
        """ return the indent for the next line using the 'smart' algorithm"""
        currentPos = scimoz.currentPos
        timeline.enter('_getSmartBraceIndent')
        if (scimoz.getColumn(currentPos) == 0):
            return None
        try:
            analysis = self._analyzeIndentNeededAtPos(scimoz, currentPos, continueComments)
            if analysis == None:
                return self._getPlainIndent(scimoz)
            # did we get a special prefix?  If yes, use that as the indent
            if analysis == 1: # doing an indent
                # need to find the beginning of the logical statement.
                lineNo = self._statementStartingLineFromPos(scimoz, currentPos-1)
                indentlog.info("doing an INDENT to the indent of line: %d" % lineNo)
                currentIndentWidth = self._getIndentWidthForLine(scimoz, lineNo)
                nextIndentWidth = (divmod(currentIndentWidth, scimoz.indent)[0] + 1) * scimoz.indent
                return self._makeIndentFromWidth(scimoz, nextIndentWidth)
            elif analysis == -1: # doing an dedent
                lineNo = self._statementStartingLineFromPos(scimoz, currentPos-1)
                #lineNo = scimoz.lineFromPosition(currentPos)
                indentlog.info("doing an DEDENT from the indent of line: %d" % lineNo)
                currentIndentWidth = self._getIndentWidthForLine(scimoz, lineNo)
                indent = scimoz.indent
                if indent == 0:
                    log.error('indent was 0, defaulting to 8')
                    indent = 8 # XXX
                indentLevel, extras = divmod(currentIndentWidth, indent)
                if indentLevel and not extras:
                    indentLevel -= 1
                nextIndentWidth = indentLevel * scimoz.indent
                return self._makeIndentFromWidth(scimoz, nextIndentWidth)
            # return whatever we got
            return analysis
        finally:
            timeline.leave('_getSmartBraceIndent')

    _getSmartPythonIndent = _getSmartBraceIndent
    _getSmartXMLIndent = _getSmartBraceIndent
    
    def _getPlainIndent(self, scimoz):
        """ return the indent for the next line using the 'plain' algorithm:
             return the 'equivalent' indent as the current logical line """
        log.info("doing _getPlainIndent")
        lineNo = self._statementStartingLineFromPos(scimoz, scimoz.currentPos)
        return self._makeIndentFromWidth(scimoz,
                                         len(self._getIndentForLine(scimoz, lineNo, scimoz.currentPos).expandtabs(scimoz.tabWidth)))

    def _getIndentWidthForLine(self, scimoz, lineNo, upto=None):
        indent = self._getIndentForLine(scimoz, lineNo, upto)
        return len(indent.expandtabs(scimoz.tabWidth))

    def _getIndentForLine(self, scimoz, lineNo, upto=None):
        lineStart = scimoz.positionFromLine(lineNo)
        if upto:
            lineEnd = min(upto, scimoz.getLineEndPosition(lineNo))
        else:
            lineEnd = scimoz.getLineEndPosition(lineNo)
        line = scimoz.getTextRange(lineStart, lineEnd)
        indentLength = len(line)-len(line.lstrip())
        indent = line[:indentLength]
        return self._makeIndentFromWidth(scimoz,
                                         len(indent.expandtabs(scimoz.tabWidth)))

    def _getIndentWidthFromIndent(self, scimoz, indentstring):
        """ Given an indentation string composed of tabs and spaces,
        figure out how long it is (get the tabWidth from scimoz) """
        expandedstring = indentstring.expandtabs(scimoz.tabWidth)
        return len(expandedstring)

    def _makeIndentFromWidth(self, scimoz, width):
        """ returns a string consisting of however many tabs and spaces
        are needed to make a width of `width`, using the useTabs pref
        and the scimoz tabWidth
        """
        if scimoz.useTabs:
            tabWidth = scimoz.tabWidth
            if tabWidth == 0: tabWidth = 8 # XXX
            numtabs, numspaces = divmod(width, tabWidth)
            return '\t'*numtabs + ' '*numspaces
        else:
            return ' '*width

    def keyPressed(self, ch, scimoz):
        charPos = scimoz.currentPos - 1
        style = scimoz.getStyleAt(charPos) & self.stylingBitsMask
        indentlog.debug("ch=%s, style=%d, indent_close_chars=%r, indent_close_styles=%r",
                       ch, style, self._indent_close_chars, self._indent_close_styles)
        if (ch in self._indent_close_chars and 
            style in self._indent_close_styles):
            indentlog.debug("got a closing brace")
            # need to find the matching brace
            braceMatch = scimoz.braceMatch(charPos)
            if braceMatch == -1:
                indentlog.debug("no matching brace")
                # no match, do nothing
                return
            startLine = self._statementStartingLineFromPos(scimoz, braceMatch)
            indent = self._getIndentForLine(scimoz, startLine)
        elif (ch in self._lineup_close_chars and 
            style in self._lineup_close_styles):
            braceMatch = scimoz.braceMatch(charPos)
            if braceMatch == -1:
                # no match, do nothing
                return
            
            indent = self._makeIndentFromWidth(scimoz, scimoz.getColumn(braceMatch))
        elif style in self._comment_styles and 'block' in self.commentDelimiterInfo:
            blockCommentEndMarker = self.commentDelimiterInfo['block'][0][-1]
            if blockCommentEndMarker.endswith(ch):
                indents = self._finishedBlockComment(scimoz, charPos+1)
                if indents is not None:
                    indent = indents[1]
                    lineNo = scimoz.lineFromPosition(charPos-1)
                    startOfLine = scimoz.positionFromLine(lineNo)
                    stuffToLeft = scimoz.getTextRange(startOfLine, 1 + charPos - len(blockCommentEndMarker))
                    if not stuffToLeft.strip(): # XXX pref: electric-comment
                        # we can align the comment closing before we do the newline
                        scimoz.targetStart = scimoz.positionFromLine(lineNo)
                        scimoz.targetEnd = charPos + 1 - len(blockCommentEndMarker)
                        scimoz.replaceTarget(len(indent), indent)
                else:
                    return
            else:
                return
        else: # it's just a regular character, just return.
            return
        # at this point we have the indent
        start = scimoz.positionFromLine(scimoz.lineFromPosition(charPos))
        end = charPos
        stuffToLeft = scimoz.getTextRange(start, end)
        if stuffToLeft.strip():
            return # there's stuff to our left, just let it be
        scimoz.targetStart = start
        scimoz.targetEnd = end
        scimoz.replaceTarget(len(indent), indent)

    def lookForHotspots(self, scimoz):
        # here we'll look for common things like URLs
        pos = scimoz.currentPos
        curLineNo = scimoz.lineFromPosition(pos)
        lineStart = scimoz.positionFromLine(curLineNo)
        lineEnd = scimoz.getLineEndPosition(curLineNo)
        curLine = scimoz.getTextRange(lineStart, lineEnd)
        firstHit = curLine.find('http://')
        mask = int('1'*self.styleBits, 2)
        styleNum = int('1'*self.styleBits, 2)-1
        scimoz.styleSetHotSpot(styleNum, 1)
        scimoz.styleSetFore(styleNum, 0x666666)
        scimoz.styleSetBack(styleNum, 0xeeeeee)
        scimoz.styleSetBold(styleNum, 1)
        scimoz.setHotspotActiveUnderline(1)
        while firstHit != -1:
            start = firstHit
            end = start + len(curLine[firstHit:].split()[0])
            # Set up the style
            scimoz.startStyling(lineStart + start, mask);
            scimoz.setStyling(end-start, styleNum);
            curLine = curLine[end:]
            firstHit = curLine.find('http://')


def _findIndent(scimoz, bitmask, chars, styles, comment_styles, tabWidth):
    """
    This code is fairly sophisticated, and a tad tricky.  Here's how it works.
    We're looking for an "indenting line" followed by an "indented line".
    Indenting lines are defined as lines that end in a special character in a special style,
    ignoring trailing whitespace and comments.  For example,
            if foo() { // asdasd
    is an indenting line in C++, but:
            if foo() // {
    is not, since the { character would not be styled as an 'indenting char'.

    Indented lines are lines that contain non-comment and non-whitespace characters
    which are indented by a positive amount compared to the indented line.  Thus
    in the code:

        [scriptable, uuid(96A8CC18-168B-4137-8435-E633D13F7925)]
        interface koILanguageService : nsISupports {
                // template interface for language services
        };
        
        [scriptable, uuid(29098AA5-7F16-4671-823D-FC0817250A7C)]
        interface koILanguage : nsISupports {
          readonly attribute string name;

    The _last_ line is the first valid 'indented' line, and so the computed
    indent is 2 (the first interface definition contains only comments, so is
    ignored.
  
    The code ignores indenting lines that are not followed by an indented line 
    
    The arguments are:
        scimoz:
            the scintilla object with the buffer being processed
        chars:
            a sequence of _unicode_ characters which are used in the
            specific language to indicate block openings (typically u'{')
        styles:
            a sequence of Scintilla style numbers which are valid for
            block-opening characters
        comment_styles:
            a sequence of Scintilla style numbers which corresponds
            to comments or other code to be ignored
        tabWidth:
            what a tab character should be counted as (almost always 8).
    
    At most the first 100 lines are looked at.
    
    As a side effect, the first 100 lines of the buffer will be 'colourised' if they
    are not already.
    """
    timeline.enter('_findIndent')
    try:
        indenting = None
        # first, colourise the first 100 lines at most so we have
        # styling information.
        N = min(100, scimoz.lineCount-1)
        end = scimoz.getLineEndPosition(N)
        scimoz.colourise(0, end)
        data = scimoz.getStyledText(0, end)
        # data is a list of (character, styleNo)
    
        for lineNo in range(N):
            # the outer loop tries to find the 'indenting' line.
            if not scimoz.getLineIndentation(lineNo+1): # skip unindented lines
                continue
            lineEndPos = scimoz.getLineEndPosition(lineNo)
            lineStartPos = scimoz.positionFromLine(lineNo)
            try:
                # we'll look for each character in the line, going from
                # the back, for an 'indenting' character
                for pos in range(lineEndPos, lineStartPos-1, -1):
                    char = data[pos*2]
                    if char in string.whitespace: # skip whitespace
                        continue
                    style = ord(data[pos*2+1]) & bitmask
                    if style in comment_styles: # skip comments
                        continue
                    if (char in chars) and (style in styles):
                        # we found that the first 'interesting'
                        # character from the right of the line is an
                        # indent-causing character
                        indenting = scimoz.getTextRange(lineStartPos, lineEndPos)
                        log.info("Found indenting line: %r" % indenting)
                        # look for an indented line after this line
                        guess, foundTabs = _findIndentedLine(scimoz,
                                                             bitmask,
                                                             N, lineNo, indenting,
                                                             comment_styles, tabWidth, data)
                        if guess is not None:
                            # if the indent is a divisor of the tab width, then we should check
                            # if there are tabs used for indentation
                            usesTabs = 0
                            if tabWidth % guess == 0:
                                for lineNo in range(N):
                                    lineEndPos = scimoz.getLineEndPosition(lineNo)
                                    lineStartPos = scimoz.positionFromLine(lineNo)
                                    # skip lines that aren't indented at all
                                    if scimoz.getWCharAt(lineStartPos) not in ' \t':
                                        continue
                                    line = scimoz.getTextRange(lineStartPos, lineEndPos)
                                    front = line[:len(line)-len(line.lstrip())]
                                    if '\t' in front or u'\t' in front:
                                        usesTabs = 1
                                        break
                            return guess, usesTabs
                        else:
                            # probably an empty block
                            raise "NextLine"
                    else:
                        # We've found a character which is not a block opener
                        # so this can't be an indenting line.
                        raise "NextLine"
            except "NextLine":
                continue
        log.info("Couldn't find indentation information from the file")
        return 0, 0
    finally:
        timeline.leave('_findIndent')

def _findIndentedLine(scimoz, bitmask, N, lineNo, indenting, comment_styles, tabWidth, data):
    """
    This function looks through the 'scimoz' buffer until at most the line number
    'N' for a line which is 'indented' relative to the 'indenting' line, ignoring
    characters styled as one of the styles in 'comment_styles'.
    """
    indentedLineNo = lineNo + 1
    guess = None
    for indentedLineNo in range(indentedLineNo, N):
        if not scimoz.getLineIndentation(indentedLineNo): # skip unindented lines
            continue
        lineEndPos = scimoz.getLineEndPosition(indentedLineNo)
        lineStartPos = scimoz.positionFromLine(indentedLineNo)
        # we want to skip characters that are just comments or just whitespace
        for pos in range(lineEndPos, lineStartPos-1, -1):
            char = data[pos*2]
            if char in string.whitespace: # skip whitespace
                continue
            style = ord(data[pos*2+1]) & bitmask
            if style in comment_styles: # skip comments
                continue
            # We have an indenting and an indented.
            indented = scimoz.getTextRange(lineStartPos, lineEndPos)
            log.info("opener: %r" % indenting)
            log.info("indented: %r" % indented)
            raw, indentsmall, foundTabsInOpener = classifyws(indenting, tabWidth)
            raw, indentlarge, foundTabsInIndented = classifyws(indented, tabWidth)
            foundTabs = foundTabsInOpener or foundTabsInIndented
            guess = indentlarge - indentsmall
            # If the guess is reasonable, use it -- indents less than
            # 2 don't qualify.
            if guess > 1:
                return guess, foundTabs
            else:
                log.warn("Found non-positive guess")
                return None, None
    return None, None
    
# taken from IDLE
# Look at the leading whitespace in s.
# Return pair (# of leading ws characters,
#              effective # of leading blanks after expanding
#              tabs to width tabwidth)

def classifyws(s, tabwidth):
    foundTabs = 0
    raw = effective = 0
    for ch in s:
        if ch == ' ':
            raw = raw + 1
            effective = effective + 1
        elif ch == '\t':
            foundTabs = 1
            raw = raw + 1
            effective = (effective / tabwidth + 1) * tabwidth
        else:
            break
    return raw, effective, foundTabs

