# Copyright (c) 2000-2003 ActiveState Corporation.
# See the file LICENSE.txt for licensing information.

"""Provide to Python code what uriparse.js provides to JavaScript code.

As for uriparse.js, koIFileEx is used to handle path/URI conversion.
"""

import os
import sys
import threading
import urlparse
import re

from xpcom import components


#---- globals

mutex = threading.Lock()
_gKoFileEx = None # koFileEx singleton instance



#---- internal support routines

def _getKoFileEx():
    global _gKoFileEx
    if _gKoFileEx is None:
        _gKoFileEx = components.classes["@activestate.com/koFileEx;1"]\
                               .createInstance(components.interfaces.koIFileEx)
    _gKoFileEx.URI = None # clear the instance data
    return _gKoFileEx



#---- public interface

# Get the URI representation of the given local file path.
#
#  "localPath" must be a local file path.
#
# Returns the URI for the given path or raises a ValueError if "localPath" is
# not a local path.
#
# Examples:
#  D:\trentm\foo.txt -> file:///D:/trentm/foo.txt
#  \\planer\d\trentm\tmp\foo.txt -> file://planer/d/trentm/tmp/foo.txt
#  file:///D:/trentm/foo.txt -> throws exception
#  ftp://ftp.activestate.com/ActivePython -> throws exception
#
def localPathToURI(localPath):
    mutex.acquire()
    try:
        koFileEx = _getKoFileEx()
        koFileEx.path = localPath
        if os.path.normpath(koFileEx.path) != os.path.normpath(localPath):
            raise ValueError("'%s' does not appear to be a proper local path"
                             % localPath)
        return koFileEx.URI
    finally:
        mutex.release()


# Get the local file path for the given URI.
#
#  "uri" may be a URI for a local file or a local path.
#
# Returns the local file path or raises an exception if there is no local file
# representation for that URI. Note: I would rather this explicitly raised if
# "uri" were a local path, but koIFileEx does not work that way.
#
# Examples:
#  D:\trentm\foo.txt -> D:\trentm\foo.txt
#  file:///D:/trentm/foo.txt -> D:\trentm\foo.txt
#  \\planer\d\trentm\tmp\foo.txt -> file://planer/d/trentm/tmp/foo.txt
#  file://planer/d/trentm/tmp/foo.txt -> file://planer/d/trentm/tmp/foo.txt
#  ftp://ftp.activestate.com/ActivePython -> throws exception
#
def URIToLocalPath(uri):
    mutex.acquire()
    try:
        koFileEx = _getKoFileEx()
        koFileEx.URI = uri
        if koFileEx.scheme != "file":
            raise ValueError("'%s' does not have a local path" % uri)
        return koFileEx.path;
    finally:
        mutex.release()


# Fix the URI using Mozilla's URI fixup service.
#
# Known fixes:
#   - quotes spaces in the URI
#       file:///spam and eggs.txt -> file:///spam%20and%20eggs.txt
#   - adds two leading slashes to file URI with only one
#       file:/foo/bar.txt -> file:///foo/bar.txt
#
def fixupURI(uri):
    fixupSvc = components.classes["@mozilla.org/docshell/urifixup;1"]\
                         .getService(components.interfaces.nsIURIFixup)
    try:
        fixupURI = fixupSvc.createFixupURI(uri,
                        components.interfaces.nsIURIFixup.FIXUP_FLAG_NONE);
        if fixupURI.spec:
            uri = fixupURI.spec;
            re.sub("^(file:/+\w)|", "\1:", uri)
    except Exception, ex:
        log.debug("nsIURIFixup could not fixup '%s': %s", uri, ex)
        # Leave uri alone (presumably to fail below!)
    return uri


def _cleanfileURL(url):
    if url.startswith('file:/'):
        if not url.startswith('file:///'):
            url = 'file:///' + url[len('file:/'):]
    return url
    
# The "RelativizeURL" method is an ugly abomination. Do NOT use it.
# It is currently only used in koProject.py. I would have removed this
# dependency but I am frightened of the implementation. --TM
def RelativizeURL(origbaseurl, origurl):
    "Take a URL, and return a relative path to _our_ URL if possible, and return it unchanged otherwise"
    # XXXXXXXXXXXXXXX This algorithm is not smart enough.  If you change this algorithm, also change the "joining"
    # done in koProject.py (see XXX)
    baseurl = _cleanfileURL(origbaseurl)
    url = _cleanfileURL(origurl)
    (base_addressing_scheme, base_network_location, base_path, base_parameters, base_query, base_fragment_identifier) = urlparse.urlparse(url)
    (addressing_scheme, network_location, path, parameters, query, fragment_identifier) = urlparse.urlparse(url)
    if addressing_scheme not in ('ftp', 'file'):
        # we only relativize FTP and local files
        #print "NOT THE RIGHT addressing_scheme", addressing_scheme
        return url
    if addressing_scheme != base_addressing_scheme:
        #print "NOT THE SAME addressing_scheme", addressing_scheme, base_addressing_scheme
        return url
    if network_location != base_network_location:
        #print "NOT THE SAME network_location", network_location, base_network_location
        return url
    if addressing_scheme == 'file':
        # we're dealing with files on _this_ computer, so we should use the platform-native file comparisons methods
        base_file = URIToLocalPath(baseurl)
        other_file = URIToLocalPath(url)
        norm_base_file = os.path.normcase(base_file)
        norm_other_file = os.path.normcase(other_file)
        common_prefix = os.path.commonprefix((norm_base_file, norm_other_file))
        if common_prefix == norm_base_file:
            # the other is a child of the base
            return url[len(baseurl):]
        # if I don't know how to at this point, I give up and return it as is (absolute)
        return url


if __name__ == "__main__":
    def test_localpath(localPath):
        try:
            uri = localPathToURI(localPath)
        except Exception, ex:
            uri = ex
        print "localPathToURI('%s'): '%s'" % (localPath, uri)
        if not isinstance(uri, Exception):
            try:
                path = URIToLocalPath(uri)
            except Exception, ex:
                path = ex
            print "URIToLocalPath('%s'): '%s'" % (uri, path)

    def test_url(originalURL):
        try:
            path = URIToLocalPath(originalURL)
        except Exception, ex:
            path = ex
        print "URIToLocalPath('%s'): '%s'" % (originalURL, path)
        if not isinstance(path, Exception):
            try:
                uri = localPathToURI(path)
            except Exception, ex:
                uri = ex
            print "localPathToURI('%s'): '%s'" % (path, uri)

        fixed = fixupURI(originalURL)
        print "fixupURI('%s'): '%s'" % (originalURL, fixed)

    print "Testing 'uriparse.py'."
    tests = [
        ("path", "Normal windows local path", r"c:\test\test.txt"),
        ("path", "Windows UNC path", r"\\planer\d\trentm\tmp\foo.txt"),
        ("url",  "Windows UNC URI (2 slashes)", "file://planer/d/trentm/tmp/foo.txt"),
        ("url",  "Windows UNC URI (5 slashes)", "file://///planer/d/trentm/tmp/foo.txt"),
        ("url",  "File URI with spaces", "file://breakfast/spam and eggs.txt"),
        #("url",  "Windows UNC URI (1 slashes)", "file:/planer/d/trentm/tmp/foo.txt"),
        #("url",  "Windows UNC URI (3 slashes)", "file:///planer/d/trentm/tmp/foo.txt"),
        #("url",  "Windows UNC URI (4 slashes)", "file:////planer/d/trentm/tmp/foo.txt"),
    ]
    for scheme, desc, testpath in tests:
        if scheme == "path":
            print
            print "---- %s: '%s'" % (desc, testpath)
            test_localpath(testpath)
        elif scheme == "url":
            print
            print "---- %s: '%s'" % (desc, testpath)
            test_url(testpath)

    print
    path = RelativizeURL(r'c:\test',r'..\test\test.txt')
    print "RelativizeURL('c:\\test', '..\\test\\test.txt'): '%s'" % path

